/***************************************************************************
 $RCSfile: api.cpp,v $
                             -------------------
    cvs         : $Id: api.cpp,v 1.25 2004/05/27 20:18:56 aquamaniac Exp $
    begin       : Sat Jun 08 2002
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "assert.h"

#include "api.h"
#include <hbcistring.h>
#include <mediumrdhbase.h>
#include <gwenhywfar/gwenhywfar.h>
#include <gwenhywfar/directory.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/debug.h>
#include "init_plugins.h"
#include <openhbci2/msgengineplugin.h>
#include <openhbci2/transporterplugin.h>
#include <time.h>

#include <fstream>

namespace HBCI {


unsigned int API::_instances=0;


API::API(bool readonly,
	 bool retrievalonly)
  :Hbci(readonly, retrievalonly)
  ,_timeout(60)
{
  Error err;

  if (!_instances) {
    // init Gwen (since auto-init doesn't work on Solaris and OSX)
    GWEN_Init();
  }
  _instances++;

  _loader=new Loader(this);
  _loader.setDescription("API::_loader");
  _loader.setObjectDescription("Loader");

  err=registerLinkedPlugins(this, &pluginManager());
  if (!err.isOk()) {
    fprintf(stderr,"Error registering plugins: %s\n",
	    err.errorString().c_str());
    throw Error("API::API",err);
  }
}


API::~API() {
  // Manually delete the Pointer<Bank> list since we own these
  // objects but due to cross-referencing of Bank<->Account *and*
  // Bank<->User the Bank's reference counters won't reach zero
  // automatically.
  list<Pointer<Bank> >::iterator it;

  for (it=_banks.begin();
       it!=_banks.end();
       it++) {
    // delete object without destroying the admin structures. This keeps
    // the pointer itself valid, but if some pointer tries to use this
    // pointer we will get an exception
    (*it).release();
  }

  if (_instances) {
    _instances--;
    if (!_instances)
      GWEN_Fini();
  }
}


void API::addBank(Pointer<Bank> b){
    Pointer<Bank> nb;

    for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
        iter!=_banks.end(); iter++) {
        /* first check if the very same object already exists.
         * It's not a problem if it does, since it is the SAME.
         */
        if (*iter==b) {
#if DEBUGMODE>0
            fprintf(stderr,
                    "Will not add bank, since it already IS listed.\n");
#endif
            return;
        }
        /* then check if there already is an object with the same id data.
         * This WOULD be a problem, since that would corrupt the list.
         * The second occurence of an equal account could rarely be accessed.
         */
	if ((*iter).ref().country()==b.ref().country())
	  if ((*iter).ref().bankCode()==b.ref().bankCode())
	    throw Error("API::addBank()",
			ERROR_LEVEL_NORMAL,
			HBCI_ERROR_CODE_EXISTS,
			ERROR_ADVISE_DONTKNOW,
			"bank already exists.");
    } // for

    // otherwise simply add the bank
    nb=b;
    nb.setDescription("Entry of API::_banks");
    _banks.push_back(nb);
}


Pointer<Bank> API::findBank(int country,
			    const string &instcode) const {
  for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
      iter!=_banks.end(); iter++) {
    if ((*iter).ref().country()==country)
      if (-1!=GWEN_Text_ComparePattern((*iter).ref().bankCode().c_str(),
                                       instcode.c_str(), 0))
        return (*iter);
  }
  return 0;
}



Pointer<User> API::findUser(int country,
                            const string &instcode,
                            const string &userId) const {
  Pointer<User> user;

  for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
      iter!=_banks.end(); iter++) {
    if ((*iter).ref().country()==country)
      if (-1!=GWEN_Text_ComparePattern((*iter).ref().bankCode().c_str(),
                                       instcode.c_str(), 0)) {
        user=(*iter).ref().findUser(userId);
        if (user.isValid())
          return user;
      }
  } // for bank

  DBG_INFO(0, "User \"%d/%s/%s\" not found",
           country, instcode.c_str(), userId.c_str());
  return 0;
}



Pointer<Customer> API::findCustomer(int country,
                                    const string &instcode,
                                    const string &custId) const {
  Pointer<Customer> cust;

  for(list<Pointer<Bank> >::const_iterator iter=_banks.begin();
      iter!=_banks.end(); iter++) {
    if ((*iter).ref().country()==country)
      if (-1!=GWEN_Text_ComparePattern((*iter).ref().bankCode().c_str(),
                                       instcode.c_str(), 0)) {
        cust=(*iter).ref().findCustomer(custId);
        if (cust.isValid())
          return cust;
      }
  } // for bank

  DBG_INFO(0, "Customer \"%d/%s/%s\" not found",
           country, instcode.c_str(), custId.c_str());
  return 0;
}



int API::totalUsers () const 
{
    list<Pointer<Bank> >::const_iterator iter, end;
    int ret = 0;

    end = banks().end();
    for ( iter = banks().begin(); iter != end; iter++ ) {
	ret += (*iter).ref().users().size();
    }
    return ret;
}


Error API::loadEnvironment(const string &f,
			   unsigned int fl){
  GWEN_DB_NODE *cfg;
  Error err;

  cfg=GWEN_DB_Group_new("hbci");
  if (GWEN_DB_ReadFile(cfg, f.c_str(),
                       GWEN_DB_FLAGS_DEFAULT|
                       GWEN_PATH_FLAGS_CREATE_GROUP)) {
    GWEN_DB_Group_free(cfg);
    return Error("API::loadEnvironment()",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "Could not read config file.",
		 f);
  }
  err=_loader.ref().loadAll(cfg, fl);
  GWEN_DB_Group_free(cfg);

  return err;
}


Error API::saveEnvironment(const string &f,
			   unsigned int fl) const {
  GWEN_DB_NODE *cfg;
  Error err;
  int rv;

  cfg=GWEN_DB_Group_new("HBCI");

  err=_loader.ref().saveAll(cfg, fl);
  if (!err.isOk()) {
    GWEN_DB_Group_free(cfg);
    return err;
  }
  rv=GWEN_DB_WriteFile(cfg, f.c_str(),
                       GWEN_DB_FLAGS_DEFAULT
                       //GWEN_DB_FLAGS_COMPACT
                      );
  GWEN_DB_Group_free(cfg);
  if (rv)
    return Error("API::loadEnvironment()",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "Could not read config file.",
		 f);
  return Error();
}


Error API::_openConnection(Pointer<Transporter> tx){
  time_t startTime;
  Error err;
  Pointer<Interactor> ia;

  ia=interactor();
  if (debugLevel()>15)
    fprintf(stderr,"MessageEngineXML::_openConnection\n");

  // get start time for timeout calculation
  startTime=time(0);

  if (tx.ref().isOpen())
    return Error();

  // initiate connect
  err=tx.ref().startOpen();
  if (!err.isOk())
    return Error("MessageEngineXML::_openConnection", err);

  // wait for connect
  while (ia.ref().keepAlive()) {
    err=tx.ref().checkOpen();
    if (err.isOk())
      break;
    if (err.code()!=HBCI_ERROR_CODE_SOCKET_ERROR_TIMEOUT &&
        err.code()!=HBCI_ERROR_CODE_SOCKET_ERROR_INTERRUPT)
      return Error("MessageEngineXML::_openConnection", err);

    if (difftime(time(0), startTime)>_timeout) {
      if (debugLevel()>10)
        fprintf(stderr,
                "MessageEngineXML::_openConnection: timeout\n");
      return Error("MessageEngineXML::_openConnection", err);
    }
  } // while

  return Error();
}


Error API::_handleMessageQueue(Pointer<MessageEngine> engine,
			       Pointer<MsgQueue> mbox){
  Error err;

  try {
    if (mbox.ref().empty()) {
      // empty queue, finished ;-)
      return Error();
    }

    // execute this queue
    err=engine.ref().executeQueue(mbox);
    if (!err.isOk())
      return Error("API::_handleMessageQueue()", err);

  }
  catch(Error xerr) {
    err=xerr;
  }
  if (!err.isOk())
    _logMessage(1,"RESULT (handleMessage): "+err.errorString());
  return err;
}


void API::_logMessage(int loglevel, const string &msg) {
  if (interactor().isValid()) {
    if (loglevel<debugLevel())
      interactor().ref().logMessage(msg);
  }
}


Error API::_handleJobQueue(Pointer<Outbox> outbox,
			   Pointer<MessageEngine> engine,
			   Pointer<Customer> cust,
                           const list<Pointer<OutboxJob> > &jq){
  list<Pointer<OutboxJob> >::const_iterator it;
  list<Pointer<OutboxJob> >::const_iterator tmpit;
  Error err;
  int jobErrors=0;
  Pointer<MsgQueue> mbox;
  list<Pointer<OutboxJob> > enqueuedJobs;
  GWEN_DB_NODE *response;
  Pointer<Bank> bank;

  mbox.setDescription("API::_handleUserQueue: mbox");
  bank=cust.ref().user().ref().bank();
  assert(bank.isValid());

  it=jq.begin();
  while(1) {
    unsigned int jobsToDo;

    DBG_INFO(0, "Filling queue");
    // create new message queue
    mbox=engine.ref().queueFactory(cust);
    // enqueue all jobs, if possible
    jobsToDo=0;

    // count jobs
    jobsToDo=0;
    for (tmpit=jq.begin(); tmpit!=jq.end(); tmpit++) {
      if ((*tmpit).ref().status()==HBCI_JOB_STATUS_TODO) {
        DBG_NOTICE(0, " Job \"%s\" is pending",
                   (*tmpit).ref().name().c_str());
        jobsToDo++;
      }
    } // for
    if (!jobsToDo) {
      // all jobs done, break while loop
      DBG_NOTICE(0, "Queue done.");
      break;
    }

    DBG_NOTICE(0, "Doing %d jobs", jobsToDo);
    // enqueue as many jobs as possible
    interactor().ref().actionStarted(ACT_FILLINGQUEUE,
                                     "Preparing queue.");
    for (;;) {
      if (jobsToDo<1)
        break;

      if (it==jq.end())
        it=jq.begin();

      if ((*it).ref().status()==HBCI_JOB_STATUS_TODO) {
        HBCI_QUEUECHECK_RESULT qr;

        jobsToDo--;
        qr=(*it).ref().checkEnqueue(mbox);
        if (qr==HBCI_QUEUECHECK_RESULT_OK) {
          // enqueue job in mbox
          err=(*it).ref().enqueueJob(mbox);
          if (!err.isOk()) {
            _logMessage(1,"RESULT (EnqueueJob): "+err.errorString());
            if (err.advise()==ERROR_ADVISE_ABORT)
              return err;
            else
              jobErrors++;
          }
          else {
            // save job for later
            DBG_NOTICE(0, "Enqueued job \"%s\"", (*it).ref().name().c_str());
            enqueuedJobs.push_back(*it);
          }
        }
        else if (qr==HBCI_QUEUECHECK_RESULT_JOB) {
          // job can not be added, so try it next turn
          DBG_NOTICE(0, "Job \"%s\" does not fit into queue",
                     (*it).ref().name().c_str());
        }
        else if (qr==HBCI_QUEUECHECK_RESULT_QUEUE_FULL) {
          // current queue is full, send it
          DBG_NOTICE(0, "Queue is full, will send it");
          break;
        }
        else {
          DBG_NOTICE(0, "Unknown status");
          (*it).ref().setStatus(HBCI_JOB_STATUS_DONE);
          (*it).ref().setResult(HBCI_JOB_RESULT_FAILED);
        }
      } // if job is to be executed
      it++;
    } // for
    DBG_NOTICE(0, "jobs done");

    if (enqueuedJobs.empty()) {
      // could not enque any job, this must be an error
      DBG_INFO(0, "Could not enqueue any job.");
      return Error("API::_handleJobQueue()",
		   ERROR_LEVEL_NORMAL,
		   1,
		   ERROR_ADVISE_DONTKNOW,
		   "could not add any job to queue");
    }

    // now send the message queue to the server
    DBG_INFO(0, "Sending queue");

    interactor().ref().actionStarted(ACT_SENDINGMESSAGE,
                                     "Sending queue.");
    err=_handleMessageQueue(engine, mbox);

    // copy data from mbox to Outbox
    DBG_INFO(0, "Copying queue response to Outbox");
    response=mbox.ref().queueResponse();
    assert(response);
    // extract bank information
    interactor().ref().actionStarted(ACT_UPDATESYSTEM,
                                 "Updating system.");
    _processOutboxResponse(response, bank);
    outbox.ref().addResponse(response);
    outbox.ref().addResults(mbox.ref().results());
    GWEN_DB_Group_free(response);

    DBG_INFO(0, "Evaluating all executed jobs");
    // evaluate all jobs which have been executed
    interactor().ref().actionStarted(ACT_CHKRESULT,
                                 "Checking job result.");
    for (tmpit=enqueuedJobs.begin(); tmpit!=enqueuedJobs.end(); tmpit++) {
      // evaluate this job
      err=(*tmpit).ref().evaluate(outbox);
      if (!err.isOk())
        jobErrors++;
    } // for
    enqueuedJobs.clear();
  } // while

  if (jobErrors>0)
    return ( (jobErrors == 1) ?
	     err :
	     Error("API::_handleJobQueue()",
		   ERROR_LEVEL_NORMAL,
		   1,
		   ERROR_ADVISE_DONTKNOW,
		   "more than one job failed"));
  return Error();
}


Error API::_handleUserQueue(Pointer<Outbox> outbox,
			    Pointer<MessageEngine> engine,
                            Pointer<customerQueue> q,
                            Pointer<Transporter> conn){
  list<Pointer<OutboxJob> > jq;
  Error err;
  Error err2;
  int j;
  Pointer<Customer> cust;

  // get queue size
  j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
  if (!j)
    // nothing to be done
    return Error();

  // get corresponding customer
  cust=q.ref().customer();
  if (!cust.isValid())
    return Error("API::_handleUserQueue",
		 ERROR_LEVEL_INTERNAL,
		 HBCI_ERROR_CODE_INEXISTENT,
		 ERROR_ADVISE_ABORT,
		 "no customer in customerQueue");

  if (Hbci::debugLevel()>5)
    fprintf(stderr,"HANDLEUSERQUEUE: customer is %s\n",
	    cust.ref().custId().c_str());

  jq=q.ref().jobs();


  // open connection
  interactor().ref().actionStarted(ACT_CONTACTINGSERVER,
                               "Connecting.");
  err=_openConnection(conn);
  if (!err.isOk())
    return Error("API::_handleUserQueue", err);

  // start session
  err=engine.ref().startSession(cust, conn);
  if (!err.isOk())
    return Error("API::_handleUserQueue", err);

  // handle jobs
  err=_handleJobQueue(outbox, engine, cust, jq);

  // end session
  err2=engine.ref().endSession();
  // close connection (in case the engine did not)
  DBG_INFO(0, "Closing connection");
  interactor().ref().actionStarted(ACT_CLOSECONNECTION,
                                   "Disconnecting.");
  conn.ref().close();

  // error from job queue takes precedence
  if (!err.isOk()) {
    DBG_INFO(0, "Error in HandleJobQueue: %s",
             err.errorString().c_str());
    if (err.advise()==ERROR_ADVISE_ABORT)
      return err;
  }

  if (!err2.isOk())
    return Error("API::_handleUserQueue", err2);

  return err;
}


Error API::_handleBankQueue(Pointer<Outbox> outbox,
			    Pointer<bankQueue> q){
  list<Pointer<customerQueue> > cq;
  list<Pointer<customerQueue> >::const_iterator it;
  Pointer<Transporter> conn;
  int j;
  Pointer<Bank> bank;
  Error err;
  int jobErrors=0;
  Pointer<MessageEngine> engine;

  conn.setDescription("API::_handleBankQueue: conn");

  // get queue size
  j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
  if (!j)
    // nothing to be done
    return Error();

  // get corresponding bank
  bank=q.ref().bank();
  if (!bank.isValid())
    return Error("API::_handleBankQueue()",
		 ERROR_LEVEL_INTERNAL,
		 HBCI_ERROR_CODE_INEXISTENT,
		 ERROR_ADVISE_ABORT,
		 "no bank in bankQueue");
#if DEBUGMODE>0
  fprintf(stderr,"HANDLEBANKQUEUE: bank is %s\n",
	  bank.ref().bankCode().c_str());
#endif
  engine=bank.ref().messageEngine();
  cq=q.ref().customerQueues();
  conn=bank.ref().transporter();
  for (it=cq.begin(); it!=cq.end(); it++) {

    // handle user jobs
    err=_handleUserQueue(outbox, engine, *it, conn);
    // close connection (in case the engine did not)
    DBG_INFO(0, "Closing connection");
    conn.ref().close();
    if (!err.isOk()) {
      _logMessage(1,"HandleUserQueue: "+err.errorString());
      jobErrors++;
      if (err.advise()==ERROR_ADVISE_ABORT)
	break;
    }
  } // for

  // check for errors
  if (jobErrors>0)
    return ( (jobErrors == 1) ?
	     err :
	     Error("API::_handleBankQueue()",
		   ERROR_LEVEL_NORMAL,
		   1,
		   ERROR_ADVISE_DONTKNOW,
		   "more than one job failed"));
  return Error();
}


Error API::_handleQueue(Pointer<Outbox> q){
  list<Pointer<bankQueue> > bq;
  list<Pointer<bankQueue> >::const_iterator it;
  int j;
  int jobErrors=0;
  Error err;

  // get queue size
  j=q.ref().sizeByStatus(HBCI_JOB_STATUS_TODO);
  if (!j)
    // nothing to be done
    return Error();
  bq=q.ref().bankQueues();

  // add jobs for connection opening and closing
  j+=q.ref().bankCount()*2;

  for (it=bq.begin(); it!=bq.end(); it++) {
    err=_handleBankQueue(q, *it);
    if (!err.isOk()){
      _logMessage(1,"Bank: "+err.errorString());
      jobErrors++;
    }
  }

  if (jobErrors>0)
    return ( (jobErrors == 1) ?
	     err :
	     Error("API::_handleQueue()",
		   ERROR_LEVEL_NORMAL,
		   1,
		   ERROR_ADVISE_DONTKNOW,
		   "more than one job failed"));
  return Error();
}


Error API::executeQueue(Pointer<Outbox> queue){
  Error err;
  Pointer<Interactor> ia;

  ia=interactor();
  assert(ia.isValid());
  // FIXME: Try to get another way to determine the transport type
  if (!ia.ref().msgOpenConnection(TransportTypeTCP)) {
    return Error("API::executeQueue",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_ABORT,
                 "User aborted connection opening.");
  }
  err= _handleQueue(queue);
  // FIXME: Try to get another way to determine the transport type
  ia.ref().msgCloseConnection(TransportTypeTCP);
  // unmount the last medium (if there is one mounted)
  unmountCurrentMedium();
  return err;
}



list<Pointer<User> > 
API::getUsers(int country,
                  const string &instid,
                  const string &userid) const {
    list<Pointer<Bank> >::const_iterator it;
    list<Pointer<User> > custs;
    list<Pointer<User> >::const_iterator cit;;

    for (it=_banks.begin(); it!=_banks.end();it++) {
      if ((country==0 || (*it).ref().country()) &&
	  (-1!=GWEN_Text_ComparePattern((*it).ref().bankCode().c_str(),
					instid.c_str(),0))) {
	// add matching customers to mine
	for (cit=(*it).ref().users().begin();
	     cit!=(*it).ref().users().end(); cit++)
	  if (-1!=GWEN_Text_ComparePattern((*cit).ref().userId().c_str(),
					   userid.c_str(),false))
	    custs.push_back(*cit);
      } // if bank matches
    } // for
    return custs;
}


list<Pointer<Customer> >
API::getCustomers(int country,
                  const string &instid,
                  const string &custid) const {
  list<Pointer<Bank> >::const_iterator it;
  list<Pointer<User> >::const_iterator uit;;
  list<Pointer<Customer> >::const_iterator cit;;
  list<Pointer<Customer> > custs;

  DBG_DEBUG(0, "Looking for %s/%s", instid.c_str(), custid.c_str());
  // iterate through banks
  for (it=_banks.begin(); it!=_banks.end();it++) {
    if ((country==0 || (*it).ref().country()) &&
        (-1!=GWEN_Text_ComparePattern((*it).ref().bankCode().c_str(),
                                      instid.c_str(), 0))) {
      DBG_DEBUG(0, "Bank %s", (*it).ref().bankCode().c_str());
      // iterate through users
      for (uit=(*it).ref().users().begin();
           uit!=(*it).ref().users().end(); uit++) {
        DBG_DEBUG(0, "User %s", (*uit).ref().userId().c_str());
        // add matching customers to mine
        for (cit=(*uit).ref().customers().begin();
             cit!=(*uit).ref().customers().end(); cit++) {
          DBG_DEBUG(0, "Customer %s", (*cit).ref().custId().c_str());
          if (-1!=GWEN_Text_ComparePattern((*cit).ref().custId().c_str(),
                                           custid.c_str(), 0)) {
            DBG_DEBUG(0, "Customer matches");
            custs.push_back(*cit);
          }
          else {
            DBG_DEBUG(0, "Customer doesn't match");
          }
        } // for customer
      } // for user
    } // if bank matches
  } // for bank
  return custs;
}


Pointer<Medium> API::findMedium(const string &name) const {
  list<Pointer<Bank> >::const_iterator bit;
  list<Pointer<User> >::const_iterator cit;
  Pointer<Medium> medium;

  for (bit=_banks.begin(); bit!=_banks.end();bit++) {
    // add matching accounts to mine
    for (cit=(*bit).ref().users().begin();
	 cit!=(*bit).ref().users().end(); cit++) {
      medium=(*cit).ref().medium();
      if (medium.isValid())
	if (medium.ref().mediumName()==name)
	  return medium;
    } // for

  } // for

  // not found
  return 0;
}


Pointer<MediumPlugin> API::_findMediumPlugin(const string &name) {
  Pointer<Plugin> p;

  p=findPlugin(name, PluginTypeMedium);
  if (p.isValid())
    return p.cast<MediumPlugin>();

  return 0;
}


Pointer<MediumPlugin> API::_ensureMediumPlugin(const string &mtype) {
  Pointer<Plugin> tmp;

  DBG_INFO(0, "Ensuring medium plugin \"%s\"", mtype.c_str());
  tmp=ensurePlugin(mtype, PluginTypeMedium);
  if (tmp.isValid()) {
    return tmp.cast<MediumPlugin>();
  }
  return 0;
}


Pointer<Medium> API::mediumFromConfig(GWEN_DB_NODE *cfg) {
  Pointer<MediumPlugin> tmp;
  string mtype;
  Error err;

  mtype=GWEN_DB_GetCharValue(cfg, "mediumtype",0, "");
  if (mtype.empty())
    throw Error("API::mediumFromConfig",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_INVALID,
		ERROR_ADVISE_DONTKNOW,
		"no medium type given");

  tmp=_ensureMediumPlugin(mtype);
  DBG_INFO(0, "Letting plugin create medium");
  return tmp.ref().mediumFromConfig(cfg);
}


Error API::mediumToConfig(Pointer<Medium> m,
			  GWEN_DB_NODE *cfg) {
  Pointer<MediumPlugin> tmp;

  //fprintf(stderr,"Saving medium %s\n", m.ref().mediumTypeName().c_str());
  tmp=_findMediumPlugin(m.ref().mediumTypeName());
  if (!tmp.isValid()) {
    return Error("API::mediumToConfig",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_INVALID,
		 ERROR_ADVISE_DONTKNOW,
		 "medium type not found");
  }

  return tmp.ref().mediumToConfig(m, cfg);
}


MediumType API::mediumType(const string &mtype) {
  Pointer<MediumPlugin> tmp;

  tmp=_ensureMediumPlugin(mtype);
  return tmp.ref().mediumType();

}


Pointer<Medium> API::mediumFactory(const string &name,
				   const string &mtype) {
  Pointer<MediumPlugin> tmp;
  Pointer<Medium> m;

  m=findMedium(name);
  if (m.isValid()) {
    if (debugLevel()>10)
      fprintf(stderr, "Reusing existing medium for \"%s\"\n", name.c_str());
    m.setDescription("API::mediumFactory (reused)");
    return m;
  }
  tmp=_ensureMediumPlugin(mtype);
  return tmp.ref().mediumFactory(name);
}


Pointer<Medium> API::createNewMedium(const string &mtype,
                                     bool readonly,
				     int country,
				     const string &bankId,
				     const string &userId,
				     const string &name){
  Pointer<MediumPlugin> tmp;

  // first check arguments
  if (bankId.empty() || userId.empty() || mtype.empty())
    throw Error("API::createNewMedium()",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_INVALID,
		ERROR_ADVISE_DONTKNOW,
		"empty argument");

  tmp=_ensureMediumPlugin(mtype);
  return tmp.ref().createNewMedium(readonly, country, bankId, userId, name);
}



Pointer<MessageEngine> API::engineFactory(int country,
					  const string &bankCode,
                                          int hversion,
					  const string &engineType){
  Pointer<Plugin> tmp;
  Pointer<MessageEnginePlugin> ep;
  Pointer<MessageEngine> msgEngine;

  tmp=pluginManager().ensurePlugin(engineType, PluginTypeMessage);
  if (!tmp.isValid())
    return 0;
  ep=tmp.cast<MessageEnginePlugin>();
  msgEngine=ep.ref().engineFactory(country, bankCode, hversion);
  msgEngine.setDescription("API::engineFactory: MessageEngine");
  return msgEngine;
}


Pointer<MessageEngine> API::engineFromConfig(int country,
					     const string &bankCode,
                                             int hversion,
					     GWEN_DB_NODE *cfg){
  Pointer<Plugin> tmp;
  Pointer<MessageEnginePlugin> ep;
  Pointer<MessageEngine> msgEngine;
  string mtype;
  Error err;

  mtype=GWEN_DB_GetCharValue(cfg, "enginetype",0, "");
  if (mtype.empty())
    throw Error("API::engineFromConfig",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_INVALID,
		ERROR_ADVISE_DONTKNOW,
		"no engine type given");

  tmp=pluginManager().ensurePlugin(mtype, PluginTypeMessage);
  if (!tmp.isValid())
    return 0;
  ep=tmp.cast<MessageEnginePlugin>();
  msgEngine=ep.ref().engineFromConfig(country, bankCode, hversion, cfg);
  msgEngine.setDescription("API::engineFactory: MessageEngine");
  return msgEngine;
}


Error API::engineToConfig(Pointer<MessageEngine> m,
			  GWEN_DB_NODE *cfg){
  Pointer<Plugin> tmp;
  Pointer<MessageEnginePlugin> ep;

  tmp=pluginManager().findPlugin(m.ref().engineType(), PluginTypeMessage);
  if (!tmp.isValid())
    return Error("API::engineToConfig",
		 ERROR_LEVEL_INTERNAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "Plugin for MessageEngine not found.");

  ep=tmp.cast<MessageEnginePlugin>();
  return ep.ref().engineToConfig(m, cfg);
}




Pointer<Transporter> API::transporterFactory(const string &addr,
					     const string &port,
                                             int timeout,
					     const string &transporterType){
  Pointer<Plugin> tmp;
  Pointer<TransporterPlugin> ep;
  Pointer<Transporter> transporter;

  tmp=pluginManager().ensurePlugin(transporterType, PluginTypeTransport);
  if (!tmp.isValid())
    return 0;
  ep=tmp.cast<TransporterPlugin>();
  transporter=ep.ref().transporterFactory(addr, port, timeout);
  transporter.setDescription("API::transporterFactory: Transporter");
  return transporter;
}


Pointer<Transporter> API::transporterFromConfig(const string &address,
						const string &port,
						GWEN_DB_NODE *cfg){
  Pointer<Plugin> tmp;
  Pointer<TransporterPlugin> ep;
  Pointer<Transporter> transporter;
  string mtype;
  Error err;

  mtype=GWEN_DB_GetCharValue(cfg, "transporttype",0, "");
  if (mtype.empty())
    throw Error("API::transporterFromConfig",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_INVALID,
		ERROR_ADVISE_DONTKNOW,
		"no transporter type given");

  tmp=pluginManager().ensurePlugin(mtype, PluginTypeTransport);
  if (!tmp.isValid())
    return 0;
  ep=tmp.cast<TransporterPlugin>();
  transporter=ep.ref().transporterFromConfig(address, port, cfg);
  transporter.setDescription("API::transporterFactory: Transporter");
  return transporter;
}


Error API::transporterToConfig(Pointer<Transporter> m,
			       GWEN_DB_NODE *cfg){
  Pointer<Plugin> tmp;
  Pointer<TransporterPlugin> ep;

  tmp=pluginManager().findPlugin(m.ref().transporterTypeName(),
				 PluginTypeTransport);
  if (!tmp.isValid())
    return Error("API::transporterToConfig",
		 ERROR_LEVEL_INTERNAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "Plugin for Transporter not found.");

  ep=tmp.cast<TransporterPlugin>();
  return ep.ref().transporterToConfig(m, cfg);
}










Pointer<User> API::userFactory(Pointer<Bank> b,
			       Pointer<Medium> m,
			       const string &userid,
			       const string &userName) {
  return new User(b,m,userid, userName);
}


Pointer<Customer> API::customerFactory(Pointer<User> u,
				       const string &id,
				       const string &custName) {
  return new Customer(u,id, custName);
}


Pointer<Bank> API::bankFactory(int country,
                               const string &bankCode,
			       const string &server,
			       const string &port,
                               int hbciVersion,
                               const string &name) {
  Pointer<Bank> newBank =
    new Bank(this,country,bankCode,server,port, hbciVersion, name);
  newBank.setObjectDescription("Bank (created by API::bankFactory)");
  return newBank;
}


void API::removeBank(Pointer<Bank> b) {
    list<Pointer<Bank> >::iterator it1;

    // find bank in list and remove it
    for(it1=_banks.begin();
        it1!=_banks.end(); it1++) {
        if ((*it1)==b) {
            _banks.erase(it1);
            break;
        }
    } // for
}




void API::_addPluginsFromDirs(Pointer<MediumPluginList> pl,
			      const list<string> &dnames) {
  string path;
  string tmp;
  Error err;
  bool goon;
  int soversion;
  list<string>::const_iterator sit;


  for (soversion=OPENHBCI_SO_VERSION;
       soversion>=OPENHBCI_SO_VERSION-OPENHBCI_SO_AGE;
       soversion--) {
    for (sit=dnames.begin();
	 sit!=dnames.end();
	 sit++) {
      GWEN_DIRECTORYDATA *dir;

      if (debugLevel()>10)
	fprintf(stderr, "Checking directory \"%s\"\n", (*sit).c_str());
      path=*sit;
      path+="/";
      path+=String::num2string(soversion);
      path+="/media";

      dir=GWEN_Directory_new();
      if (GWEN_Directory_Open(dir, path.c_str())) {
	GWEN_Directory_free(dir);
	if (debugLevel()>5)
	  fprintf(stderr, "Error opening directory \"%s\"\n",
		  path.c_str());
	continue;
      }

      goon=true;
      while(goon) {
	char entry[256];

        if (GWEN_Directory_Read(dir,entry, sizeof(entry)))
	  goon=false;
	else {
	  if (-1!=GWEN_Text_ComparePattern(entry,
					   "*"PLUGIN_EXT,
					   0)) {
	    Pointer<PluginFile> pluginFile;
	    Pointer<Plugin> plugin;

	    tmp=path+"/"+entry;

	    err=pluginManager().loadPlugin(tmp,
					   PluginTypeMedium,
					   pluginFile,
					   plugin);
	    if (!err.isOk()) {
	      if (debugLevel()>5)
		fprintf(stderr, "Error loading pluginfile \"%s\" (%s)\n",
			entry,
			err.errorString().c_str());
	    }
	    else {
	      if (debugLevel()>5)
		fprintf(stderr, "Loaded pluginfile \"%s\"\n",
			tmp.c_str());
	    }
	    if (pl.ref().addPlugin(plugin.cast<MediumPlugin>()))
	      pl.ref().addPluginFile(pluginFile);
	  } // if filename ends with system EXT pattern
	}
      } // while

      if (GWEN_Directory_Close(dir)) {
	fprintf(stderr, "Error closing directory \"%s\"\n",
		path.c_str());
      }
      GWEN_Directory_free(dir);
    } // for dnames
  } // for versions

  return;
}



Pointer<MediumPluginList> API::enumerateMediumPlugins() {
  Pointer<MediumPluginList> pl;
  list<Pointer<Plugin> >::const_iterator it;
  list<string> dnames;

  pl=new MediumPluginList();

  /* add all plugins we already know */
  for (it=pluginManager().plugins().begin();
       it!=pluginManager().plugins().end(); it++) {
    if ((*it).ref().pluginType()==PluginTypeMedium)
      pl.ref().addPlugin((*it).cast<MediumPlugin>());
  }

  /* check for loadable media, if supported */
#ifdef PLUGINLOADING_SUPPORTED
  /* */

  dnames.push_back(PLUGIN_PATH);
  dnames.push_back("/usr/share/openhbci2/plugins");
  dnames.push_back("/usr/local/share/openhbci2/plugins");
  _addPluginsFromDirs(pl, dnames);
#endif // #ifdef PLUGINLOADING_SUPPORTED

  return pl;
}



Error API::_processOutboxResponse(GWEN_DB_NODE *db, Pointer<Bank> bank){
  GWEN_DB_NODE *n;
  Error err;

  assert(db);
  n=GWEN_DB_GetFirstGroup(db);
  while (n){
    if (strcasecmp(GWEN_DB_GroupName(n), "bpd")==0) {
      // found bank parameter data
      Pointer<Bank> b;
      int country;
      string bankCode;
      string bankName;

      country=GWEN_DB_GetIntValue(n, "country", 0, 280);
      bankCode=GWEN_DB_GetCharValue(n, "bankcode", 0, "");
      bankName=GWEN_DB_GetCharValue(n, "name", 0, "");
      b=findBank(country, bankCode);
      if (!b.isValid()) {
        DBG_WARN(0, "Bank \"%s\" (%d/%s) not found",
                 bankName.c_str(), country, bankCode.c_str());
      }
      else {
        int v;
        int i;

        if (!bankName.empty())
          b.ref().setName(bankName);
        for (i=0; i<16; i++) {
          v=GWEN_DB_GetIntValue(n, "versions/version", i, -1);
          if (v==-1)
            break;
          if (i==0)
            b.ref().clearHbciVersions();
          b.ref().addHbciVersion(v);
        } // for
      } // if bank is valid
    } // if bpd
    else if (strcasecmp(GWEN_DB_GroupName(n), "bankmsg")==0) {
      // store originator of the message
      DBG_DEBUG(0, "Got a message");
      GWEN_DB_SetIntValue(n,
                          GWEN_DB_FLAGS_DEFAULT,
                          "country",
                          bank.ref().country());
      GWEN_DB_SetCharValue(n,
                           GWEN_DB_FLAGS_DEFAULT,
                           "bankcode",
                           bank.ref().bankCode().c_str());
    }

    n=GWEN_DB_GetNextGroup(n);
  } // while n

  return Error();
}


Error API::configHbciVersion(const string& filename, 
			     unsigned& major, unsigned& minor)
{
  GWEN_DB_NODE *cfg;
  Error err;
  bool identified = false;

  cfg=GWEN_DB_Group_new("hbci");
  if (GWEN_DB_ReadFile(cfg, filename.c_str(),
                       GWEN_DB_FLAGS_DEFAULT|
                       GWEN_PATH_FLAGS_CREATE_GROUP)) {
      // File could not be loaded by Gwen
      std::ifstream istr(filename.c_str());
      if (istr) {
	  std::string tmp;
	  istr >> tmp;
	  // FIXME: add better file recognition here
	  if (tmp.find("nextjobid") != string::npos) {
	      major = 0;
	      minor = 9;
	      identified = true;
	  }
      }
      err = Error("API::configHbciVersion",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_FILE_NOT_FOUND,
		  ERROR_ADVISE_DONTKNOW,
		  "The queried config file does not exist.");
  }
  else {
      major = 1;
      minor = 9;
      identified = true;
  }

  if (!identified && err.isOk())
      err = Error("API::configHbciVersion",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_UNSUPPORTED,
		  ERROR_ADVISE_DONTKNOW,
		  "The queried config file could not be recognized.");

  GWEN_DB_Group_free(cfg);
  return err;
}



Error API::loadEnvironmentFromStream(GWEN_BUFFEREDIO* bio,
                                     unsigned int fl){
  Error err;
  GWEN_DB_NODE *cfg;

  cfg=GWEN_DB_Group_new("HBCI");
  if (GWEN_DB_ReadFromStream(cfg,
			     bio,
			     GWEN_DB_FLAGS_COMPACT|
			     GWEN_PATH_FLAGS_CREATE_GROUP)) {
    GWEN_DB_Group_free(cfg);
    return Error("API::loadEnvironmentFromStream()",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "Could not read stream."/*,
		 f*/);
  }
  GWEN_BufferedIO_Flush(bio);
  err=_loader.ref().loadAll(cfg, fl);
  GWEN_DB_Group_free(cfg);
  return err;
}


Error API::saveEnvironmentToStream(GWEN_BUFFEREDIO * bio,
				   unsigned int fl) const {
  GWEN_DB_NODE *cfg;
  Error err;
  int rv;

  cfg=GWEN_DB_Group_new("HBCI");
  err=_loader.ref().saveAll(cfg, fl);
  if (!err.isOk()) {
    GWEN_DB_Group_free(cfg);
    return err;
  }
  rv=GWEN_DB_WriteToStream(cfg, bio, GWEN_DB_FLAGS_COMPACT);
  GWEN_BufferedIO_Flush(bio);
  GWEN_DB_Group_free(cfg);
  if (rv)
    return Error("API::saveEnvironmentToStream()",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "Could not save config to buffer."
		);
  return Error();
}




} // namespace HBCI


