/***************************************************************************
 api_c.cpp
                             -------------------
    begin       : October 2002
    copyright   : (C) 2002, 2003 by Christian Stimming
    email       : stimming@tuhh.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "assert.h"

#include "api.h"
#include "api_c.h"
#include <hbcistring.h>
#include <openhbci2/error.h>


// ------------------ C Wrappers

// Helpers for non-deleting Pointer


static HBCI::Pointer<HBCI_Bank> bankPointer(HBCI_Bank *b){
  HBCI::Pointer<HBCI_Bank> bp = b;
  bp.setAutoDelete(false);
  return bp;
}



static HBCI::Pointer<HBCI_User> userPointer(HBCI_User *u){
  HBCI::Pointer<HBCI_User> up = u;
  up.setAutoDelete(false);
  return up;
}


// class HBCI_API
HBCI_API *HBCI_API_new(int readonly, int retrievalOnly){
  return new HBCI_API(readonly, retrievalOnly);
}



void HBCI_API_delete(HBCI_API *h){
  delete h;
}


HBCI_Hbci *HBCI_API_Hbci(HBCI_API *h){
  return h;
}



const list_HBCI_Bank *HBCI_API_bankList(const HBCI_API *h){
  assert(h);
  return &(h->banks());
}



HBCI_Bank *HBCI_API_findBank(const HBCI_API *h, int country,
			     const char *bankCode){
  assert(h);
  return (h->findBank(country, string(bankCode ? bankCode : ""))).ptr();
}



HBCI_Medium *HBCI_API_findMedium(const HBCI_API *h,
				 const char *name){
  assert(h);
  return (h->findMedium(string(name ? name : ""))).ptr();
}



HBCI_Error *HBCI_API_addBank(HBCI_API *h, HBCI_Bank *b, int autoDelete){
  assert(h);
  assert(b);
  HBCI::Pointer<HBCI_Bank> bp = b;
  bp.setAutoDelete(autoDelete);
  try {
    h->addBank(bp);
    return 0;
  }
  catch(HBCI_Error err) {
    return new HBCI_Error(err);
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_addBank: Caught unknown exception.\n");
    return new HBCI_Error("HBCI_API_addBank",
			  ERROR_LEVEL_CRITICAL,
			  0,
			  ERROR_ADVISE_ABORT,
			  "Caught unknown exception.", "");
  }
}



HBCI_Bank *HBCI_API_bankFactory(HBCI_API *h,
				int country,
				const char *bankCode,
				const char *server,
				const char *port){
  assert(h);
  HBCI::Pointer<HBCI_Bank> bp;

  bp = h->bankFactory(country,
		      string(bankCode ? bankCode : ""),
		      string(server ? server : ""),
		      string(port ? port : ""));
  bp.setAutoDelete(false);
  return bp.ptr();
}


HBCI_MessageEngine *HBCI_API_engineFactory(HBCI_API *h,
					   int country,
					   const char *bankCode,
					   int hversion,
					   const char *engineType)
{
    assert(h);
    HBCI::Pointer<HBCI::MessageEngine> mp;

    mp = h->engineFactory(country, 
			  string(bankCode ? bankCode : ""),
			  hversion,
			  string(engineType ? engineType : ""));
    mp.setAutoDelete(false);
    return mp.ptr();
}
HBCI_Transporter *
HBCI_API_transporterFactory(HBCI_API *h,
			    const char *addr,
			    const char *port,
			    int timeout,
			    const char *transporterType)
{
    assert(h);
    HBCI::Pointer<HBCI::Transporter> mp;

    mp = h->transporterFactory(string(addr ? addr : ""),
			       string(port ? port : ""),
			       timeout,
			       string(transporterType ? transporterType : ""));
    mp.setAutoDelete(false);
    return mp.ptr();
}
HBCI_User *HBCI_API_userFactory(HBCI_Bank *b,
				HBCI_Medium *m,
				int mediumAutoDelete,
				const char *userid,
				const char *username){
  assert(b);
  assert(m);
  HBCI::Pointer<HBCI_Medium> mp = m;
  mp.setAutoDelete(mediumAutoDelete);
  HBCI::Pointer<HBCI_User> up =
    HBCI_API::userFactory(bankPointer(b),
			  mp,
			  string(userid ? userid : ""),
			  string(username ? username : ""));
  up.setAutoDelete(false);
  return up.ptr();
}



HBCI_Customer *HBCI_API_customerFactory(HBCI_User *u,
					const char *id,
					const char *custname) {
  assert(u);
  HBCI::Pointer<HBCI_Customer> cp =
    HBCI_API::customerFactory(userPointer(u), 
			      string(id ? id : ""),
			      string(custname ? custname : ""));
  cp.setAutoDelete(false);
  return cp.ptr();
}



HBCI_Medium *HBCI_API_createNewMedium(HBCI_API *h,
				      const char *mtype,
				      int readonly,
				      int country,
				      const char *bankId,
				      const char *userid,
				      const char *name,
				      HBCI_Error **err)
{
  assert(h);
  assert(err);
  try {
    HBCI::Pointer<HBCI::Medium> mp =
      h->createNewMedium(string(mtype ? mtype : ""),
			 readonly,
			 country,
			 string(bankId ? bankId : ""),
			 string(userid ? userid : ""),
			 string(name ? name : ""));
    mp.setAutoDelete(false);
    return mp.ptr();
  }
  catch(HBCI_Error xerr) {
    fprintf(stderr,"HBCI_API_createNewMedium: Caught error: %s\n",
	    xerr.errorString().c_str());
    *err = new HBCI_Error(xerr);
    return 0;
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_createNewMedium: Caught unknown exception.\n");
    *err = new HBCI_Error("HBCI_API_createNewMedium",
			  ERROR_LEVEL_CRITICAL,
			  0,
			  ERROR_ADVISE_ABORT,
			  "Caught unknown exception.", "");
    return 0;
  }
}



MediumType HBCI_API_mediumType(HBCI_API *h,
			       const char *mtype){
  assert(h);
  try {
    return h->mediumType(mtype ? string(mtype) : "");
  }
  catch (HBCI_Error err) {
    //fprintf(stderr,"HBCI_API_mediumType: Caught error: %s\n",
    //    err.errorString().c_str());
    return MediumTypeUnknown;
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_mediumType: Caught unknown exception.\n");
    return MediumTypeUnknown;
  }
}



HBCI_MediumPluginList *HBCI_API_enumerateMediumPlugins(HBCI_API *h){
  assert(h);
  try {
    HBCI::Pointer<HBCI::MediumPluginList> p =
      h->enumerateMediumPlugins();
    p.setAutoDelete(false);
    return p.ptr();
  }
  catch (HBCI_Error err) {
    fprintf(stderr,"HBCI_API_enumerateMediumPlugins: Caught error: %s\n",
	    err.errorString().c_str());
    return NULL;
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_enumerateMediumPlugins: "
	    "Caught unknown exception.\n");
    return NULL;
  }
}



HBCI_Error *HBCI_API_loadEnvironment(HBCI_API *h,
				     const char *filename){
  assert(h);
  try {
    return new HBCI_Error(h->loadEnvironment(string(filename ? filename : "")));
  }
  catch (HBCI_Error err) {
    fprintf(stderr,"HBCI_API_loadEnvironment: Caught error: %s\n",
	    err.errorString().c_str());
    return new HBCI_Error(err);
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_loadEnvironment: Caught unknown exception.\n");
    return new HBCI_Error("HBCI_API_loadEnvironment",
			  ERROR_LEVEL_CRITICAL,
			  0,
			  ERROR_ADVISE_ABORT,
			  "Caught unknown exception.", "");
  }
}



HBCI_Error *HBCI_API_saveEnvironment(const HBCI_API *h,
				     const char *filename){
  assert(h);
  try {
    return new HBCI_Error(h->saveEnvironment(string(filename ? filename : "")));
  }
  catch (HBCI_Error err) {
    fprintf(stderr,"HBCI_API_saveEnvironment: Caught error: %s\n",
	    err.errorString().c_str());
    return new HBCI_Error(err);
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_saveEnvironment: Caught unknown exception.\n");
    return new HBCI_Error("HBCI_API_saveEnvironment",
			  ERROR_LEVEL_CRITICAL,
			  0,
			  ERROR_ADVISE_ABORT,
			  "Caught unknown exception.", "");
  }
}




HBCI_Error *HBCI_API_executeQueue(HBCI_API *h,
				  HBCI_Outbox *queue)
{
  assert(h);
  assert(queue);
  HBCI::Pointer<HBCI_Outbox> op = queue;
  op.setAutoDelete(false);
  try {
    return new HBCI_Error(h->executeQueue(op));
  }
  catch (HBCI_Error err) {
    fprintf(stderr,"HBCI_API_executeQueue: Caught error: %s\n",
	    err.errorString().c_str());
    return new HBCI_Error(err);
  }
  catch (...) {
    fprintf(stderr, "HBCI_API_executeQueue: Caught unknown exception.\n");
    return new HBCI_Error("HBCI_API_executeQueue",
			  ERROR_LEVEL_CRITICAL,
			  0,
			  ERROR_ADVISE_ABORT,
			  "Caught unknown exception.", "");
  }
}



int HBCI_API_totalUsers(const HBCI_API *h){
  assert(h);
  return h->totalUsers();
}

HBCI_Error *HBCI_API_configHbciVersion(const char *filename, 
				       unsigned *major, unsigned *minor)
{
    assert(major);
    assert(minor);
    HBCI::Error err =
	HBCI::API::configHbciVersion(filename ? filename : "", *major, *minor);
    if (err.isOk())
	return NULL;
    else
	return new HBCI::Error(err);
}



