/***************************************************************************
 $RCSfile: bank.cpp,v $
 -------------------
 cvs         : $Id: bank.cpp,v 1.8 2004/01/29 11:52:13 cstim Exp $
 begin       : Wed Jun 12 2002
 copyright   : (C) 2002 by Martin Preuss
 email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef __declspec
# if BUILDING_DLL
#  define DLLIMPORT __declspec (dllexport)
# else /* Not BUILDING_DLL */
#  define DLLIMPORT __declspec (dllimport)
# endif /* Not BUILDING_DLL */
#else
# define DLLIMPORT
#endif


#include "bank.h"

namespace HBCI {


Bank::Bank(Hbci *hbci,
	   int country,
	   const string &code,
	   const string &addr,
	   const string &port,
           int hbciversion,
           const string &name)
    :_hbci(hbci)
    ,_country(country)
    ,_bankCode(code)
    ,_addr(addr)
    ,_port(port)
    ,_hbciVersion(hbciversion)
    ,_name(name)
{
  _msgEngine.setDescription("Bank::_msgEngine");
  _transporter.setDescription("Bank::_transporter");
}


Bank::~Bank(){
  // Manually delete the Pointer<User> list since we own these
  // objects but due to cross-referencing User<->Bank the reference
  // counters won't reach zero automatically.
  list<Pointer<User> >::iterator uit;
  for (uit=_users.begin();
       uit!=_users.end();
       uit++) {
    (*uit).release();
  }
  _msgEngine.release();
  _transporter.release();
}


Pointer<User> 
Bank::findUser(const string &userid) const {
    list<Pointer<User> >::const_iterator it;

    for (it=_users.begin();
         it!=_users.end();
         it++) {
        if ((*it).ref().userId()==userid)
            return (*it);
    }
    return 0;
}


Pointer<Customer>
Bank::findCustomer(const string &id) const {
    list<Pointer<User> >::const_iterator it;
    Pointer<Customer> cu;

    for (it=_users.begin();
         it!=_users.end();
         it++) {
        cu=(*it).ref().findCustomer(id);
        if (cu.isValid())
            return cu;
    }
    return 0;
}


const list<Pointer<User> > &Bank::users() const {
  return _users;
}


void Bank::addUser(Pointer<User> c){
  list<Pointer<User> >::const_iterator it;
  Pointer<User> nc;

  for (it=_users.begin();
       it!=_users.end();
       it++) {
    /* first check if the very same object already exists.
     * It's not a problem if it does, since it is the SAME.
     */
    if ((*it)==c) {
#if DEBUGMODE>0
      fprintf(stderr,
	      "Will not add customer, since he already IS listed.\n");
#endif
      return;
    }

    /* then check if there already is an object with the same userid.
     * This WOULD be a problem, since that would corrupt the list.
     * The second occurence of an equal customer could rarely be accessed.
     */
    if ((*it).ref().userId()==c.ref().userId())
      throw Error("Bank::addUser()",
		  ERROR_LEVEL_NORMAL,
		  0,
		  ERROR_ADVISE_DONTKNOW,
		  "user already exists.");
  }

  // otherwise add customer
  nc=c;
  nc.setDescription("Entry of Bank::_users");
  _users.push_back(nc);
}


void Bank::removeUser(Pointer<User> c){
  list<Pointer<User> >::iterator it1;

  for (it1=_users.begin();
       it1!=_users.end();
       it1++) {
    if (*it1==c) {
      _users.erase(it1);
      break;
    }
  } // for
}



void Bank::setHbciVersion(int v) {
  _hbciVersion=v;
  if (_msgEngine.isValid()) {
    _msgEngine.ref().setHbciVersion(v);
  }
}



void Bank::addHbciVersion(int v){
  list<int>::iterator it;

  for (it=_versions.begin(); it!=_versions.end(); it++) {
    if ((*it)==v)
      break;
  } // for
  _versions.push_back(v);
}



void Bank::clearHbciVersions(){
  _versions.clear();
}



} /* namespace HBCI */


const HBCI_Hbci *HBCI_Bank_hbci(const HBCI_Bank *h)
{
    assert(h);
    return h->hbci();
}
int HBCI_Bank_country(const HBCI_Bank *h)
{
    assert(h);
    return h->country();
}

const char *HBCI_Bank_bankCode(const HBCI_Bank *h)
{ 
    assert(h); 
    return h->bankCode().c_str();
}
const char *HBCI_Bank_name(const HBCI_Bank *h)
{
    assert(h); 
    return h->name().c_str();
}
const char *HBCI_Bank_addr(const HBCI_Bank *h)
{
    return HBCI_Bank_address(h);
}
const char *HBCI_Bank_address(const HBCI_Bank *h)
{
    assert(h);
    return h->address().c_str();
}

const HBCI_User *HBCI_Bank_findUser(const HBCI_Bank *h, const char *userid)
{
    assert(h); 
    return (h->findUser(string(userid ? userid : ""))).ptr();
}
const list_HBCI_User *HBCI_Bank_users(const HBCI_Bank *h)
{
    assert(h);
    return &(h->users());
}
HBCI_Error *HBCI_Bank_addUser(HBCI_Bank *h, HBCI_User *c, int autoDelete)
{
  assert(h);
  assert(c);
  HBCI::Pointer<HBCI::User> cp = c;
  cp.setAutoDelete(autoDelete);
  try {
    h->addUser(cp);
    return 0;
  }
  catch(HBCI_Error err) {
    return new HBCI_Error(err);
  }
}
void HBCI_Bank_setMessageEngine(HBCI_Bank *h,
				HBCI_MessageEngine *t,
				int autoDelete)
{
  assert(h);
  assert(t);
  HBCI::Pointer<HBCI::MessageEngine> cp = t;
  cp.setAutoDelete(autoDelete);
  try {
    h->setMessageEngine(cp);
    //return 0;
  }
  catch(HBCI_Error err) {
      std::cerr << "HBCI_Bank_setMessageEngine: caught error: " 
		<< err.errorString() << std::endl;
      //return new HBCI_Error(err);
  }
}

void HBCI_Bank_setTransporter(HBCI_Bank *h,
			      HBCI_Transporter *t,
			      int autoDelete)
{
  assert(h);
  assert(t);
  HBCI::Pointer<HBCI::Transporter> cp = t;
  cp.setAutoDelete(autoDelete);
  try {
    h->setTransporter(cp);
    //return 0;
  }
  catch(HBCI_Error err) {
      std::cerr << "HBCI_Bank_setTransporter: caught error: " 
		<< err.errorString() << std::endl;
      //return new HBCI_Error(err);
  }
}

int HBCI_Bank_hbciVersion(const HBCI_Bank *h)
{ 
    assert(h); 
    return h->hbciVersion();
}
void HBCI_Bank_setHbciVersion(HBCI_Bank *h, int v)
{ 
    assert(h); 
    h->setHbciVersion(v);
}
list_int *HBCI_Bank_supportedVersions(const HBCI_Bank *h)
{
    return HBCI_Bank_hbciVersions(h);
}
list_int *HBCI_Bank_hbciVersions(const HBCI_Bank *h)
{ 
    assert(h); 
    return new list_int(h->hbciVersions());
}



void list_HBCI_Bank_delete(list_HBCI_Bank *l)
{
    delete l;
}
void list_HBCI_Bank_iter_delete(list_HBCI_Bank_iter *l)
{
    delete l;
}
int list_HBCI_Bank_size(const list_HBCI_Bank *l)
{
    assert(l);
    return l->size();
}
list_HBCI_Bank_iter* list_HBCI_Bank_begin(const list_HBCI_Bank *l)
{
    assert(l);
    return new list_HBCI_Bank_iter(l->begin());
}
list_HBCI_Bank_iter* list_HBCI_Bank_end(const list_HBCI_Bank *l)
{
    assert(l);
    return new list_HBCI_Bank_iter(l->end());
}
const HBCI_Bank *
list_HBCI_Bank_iter_get(const list_HBCI_Bank_iter *l)
{
    assert(l);
    return (**l).ptr();
}
void list_HBCI_Bank_iter_next(list_HBCI_Bank_iter *l)
{
    assert(l);
    l->operator++();
}
int list_HBCI_Bank_iter_equal(const list_HBCI_Bank_iter *l1, 
			      const list_HBCI_Bank_iter *l2)
{
    assert(l1);
    assert(l2);
    return (*l1) == (*l2);
}
void *list_HBCI_Bank_foreach (const list_HBCI_Bank *list, 
			      list_HBCI_Bank_cb func,
			      void *user_data)
{
    assert(list);
    assert(func);
    list_HBCI_Bank_iter iter, end;
    void *ret = 0l;

    end = list->end();
    for ( iter = list->begin(); iter != end; iter++ ) {
	ret = (*func)( (*iter).ptr(), user_data);
	if (ret)
	    break;
    }
    return ret;
}



