/***************************************************************************
 $RCSfile: bank.h,v $
 -------------------
 cvs         : $Id: bank.h,v 1.9 2004/02/11 12:46:46 cstim Exp $
 begin       : Wed Jun 12 2002
 copyright   : (C) 2002 by Martin Preuss
 email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HBCIBANKIMPL_H
#define HBCIBANKIMPL_H

#ifdef __cplusplus
namespace HBCI {
  class Bank;

}

#include <openhbci2/pointer.h>
#include <list>
#include <string>
using namespace std;


/** @short This is the C accessible type of a Bank in %HBCI.
 * @ingroup HBCI_Bankg
 */
typedef struct HBCI::Bank HBCI_Bank;
/** @short List of HBCI_Bank's
 * @ingroup HBCI_Bankg */
typedef list<HBCI::Pointer<HBCI::Bank> > list_HBCI_Bank;
/** @short Iterator for List of HBCI_Bank's
 * @ingroup HBCI_Bankg */
typedef list<HBCI::Pointer<HBCI::Bank> >::const_iterator list_HBCI_Bank_iter;
#else /* __cplusplus */
typedef struct HBCI_Bank HBCI_Bank;
typedef struct list_HBCI_Bank list_HBCI_Bank;
typedef struct list_HBCI_Bank_iter list_HBCI_Bank_iter;

#endif /* __cplusplus */

#include <openhbci2/pointer.h>
#include <openhbci2/hbci.h>
#include <openhbci2/user.h>
#include <openhbci2/msgengine.h>
#include <openhbci2/transporter.h>
#include <openhbci2/listwrappers.h>


#ifdef __cplusplus
namespace HBCI {
/**
 * @short Implementation of the abstract base class Bank.
 *
 * This class represents a full institute. It contains a list of
 * its customers as well as a list of its accounts. There is also a list
 * of institute messages which this institute has send to us.
 *
 * This class should only be visible inside OpenHBCI -- all external
 * applications are only supposed to use the interface class
 * Bank.
 *
 * @author Martin Preuss<martin@libchipcard.de>
 */
class DLLIMPORT Bank {
private:
  Hbci *_hbci;
  int _country;
  string _bankCode;
  string _addr;
  string _port;
  int _hbciVersion;
  list<Pointer<User> > _users;
  Pointer<MessageEngine> _msgEngine;
  Pointer<Transporter> _transporter;
  string _name;
  list<int> _versions;

public:
  Bank(Hbci *hbci,
       int country = 280,
       const string &code = "",
       const string &addr = "",
       const string &port = "",
       int hbciversion = HBCI_VERSION_201,
       const string &name="");

  virtual ~Bank();

  /** Returns the central HBCI object that controls everything. */
  Hbci *hbci() const { return _hbci;};

  int country() const { return _country; };

  const string &bankCode() const { return _bankCode;};

  const string &name() const { return _name;};
  void setName(const string &name) { _name=name;};

  const string &address() const { return _addr;};
  void setAddress(const string &s) { _addr=s;};

  const string &port() const { return _port;};
  void setPort(const string &s) { _port=s;};

  /**
   * Returns a list of supported HBCI versions.
   */
  const list<int> &hbciVersions() const { return _versions;};
  void clearHbciVersions();
  void addHbciVersion(int v);

  /**
   * Get the HBCI version.
   * This is the version of the HBCI protocol to be used with this bank.
   */
  int hbciVersion() const { return _hbciVersion;};

  /**
   * Set the HBCI protocol version to be used with this bank.
   */
  void setHbciVersion(int v);

  /**
   * @author Martin Preuss <martin@libchipcard.de>
   */
  Pointer<User> findUser(const string &userid) const;

  /**
   * Searches for a customer in the banks internal list.
   * No wildcards or jokers allowed.
   * @param id userid to look for
   * @author Martin Preuss <martin@libchipcard.de>
   */
  virtual Pointer<Customer>
    findCustomer(const string &id) const;

  /**
   * @author Martin Preuss <martin@libchipcard.de>
   */
  const list<Pointer<User> > &users() const;

  /**
   * @author Martin Preuss <martin@libchipcard.de>
   */
  void addUser(Pointer<User> c);

  /**
   * Removes a customer from the internal list. The object is
   * not destroyed, but you should not use it anymore.
   * @author Martin Preuss <martin@libchipcard.de>
   */
  void removeUser(Pointer<User> c);

  /** Set the MessageEngine to be used. */
  void setMessageEngine(Pointer<MessageEngine> me) { _msgEngine=me;};
  Pointer<MessageEngine> messageEngine() const { return _msgEngine;};

  /** Set the Transporter to be used. */
  void setTransporter(Pointer<Transporter> t) { _transporter=t;};
  Pointer<Transporter> transporter() const { return _transporter;};


};
} /* namespace HBCI */

#endif /* __cplusplus */

#ifdef __cplusplus
extern "C" {
#endif

  /** @defgroup HBCI_Bankg HBCI_Bank functions
   *
   * @short These are the functions for the C accessible type of
   * HBCI::Bank.*/
  /*@{*/

  /** @name Bank identification */
  /*@{*/
  /**
   *  Returns the ISO-3166 countrycode of this bank.
   *
   * In %HBCI, "280" is still used for Germany according to the %HBCI
   * specs -- even though in the year 1990 the code was changed from
   * 280 to 276 in ISO-3166.
   * @param h The object to perform this action on.
   */
  extern int HBCI_Bank_country(const HBCI_Bank *h);

  /**  Returns the bank code of this bank.
   *
   * ("Bankleitzahl", BLZ)
   * @param h The object to perform this action on.
   */
  extern const char *HBCI_Bank_bankCode(const HBCI_Bank *h);

  /**  Returns the name of this Bank.
   *
   * This name is the one that the bank sends back upon initally
   * connecting.
   * @param h The object to perform this action on.
   */
  extern const char *HBCI_Bank_name(const HBCI_Bank *h);

  extern const char *HBCI_Bank_addr(const HBCI_Bank *h);
  extern const char *HBCI_Bank_address(const HBCI_Bank *h);
  /*@}*/

  /** @name Object retrieval */
  /*@{*/
  /**
   *  Returns the list of users of this bank.
   * @param h The object to perform this action on.
   */
  extern const list_HBCI_User *HBCI_Bank_users(const HBCI_Bank *h);
  /**
   *  Searches for a user in the bank's internal list.
   *
   * No wildcards or jokers allowed.
   * @param h The object to perform this action on.
   * @param userid userid to look for
   */
  extern const HBCI_User *HBCI_Bank_findUser(const HBCI_Bank *h,
					     const char *userid);
  /**  Returns the central %HBCI object that controls the user
   * interaction and version control.
   * @param h The object to perform this action on.
   */
  extern const HBCI_Hbci *HBCI_Bank_hbci(const HBCI_Bank *h);
  /*@}*/

  /** @name Manage Objects*/
  /*@{*/
  /**
   *  Adds a customer to the list of customers of this bank.
   *
   * If there already is an entry for this customer a Error will
   * be returned.
   *
   * @param h The object to perform this action on.
   * @param c User to add.
   * @param autoDelete If TRUE, then the HBCI_Bank takes ownership of
   * the HBCI_User object and will delete it if the life time is
   * over. If FALSE, the HBCI_Bank will <i>not</i> delete the
   * HBCI_User object; instead, the caller is still responsible for
   * its deletion.
   * @return NULL if successful, or pointer to HBCI_Error upon failure.
   */
  extern HBCI_Error *HBCI_Bank_addUser(HBCI_Bank *h,
				       HBCI_User *c,
				       int autoDelete);
  /** Set the MessageEngine to be used. 
   * @param h The object to perform this action on.
   * @param t MessageEngine to be set.
   * @param autoDelete If TRUE, then the HBCI_Bank takes ownership of
   * the HBCI_MessageEngine object and will delete it if the life time is
   * over. If FALSE, the HBCI_Bank will <i>not</i> delete the
   * HBCI_MessageEngine object; instead, the caller is still responsible for
   * its deletion.
   */
  extern void HBCI_Bank_setMessageEngine(HBCI_Bank *h,
					 HBCI_MessageEngine *t,
					 int autoDelete);

  /** Set the Transporter to be used. 
   * @param h The object to perform this action on.
   * @param t Transporter to be set.
   * @param autoDelete If TRUE, then the HBCI_Bank takes ownership of
   * the HBCI_Transporter object and will delete it if the life time is
   * over. If FALSE, the HBCI_Bank will <i>not</i> delete the
   * HBCI_Transporter object; instead, the caller is still responsible for
   * its deletion.
   */
  extern void HBCI_Bank_setTransporter(HBCI_Bank *h,
				       HBCI_Transporter *t,
				       int autoDelete);
  /*@}*/

  /** @name HBCI Version */
  /*@{*/
  /**
   *  Returns the %HBCI protocol version used with this bank.
   *
   * This %HBCI version number is one out of the list<int> returned
   * by supportedVersions(). This has been set by setHbciVersion,
   * or, if it hasn't been set, it is some default value.  */
  extern int HBCI_Bank_hbciVersion(const HBCI_Bank *h);
  /**
   *  Set the %HBCI protocol version to be used with this bank.
   */
  extern void HBCI_Bank_setHbciVersion(HBCI_Bank *h, int v);
  /**
   *  Returns a list of %HBCI protocol versions that this bank supports.
   */
  extern list_int *HBCI_Bank_hbciVersions(const HBCI_Bank *h);
  extern list_int *HBCI_Bank_supportedVersions(const HBCI_Bank *h);
  /*@}*/

  /** @name C wrapper for list of Banks
   *
   * C wrapper for STL's list<Bank> type. To browse through a list,
   * get an iterator by list_HBCI_Bank_begin, increment it by
   * list_HBCI_Bank_iter_next, access its elements by
   * list_HBCI_Bank_iter_get, and see whether you're at the end of
   * the list by comparing with list_HBCI_Bank_iter_equal against
   * list_HBCI_Bank_end. */
  /*@{*/
  extern void list_HBCI_Bank_delete(list_HBCI_Bank *l);
  extern void list_HBCI_Bank_iter_delete(list_HBCI_Bank_iter *l);
  extern int list_HBCI_Bank_size(const list_HBCI_Bank *l);
  extern list_HBCI_Bank_iter* list_HBCI_Bank_begin(const list_HBCI_Bank *l);
  extern list_HBCI_Bank_iter* list_HBCI_Bank_end(const list_HBCI_Bank *l);
  extern const HBCI_Bank*
    list_HBCI_Bank_iter_get(const list_HBCI_Bank_iter *l);
  extern void list_HBCI_Bank_iter_next(list_HBCI_Bank_iter *l);
  extern int list_HBCI_Bank_iter_equal(const list_HBCI_Bank_iter *l1,
				       const list_HBCI_Bank_iter *l2);
  /** Callback function for @ref list_HBCI_Bank_foreach. */
  typedef void *(*list_HBCI_Bank_cb) (const HBCI_Bank *bank,
				      void *user_data);
  /** Traverses the list, calling the callback function 'func' on
   * each list element.  Traversal will stop when 'func' returns a
   * non-NULL value, and the routine will return with that
   * value. Otherwise the routine will return NULL.
   *
   * @param list The list to traverse.
   * @param func The function to be called with each list element.
   * @param user_data A pointer passed on to the function 'func'.
   * @return The non-NULL pointer returned by 'func' as soon as it
   * returns one. Otherwise (i.e. 'func' always returns NULL)
   * returns NULL.
   * @author Christian Stimming <stimming@tuhh.de> */
  extern void *list_HBCI_Bank_foreach (const list_HBCI_Bank *list,
				       list_HBCI_Bank_cb func,
				       void *user_data);
  /*@}*/
  /*@}*/
#ifdef __cplusplus
}
#endif


#endif /* HBCIBANKIMPL_H */



