/***************************************************************************
                          hbcirsakey.cpp  -  description
                             -------------------
    begin                : Sun Jul 1 2001
    copyright            : (C) 2001 by fabian kaiser
    email                : fabian.kaiser@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

/*
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <assert.h>

#ifdef __declspec
# if BUILDING_DLL
#  define DLLIMPORT __declspec (dllexport)
# else /* Not BUILDING_DLL */
#  define DLLIMPORT __declspec (dllimport)
# endif /* Not BUILDING_DLL */
#else
# define DLLIMPORT
#endif


#include <string>
#include <stdio.h>
#include <list>

#include "rsakey.h"
#include "hbcistring.h"

#include <openhbci2/error.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/crypt.h>
#include <gwenhywfar/md.h>
#include <gwenhywfar/padd.h>

namespace HBCI {



bool RSAKey::toDB(GWEN_DB_NODE *n) const{
  GWEN_ERRORCODE err;

  if (!_key)
    return false;
  err=GWEN_CryptKey_ToDb(_key, n, GWEN_CryptKey_IsPublic(_key));
  if (!GWEN_Error_IsOk(err))
    return false;
  return true;
}



RSAKey::RSAKey(bool cryptKey, GWEN_DB_NODE *n): _key(0){
  if (n) {
    _key=GWEN_CryptKey_FromDb(n);
    if (!_key)
      throw Error("RSAKey::RSAKey",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_INVALID,
		  ERROR_ADVISE_DONTKNOW,
		  "Could not create key from db");
  }
  else {
    _key=GWEN_CryptKey_Factory("RSA");
    throw Error("RSAKey::RSAKey",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_INVALID,
		ERROR_ADVISE_DONTKNOW,
		"Could not create RSA key");
  }
  setToSignKey(!cryptKey);
  _userId=GWEN_CryptKey_GetOwner(_key);
}



RSAKey::RSAKey(const RSAKey &k):_key(0){
  if (k._key)
    _key=GWEN_CryptKey_dup(k._key);
  _isCryptKey=k._isCryptKey;
  _userId=k._userId;
}


RSAKey::~RSAKey(){
  GWEN_CryptKey_free(_key);
}


bool RSAKey::decrypt() {
  GWEN_ERRORCODE err;
  GWEN_BUFFER *srcbuf, *dstbuf;

  srcbuf=GWEN_Buffer_new((char*)intData.data(),
			 intData.length(), intData.length(), 0);
  GWEN_Buffer_SetMode(srcbuf, 0);
  dstbuf=GWEN_Buffer_new(0,
			 intData.length(), 0, 1);
  err=GWEN_CryptKey_Decrypt(_key,
			    srcbuf,
			    dstbuf);
  if (!GWEN_Error_IsOk(err)) {
    GWEN_Buffer_free(srcbuf);
    GWEN_Buffer_free(dstbuf);
    return false;
  }

  intData=string((const char*)GWEN_Buffer_GetStart(dstbuf),
		 GWEN_Buffer_GetUsedBytes(dstbuf));
  GWEN_Buffer_free(srcbuf);
  GWEN_Buffer_free(dstbuf);
  return true;
}


bool RSAKey::encrypt() {
  GWEN_ERRORCODE err;
  GWEN_BUFFER *srcbuf, *dstbuf;
  string s;
  unsigned int keylen;

  keylen=GWEN_CryptKey_GetChunkSize(_key);
  s=string(keylen - intData.length(), '\0') + intData;
  srcbuf=GWEN_Buffer_new(0, s.length(), 0, 1);
  GWEN_Buffer_AppendBytes(srcbuf, s.data(), s.length());

  dstbuf=GWEN_Buffer_new(0,
			 intData.length(), 0, 1);
  err=GWEN_CryptKey_Encrypt(_key,
			    srcbuf,
			    dstbuf);
  if (!GWEN_Error_IsOk(err)) {
    DBG_ERROR_ERR(0, err);
    GWEN_Buffer_free(srcbuf);
    GWEN_Buffer_free(dstbuf);
    return false;
  }
  intData=string((const char*)GWEN_Buffer_GetStart(dstbuf),
		 GWEN_Buffer_GetUsedBytes(dstbuf));
  GWEN_Buffer_free(srcbuf);
  GWEN_Buffer_free(dstbuf);
  return true;
}



bool RSAKey::verify(const string &signature) {
  GWEN_MD *md;
  GWEN_BUFFER *hashbuf;
  GWEN_BUFFER *signbuf;
  GWEN_ERRORCODE err;

  /* hash data */
  md=GWEN_MD_Factory("RMD160");
  if (!md) {
    DBG_ERROR(0, "RMD160 not found");
    return false;
  }

  if (GWEN_MD_Begin(md)) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return false;
  }

  if (GWEN_MD_Update(md,
		     (const char*)intData.data(),
		     intData.length())) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return false;
  }

  if (GWEN_MD_End(md)) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return false;
  }

  hashbuf=GWEN_Buffer_new(0, GWEN_MD_GetDigestSize(md), 0, 1);
  if (GWEN_Buffer_AppendBytes(hashbuf,
                              (const char*)GWEN_MD_GetDigestPtr(md),
                              GWEN_MD_GetDigestSize(md))) {
    DBG_INFO(0, "here");
    GWEN_Buffer_free(hashbuf);
    GWEN_MD_free(md);
    return false;
  }
  GWEN_MD_free(md);

  /* padd */
  DBG_DEBUG(0, "Padding hash using ISO 9796");
  if (GWEN_Padd_PaddWithISO9796(hashbuf)) {
    DBG_INFO(0, "here");
    GWEN_Buffer_free(hashbuf);
    return false;
  }

  /* verify hash */
  signbuf=GWEN_Buffer_new((char*)signature.data(),
			  signature.length(),
			  signature.length(),
			  0);
  GWEN_Buffer_SetMode(signbuf, 0);
  err=GWEN_CryptKey_Verify(_key,
			   hashbuf,
			   signbuf);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(0, err);
    GWEN_Buffer_free(hashbuf);
    GWEN_Buffer_free(signbuf);
    return false;
  }
  GWEN_Buffer_free(hashbuf);
  GWEN_Buffer_free(signbuf);
  return true;
}



bool RSAKey::sign() {
  GWEN_MD *md;
  GWEN_BUFFER *hashbuf;
  GWEN_BUFFER *signbuf;
  GWEN_ERRORCODE err;

  DBG_INFO(0, "RSAKey: Signing data");
  /* hash data */
  DBG_DEBUG(0, "Hash data");
  md=GWEN_MD_Factory("RMD160");
  if (!md) {
    DBG_ERROR(0, "RMD160 not found");
    return false;
  }

  if (GWEN_MD_Begin(md)) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return false;
  }

  if (GWEN_MD_Update(md,
		     (const char*)intData.data(),
		     intData.length())) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return false;
  }

  if (GWEN_MD_End(md)) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return false;
  }
  DBG_DEBUG(0, "Hashing done");

  hashbuf=GWEN_Buffer_new(0,
			  GWEN_MD_GetDigestSize(md)+96,
			  0, 1);
  if (GWEN_Buffer_AppendBytes(hashbuf,
			      (const char*)GWEN_MD_GetDigestPtr(md),
			      GWEN_MD_GetDigestSize(md))) {
    DBG_INFO(0, "here");
    GWEN_Buffer_free(hashbuf);
    GWEN_MD_free(md);
    return false;
  }
  GWEN_MD_free(md);

  /* padd */
  DBG_DEBUG(0, "Padding hash using ISO 9796");
  if (GWEN_Padd_PaddWithISO9796(hashbuf)) {
    DBG_INFO(0, "here");
    GWEN_Buffer_free(hashbuf);
    return false;
  }

  /* sign hash */
  signbuf=GWEN_Buffer_new(0,
			  96,
			  0,
			  1);
  GWEN_Buffer_Rewind(hashbuf);
  err=GWEN_CryptKey_Sign(_key,
                         hashbuf,
			 signbuf);

  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(0, err);
    GWEN_Buffer_free(hashbuf);
    GWEN_Buffer_free(signbuf);
    return false;
  }

  intData=string(GWEN_Buffer_GetStart(signbuf),
		 GWEN_Buffer_GetUsedBytes(signbuf));

  GWEN_Buffer_free(hashbuf);
  GWEN_Buffer_free(signbuf);
  DBG_DEBUG(0, "Signing done");
  return true;
}



string RSAKey::getModulusData() const {
  GWEN_DB_NODE *n;
  const void *p;
  unsigned int l;
  string result;

  n=GWEN_DB_Group_new("keydata");
  if (!toDB(n)) {
    GWEN_DB_Group_free(n);
    return "";
  }

  p=GWEN_DB_GetBinValue(n,
			"data/n",
			0,
			0,0,
			&l);
  if (!p) {
    GWEN_DB_Group_free(n);
    return "";
  }
  result=string((const char*)p, l);
  GWEN_DB_Group_free(n);
  return result;
}



string RSAKey::getExpData() const {
  GWEN_DB_NODE *n;
  const void *p;
  unsigned int l;
  string result;

  n=GWEN_DB_Group_new("keydata");
  if (!toDB(n)) {
    GWEN_DB_Group_free(n);
    return "";
  }

  p=GWEN_DB_GetBinValue(n,
			"data/e",
			0,
			0,0,
			&l);
  if (!p) {
    GWEN_DB_Group_free(n);
    return "";
  }
  result=string((const char*)p, l);
  GWEN_DB_Group_free(n);
  return result;
}



void RSAKey::generateKeyPair(unsigned int keyLength,
                             RSAKey **privKey, RSAKey **pubKey) {
  GWEN_CRYPTKEY *key;
  GWEN_ERRORCODE err;
  GWEN_DB_NODE *db;

  key=GWEN_CryptKey_Factory("RSA");
  err=GWEN_CryptKey_Generate(key, keyLength);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(0, err);
    GWEN_CryptKey_free(key);
    throw Error("RSAKey::generateKeyPair",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_INVALID,
                ERROR_ADVISE_DONTKNOW,
                "Could not generate key");
  }
  // some RaiBas don't like 0
  GWEN_CryptKey_SetNumber(key, 1);
  GWEN_CryptKey_SetVersion(key, 1);

  db=GWEN_DB_Group_new("pub");
  GWEN_CryptKey_SetPublic(key, 1);
  err=GWEN_CryptKey_ToDb(key, db, 1);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(0, err);
    GWEN_DB_Group_free(db);
    GWEN_CryptKey_free(key);
    throw Error("RSAKey::generateKeyPair",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_INVALID,
                ERROR_ADVISE_DONTKNOW,
                "Error in ToDB");
  }
  *pubKey=new RSAKey(false, db);
  GWEN_DB_Group_free(db);

  db=GWEN_DB_Group_new("private");
  GWEN_CryptKey_SetPublic(key, 0);
  err=GWEN_CryptKey_ToDb(key, db, 0);
  if (!GWEN_Error_IsOk(err)) {
    DBG_INFO_ERR(0, err);
    GWEN_DB_Group_free(db);
    GWEN_CryptKey_free(key);
    throw Error("RSAKey::generateKeyPair",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_INVALID,
                ERROR_ADVISE_DONTKNOW,
                "Error in ToDB");
  }
  *privKey=new RSAKey(false, db);
  GWEN_DB_Group_free(db);
  GWEN_CryptKey_free(key);
}



string RSAKey::getIniLetterModulus() const {
  char buffer[256];
  string modulus;

  modulus=getModulusData();
  if (modulus.length()<96)
    modulus=string(96 - modulus.length(), 0x0) + modulus;
  if (GWEN_Text_ToHex((const char*)modulus.data(), modulus.length(),
		      buffer, sizeof(buffer))==0) {
    return "";
  }
  else
    return buffer;
}



string RSAKey::getIniLetterExponent() const {
  char buffer[256];
  string expo;

  expo=getExpData();
  if (expo.length()<96)
    expo = string(96 - expo.length(), 0x0) + expo;
  if (GWEN_Text_ToHex((const char*)expo.data(), expo.length(),
                      buffer, sizeof(buffer))==0) {
    return "";
  }
  else
    return buffer;
}



string RSAKey::getIniLetterHash() const {
  string expo;
  string modulus;
  string result;
  char buffer[64];

  expo=getExpData();
  modulus=getModulusData();
  result = string(128 - expo.length(), 0x0) + expo;
  result += string(128 - modulus.length(), 0x0) + modulus;
  result = _ripe(result);
  if (GWEN_Text_ToHex((const char*)result.data(), result.length(),
		      buffer, sizeof(buffer))==0) {
    return "";
  }
  else
    return buffer;
}



string RSAKey::_ripe(const string &src) const {
  GWEN_MD *md;
  string result;

  /* hash data */
  DBG_DEBUG(0, "Hash data");
  md=GWEN_MD_Factory("RMD160");
  if (!md) {
    DBG_ERROR(0, "RMD160 not found");
    return "";
  }

  if (GWEN_MD_Begin(md)) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return "";
  }

  if (GWEN_MD_Update(md,
		     (const char*)src.data(),
		     src.length())) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return "";
  }

  if (GWEN_MD_End(md)) {
    DBG_INFO(0, "here");
    GWEN_MD_free(md);
    return "";
  }
  DBG_DEBUG(0, "Hashing done");

  result=string((const char*)GWEN_MD_GetDigestPtr(md),
		GWEN_MD_GetDigestSize(md));
  GWEN_MD_free(md);
  return result;
}


int RSAKey::number() const{
  return GWEN_CryptKey_GetNumber(_key);
}



void RSAKey::setNumber(int newNumber){
  GWEN_CryptKey_SetNumber(_key, newNumber);
}



int RSAKey::version() const{
  return GWEN_CryptKey_GetVersion(_key);
}



void RSAKey::setVersion(int newVersion){
  GWEN_CryptKey_SetNumber(_key, newVersion);
}



const string &RSAKey::userId() const{
  return _userId;
}



void RSAKey::setUserId(const string &id){
  _userId=id;
  GWEN_CryptKey_SetOwner(_key, id.c_str());
}


void RSAKey::setToSignKey(bool signKey){
  if (signKey)
    GWEN_CryptKey_SetKeyName(_key, "S");
  else
    GWEN_CryptKey_SetKeyName(_key, "V");
}



bool RSAKey::isCryptoKey() const {
  return (strcasecmp(GWEN_CryptKey_GetKeyName(_key), "V")==0);
}



bool RSAKey::isPublicKey() const {
  return GWEN_CryptKey_IsPublic(_key);
}


} /* namespace HBCI */

HBCI_RSAKey *HBCI_RSAKey_new(int cryptKey, GWEN_DB_NODE *n)
{
    return new HBCI::RSAKey(cryptKey, n);
}

int HBCI_RSAKey_isPublicKey(const HBCI_RSAKey *h)
{
    assert(h);
    return h->isPublicKey();
}

int HBCI_RSAKey_isCryptoKey(const HBCI_RSAKey *h)
{
    assert(h);
    return h->isCryptoKey();
}
int HBCI_RSAKey_number(const HBCI_RSAKey *h)
{
    assert(h);
    return h->number();
}
int HBCI_RSAKey_version(const HBCI_RSAKey *h)
{
    assert(h);
    return h->version();
}
void HBCI_RSAKey_setNumber(HBCI_RSAKey *h, int newNumber)
{
    assert(h);
    h->setNumber(newNumber);
}
void HBCI_RSAKey_setVersion(HBCI_RSAKey *h, int newVersion)
{
    assert(h);
    h->setVersion(newVersion);
}
const char *HBCI_RSAKey_userId(const HBCI_RSAKey *h)
{
    assert(h);
    return h->userId().c_str();
}
char * HBCI_RSAKey_getIniLetterModulus(const HBCI_RSAKey *h)
{
    assert(h);
    return hbci_strdup(h->getIniLetterModulus());
}
char * HBCI_RSAKey_getIniLetterExponent(const HBCI_RSAKey *h)
{
    assert(h);
    return hbci_strdup(h->getIniLetterExponent());
}
char * HBCI_RSAKey_getIniLetterHash(const HBCI_RSAKey *h)
{
    assert(h);
    return hbci_strdup(h->getIniLetterHash());
}
unsigned HBCI_RSAKey_getModulusData(const HBCI_RSAKey *h, 
				    void *buffer, unsigned maxbuf)
{
    assert(h);
    assert(buffer);

    std::string r = h->getModulusData();
    assert(r.size() < maxbuf);

    memcpy(buffer, r.c_str(), r.size());
    return r.size();
}
unsigned HBCI_RSAKey_getExpData(const HBCI_RSAKey *h, 
				void *buffer, unsigned maxbuf)
{
    assert(h);
    assert(buffer);

    std::string r = h->getExpData();
    assert(r.size() < maxbuf);

    memcpy(buffer, r.c_str(), r.size());
    return r.size();
}

