/***************************************************************************
                          hbcirsakey.h  -  description
                             -------------------
    begin                : Sun Jul 1 2001
    copyright            : (C) 2001 by fabian kaiser
    email                : fabian.kaiser@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

/*
 */


#ifndef HBCIRSAKEY_H
#define HBCIRSAKEY_H

#include <openhbci2/dllimport.h>
#include <openhbci2/cryptkey.h>
#include <gwenhywfar/crypt.h>

#ifdef __cplusplus 
#include <string>

#define DEFAULT_KEY_LENGTH 768
#define DEFAULT_EXPONENT 65537
using namespace std;

namespace HBCI {

/**
 * @short Performs encryption and decryption with rsa.
 *
 * Also, padding method ISO-9796 is available.
 * @author fabian kaiser
 */
class DLLIMPORT RSAKey : public CryptKey  {
public:
  /**
   * Creates an RSA key from the given DB.
   * The DB should have the following entries:
   * <ul>
   * <li>For public keys:
   * <pre>
   *   char type="RSA"
   *   char owner="USERID"
   *   int  number="1"
   *   int  version="1"
   *   data {
   *     int  public="1"
   *     bin  n="MODULUS"
   *     bin  e="010001"
   *   } # data
   * </pre>
   * </li>
   * <li>For private keys:
   * <pre>
   *   char type="RSA"
   *   char owner="USERID"
   *   int  number="1"
   *   int  version="1"
   *   data {
   *     int  public="0"
   *     bin  n="XXX"
   *     bin  e="010001"
   *     bin  p="XXX"
   *     bin  q="XXX"
   *     bin  dmp1="XXX"
   *     bin  dmq1="XXX"
   *     bin  iqmp="XXX"
   *     bin  d="XXX"
   *   } # data
   * </pre>
   * </li>
   * </ul>
   */
  RSAKey(bool cryptKey = false, GWEN_DB_NODE *n=0);
  RSAKey(const RSAKey &k);

  ~RSAKey();

  /**
   * Stores the key data in the given node. See constructor for details.
   */
  bool toDB(GWEN_DB_NODE *n) const;

  /**
   * create a pair of rsa-keys, one public, one private
   */
  DLLIMPORT static void generateKeyPair(unsigned int keyLength,
                                        RSAKey **privKey,
                                        RSAKey **pubKey);

  /**
   * is this the public key?
   */
  bool isPublicKey() const;

  /**
   * is this a key for encrypting (or for signing)
   */
  bool isCryptoKey() const;

  /**
   * verify the signature of given data
   * (this key has to be the public one)
   */
  bool verify(const string &signature);

  /**
   * sign data
   * (this key should be the private one, otherways it would be pretty
   * senseless...)
   */
  bool sign();

  /**
   * encrypt data
   */
  bool encrypt();

  /**
   * decrypt data
   */
  bool decrypt();

  /**
   * return the version-number of the key
   */
  int number() const;

  /**
   * Set a new version-number for this key.
   */
  void setNumber(int newNumber);

  /**
   * return the sub-version-number of the key
   */
  int version() const;

  /**
   * Set a new sub-version-number for this key
   */
  void setVersion(int newVersion);

  /**
   * return the userid that belongs to this key (intitute-key: the
   * unique id for this key)
   */
  const string &userId() const;

  /**
   * set owner of this key
   */
  void setUserId(const string &id);

  /**
   * return the binary representation of the modulus if this is a public key (otherways result is undefined)
   */
  string getModulusData() const;

  /**
   * return the binary representation of the exp. (should by 3 byte long with exp = 65537)
   */
  string getExpData() const;

  /**
   * return the ini-letter-modulus that is needed to print the ini-letter
   */
  string getIniLetterModulus() const;

  /**
   * return the ini-letter-exponent that is needed to print the ini-letter
   */
  string getIniLetterExponent() const;

  /**
   * return the ini-letter-hash that is needed to print the ini-letter
   */
  string getIniLetterHash() const;


  /**
   * should this be a signature-key
   */
  void setToSignKey(bool signKey);
private:
  bool _isCryptKey;
  GWEN_CRYPTKEY *_key;
  string _userId;

  string _ripe(const string &src) const;

};

} /* namespace HBCI */

typedef HBCI::RSAKey HBCI_RSAKey;
extern "C" {
#else /* __cplusplus */
  typedef struct HBCI_RSAKey HBCI_RSAKey;
#endif /* __cplusplus */

  /** Constructor */
  HBCI_RSAKey *HBCI_RSAKey_new(int cryptKey, GWEN_DB_NODE *n);
  /**
   * is this the public key?
   */
  int HBCI_RSAKey_isPublicKey(const HBCI_RSAKey *h);

  /**
   * is this a key for encrypting (or for signing)
   */
  int HBCI_RSAKey_isCryptoKey(const HBCI_RSAKey *h);


  /**
   * return the version-number of the key
   */
  int HBCI_RSAKey_number(const HBCI_RSAKey *h);

  /**
   * return the sub-version-number of the key
   */
  int HBCI_RSAKey_version(const HBCI_RSAKey *h);

  /**
   * Set a new version-number for this key.
   */
  void HBCI_RSAKey_setNumber(HBCI_RSAKey *h, int newNumber);

  /**
   * Set a new sub-version-number for this key
   */
  void HBCI_RSAKey_setVersion(HBCI_RSAKey *h, int newVersion);

  /**
   * return the userid that belongs to this key (intitute-key: the
   * unique id for this key)
   */
  const char *HBCI_RSAKey_userId(const HBCI_RSAKey *h);

  /**
   * return the binary representation of the modulus if this is a public key (otherways result is undefined)
   *
   * @param h The object to perform this action on
   * @param buffer An allocated buffer to receive this data
   * @param maxbuf The maximum size of the given buffer
   * @return The number of bytes in the buffer that have been used
   */
  unsigned HBCI_RSAKey_getModulusData(const HBCI_RSAKey *h, 
				      void *buffer, unsigned maxbuf);

  /**
   * return the binary representation of the exp. (should by 3 byte long with exp = 65537)
   *
   * @param h The object to perform this action on
   * @param buffer An allocated buffer to receive this data
   * @param maxbuf The maximum size of the given buffer
   * @return The number of bytes in the buffer that have been used
   */
  unsigned HBCI_RSAKey_getExpData(const HBCI_RSAKey *h, 
				  void *buffer, unsigned maxbuf);

  /**
   * return the ini-letter-modulus that is needed to print the ini-letter
   */
  char * HBCI_RSAKey_getIniLetterModulus(const HBCI_RSAKey *h);

  /**
   * return the ini-letter-exponent that is needed to print the ini-letter
   */
  char * HBCI_RSAKey_getIniLetterExponent(const HBCI_RSAKey *h);

  /**
   * return the ini-letter-hash that is needed to print the ini-letter
   */
  char * HBCI_RSAKey_getIniLetterHash(const HBCI_RSAKey *h);
  

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* HBCIRSAKEY_H */
