/***************************************************************************
                          hbci.cpp  -  description
                             -------------------
    begin                : Sun May 27 2001
    copyright            : (C) 2001 by fabian kaiser
    email                : fabian.kaiser@gmx.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "hbci.h"

#include <string>
#include <stdio.h>
#include <list>

#include "assert.h"
#include "error.h"
#include "pointer.h"
#include "auth.h"
#include "interactor.h"
#include "hbcistring.h"

#include <gwenhywfar/debug.h>


namespace HBCI {

int Hbci::_debugLevel=0;


static string numToOneChar(unsigned int num){
  return (num < 10 ?
          // from 0 to 9 return the one-digit number
          String::num2string(num) :
          (num < 10 + 26 ?
           // from 10 to 35 return a lower case character
	   string(1, 'a' + (num - 10)) :
	   (num < 10 + 26 + 23 ?
	    // from 36 to 59 return an upper case character
	    string(1, 'A' + (num - 36)) :
	    // 60 and up: give up and return upper case X
	    "X")));
}


Hbci::Hbci(bool readonly, bool retrievalonly)
: _readonly(readonly)
, _retrievalOnly(retrievalonly)
  , _pluginManager(PluginManager(this))
, _plugin_path(PLUGIN_PATH){
  // set the system name and version to default values
  setSystemName(HBCI_SYSTEM_NAME);
  // set the system version -- we cannot use OPENHBCI_VERSION_STRING
  // directly because that one might easily be longer than 5
  // characters.
  setSystemVersion(String::num2string(OPENHBCI_VERSION_MAJOR)
                   + "."
                   + numToOneChar(OPENHBCI_VERSION_MINOR)
                   + "."
                   + numToOneChar(OPENHBCI_VERSION_PATCHLEVEL));

  /* this is a default authentificator, this may be overwritten by the
   * main program.*/
  _auth=new Auth(this);
  _interactor=new Interactor();
}


Hbci::~Hbci(){
}


Pointer<Plugin> Hbci::ensurePlugin(const string &name,
				   PluginType pt){
  Pointer<Plugin> p;

  DBG_INFO(0, "Ensuring plugin \"%s\"", name.c_str());
  p=_pluginManager.ensurePlugin(name, pt);
  p.setDescription("HBCI::ensurePlugin: Plugin");
  return p;
}


Pointer<Plugin> Hbci::findPlugin(const string &name,
				 PluginType pt){
  return _pluginManager.findPlugin(name, pt);
}


const string& Hbci::pluginPath() const {
    return _plugin_path;
}


Pointer<Medium> Hbci::mountCustomersMedium(int country,
                                           const string &bankCode,
                                           Pointer<Customer> cust){
  Pointer<User> u;
  Pointer<Medium> m;
  Error err;

  u=cust.ref().user();
  m=u.ref().medium();
  if (_mountedMedium.isValid()) {
    if (m!=_mountedMedium) {
      while (_mountedMedium.ref().isMounted()) {
        err=_mountedMedium.ref().unmountMedium();
        if (!err.isOk()) {
          DBG_WARN(0,
                   "Ignoring error while unmounting medium: %s",
                   err.errorString().c_str());
          break;
        }
      } // while
    }
    _mountedMedium=0;
  }

  err=m.ref().mountMedium();
  if (!err.isOk())
    throw err;
  _mountedMedium=m;

  err=_mountedMedium.ref().selectContext(country,
                                         bankCode,
                                         u.ref().userId());
  if (!err.isOk())
    throw err;
  return _mountedMedium;
}


Error Hbci::unmountCurrentMedium() {
  Error err;

  if (_mountedMedium.isValid()) {
    while (_mountedMedium.ref().isMounted()) {
      err=_mountedMedium.ref().unmountMedium();
      if (!err.isOk())
        return err;
    } // while
    _mountedMedium=0;
  }
  return Error();
}



Pointer<ParserPlugin> Hbci::getParser(const string &name) {
  Pointer<Plugin> pl;

  pl=ensurePlugin(name, PluginTypeParser);
  if (!pl.isValid())
    // most likely the PluginManager already threw an Error, but anyway...
    throw Error("Hbci::getParser",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"Could not find parser",
		name);
  return pl.cast<ParserPlugin>();
}


int Hbci::debugLevel() {
  return _debugLevel;
}


void Hbci::setDebugLevel(int l) {
  _debugLevel=l;
}




void Hbci::libraryVersion(int &major,
                          int &minor,
                          int &patchlevel) {
  major=OPENHBCI_VERSION_MAJOR;
  minor=OPENHBCI_VERSION_MINOR;
  patchlevel=OPENHBCI_VERSION_PATCHLEVEL;
};



void Hbci::libraryVersion(int &major,
                          int &minor,
                          int &patchlevel,
                          int &build) {
  libraryVersion(major,minor,patchlevel);
  build=OPENHBCI_VERSION_BUILD;
};


void Hbci::setSystemName(const string &n) {
  // only use 25 characters starting from index 0 here
  _systemName=n.substr(0,25);
};


void Hbci::setSystemVersion(const string &s) {
  // only use 5 characters starting from index 0 here
  _systemVersion=s.substr(0,5);
};




} // namespace HBCI

// ----------- C Wrappers

void HBCI_Hbci_setInteractor(HBCI_Hbci *h, HBCI_Interactor *inter, 
			int autoDelete)
{
    assert(h);
    HBCI::Pointer<HBCI_Interactor> ip = inter;
    ip.setAutoDelete(autoDelete); 
    h->setInteractor(ip);
}
HBCI_Interactor *HBCI_Hbci_interactor(const HBCI_Hbci *h)
{
    assert(h);
    return h->interactor().ptr();
}
void HBCI_Hbci_libraryVersion(int *major, int *minor, int *plevel)
{
    assert(major);
    assert(minor);
    assert(plevel);
    HBCI::Hbci::libraryVersion(*major, *minor, *plevel);
}
void HBCI_Hbci_libraryVersion_build(int *major, int *minor, int *plevel,
				    int *pbuild)
{
    assert(major);
    assert(minor);
    assert(plevel);
    assert(pbuild);
    HBCI::Hbci::libraryVersion(*major, *minor, *plevel, *pbuild);
}
const char *HBCI_Hbci_pluginPath(const HBCI_Hbci *h)
{
    assert(h);
    return h->pluginPath().c_str();
}
const HBCI_Plugin *HBCI_Hbci_ensurePlugin(HBCI_Hbci *h, const char *name,
					  PluginType pt)
{
    assert(h);
    HBCI::Pointer<HBCI::Plugin> pp = 
      h->ensurePlugin(string(name ? name : ""), pt);
    return pp.isValid() ? pp.ptr() : NULL;
}
const HBCI_Plugin *HBCI_Hbci_findPlugin(HBCI_Hbci *h, const char *name,
					PluginType pt)
{
    assert(h);
    HBCI::Pointer<HBCI::Plugin> pp = 
      h->findPlugin(string(name ? name : ""), pt);
    return pp.isValid() ? pp.ptr() : NULL;
}
const char *HBCI_Hbci_systemName(const HBCI_Hbci *h)
{
    assert(h);
    return h->systemName().c_str();
}
void HBCI_Hbci_setSystemName(HBCI_Hbci *h, const char *n)
{
    assert(h);
    h->setSystemName(string(n ? n : ""));
}
const char *HBCI_Hbci_systemVersion(const HBCI_Hbci *h)
{
    assert(h);
    return h->systemVersion().c_str();
}
void HBCI_Hbci_setSystemVersion(HBCI_Hbci *h, const char *n)
{
    assert(h);
    h->setSystemVersion(string(n ? n : ""));
}
const char *HBCI_Hbci_applicationDataDir(const HBCI_Hbci *h)
{
    assert(h);
    return h->applicationDataDir().c_str();
}
void HBCI_Hbci_setApplicationDataDir(HBCI_Hbci *h, const char *s)
{
    assert(h);
    h->setApplicationDataDir(string(s ? s : ""));
}
void HBCI_Hbci_setDebugLevel(int l)
{
    HBCI::Hbci::setDebugLevel(l);
}
int HBCI_Hbci_debugLevel()
{
    return HBCI::Hbci::debugLevel();
}

