/***************************************************************************
  $RCSfile: job.cpp,v $
  -------------------
  cvs         : $Id: job.cpp,v 1.12 2004/01/19 18:53:49 aquamaniac Exp $
  begin       : Sat Jul 05 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "job.h"
#include <gwenhywfar/debug.h>
#include "hbcistring.h"

namespace HBCI {


Job::Job(const string &name,
	 Pointer<Customer> cu,
	 const string &accountId) throw(Error)
  :_customer(cu), _name(name), _account(accountId)
{
  Pointer<User> u;
  Pointer<Bank> b;

  if (!cu.isValid()) {
    throw Error("Job::Job",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"Invalid customer pointer");
  }

  /* get some basic information */
  u=_customer.ref().user();
  if (!u.isValid()) {
    throw Error("Job::Job",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"Invalid user pointer");
  }

  b=u.ref().bank();
  if (!b.isValid()) {
    throw Error("Job::Job",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"Invalid bank pointer");
  }

  _msgEngine=b.ref().messageEngine();
  if (!_msgEngine.isValid()) {
    throw Error("Job::Job",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"Invalid message engine pointer");
  }

  // create the MsgJob
  _job=_msgEngine.ref().jobFactory(name,
                                   cu,
				   accountId);
  if (!_job.isValid()) {
    throw Error("Job::Job",
		ERROR_LEVEL_NORMAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"Unsupported job",
		name);
  }
}


Job::~Job(){
}



const char *Job::getProperty(const string &name,
			     const string &defValue) throw(Error) {
  if (!_job.isValid())
    throw Error("Job::getProperty",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().getProperty(name, defValue);
}



int Job::getIntProperty(const string &name,
			int defValue) throw(Error){
  if (!_job.isValid())
    throw Error("Job::getIntProperty",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().getIntProperty(name, defValue);
}



string Job::getBinProperty(const string &name,
			   const string &defValue) throw(Error){
  if (!_job.isValid())
    throw Error("Job::getBinProperty",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().getBinProperty(name, defValue);
}



void Job::setProperty(const string &name, const string &value) throw(Error){
  GWEN_DB_NODE *gr;

  if (!_job.isValid())
    throw Error("Job::setProperty",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  gr=_job.ref().jobData();
  assert(gr);
  gr=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_DEFAULT, "request");
  assert(gr);
  GWEN_DB_SetCharValue(gr,
                       GWEN_DB_FLAGS_OVERWRITE_VARS,
                       name.c_str(),
                       value.c_str());
}



void Job::setIntProperty(const string &name, int value) throw(Error){
  GWEN_DB_NODE *gr;

  if (!_job.isValid())
    throw Error("Job::setProperty",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  gr=_job.ref().jobData();
  assert(gr);
  gr=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_DEFAULT, "request");
  assert(gr);
  GWEN_DB_SetIntValue(gr,
                      GWEN_DB_FLAGS_OVERWRITE_VARS,
                      name.c_str(),
                      value);
}



void Job::setBinProperty(const string &name,
			 const string &value) throw(Error){
  GWEN_DB_NODE *gr;

  if (!_job.isValid())
    throw Error("Job::setBinProperty",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  gr=_job.ref().jobData();
  assert(gr);
  gr=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_DEFAULT, "request");
  assert(gr);
  GWEN_DB_SetBinValue(gr,
                      GWEN_DB_FLAGS_OVERWRITE_VARS,
                      name.c_str(),
                      (void*)(value.c_str()),
                      value.length());
}



void Job::addRequestData(const string &path, GWEN_DB_NODE *gr){
  if (!_job.isValid())
    throw Error("Job::addRequestData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().addRequestData(path, gr);
}



bool Job::needsSignature() const{
  if (!_job.isValid())
    throw Error("Job::needsSignature",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().needsSignature();
}



bool Job::needsEncryption() const{
  if (!_job.isValid())
    throw Error("Job::needsEncryption",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().needsEncryption();
}



Error Job::enqueueJob(Pointer<MsgQueue> queue){
  if (!_job.isValid())
    return Error("Job::enqueueJob",
		 ERROR_LEVEL_INTERNAL,
		 HBCI_ERROR_CODE_UNKNOWN,
		 ERROR_ADVISE_DONTKNOW,
		 "No MsgJob",
		 _name);
  return queue.ref().addJob(_job);
}



HBCI_QUEUECHECK_RESULT Job::checkEnqueue(Pointer<MsgQueue> queue){
  if (!_job.isValid()) {
    DBG_ERROR(0, "No MsgJob");
    return HBCI_QUEUECHECK_RESULT_ERROR;
  }
  return queue.ref().checkAddJob(_job);
}



void Job::addSigner(Pointer<Customer> cu){
  if (!_job.isValid())
    throw Error("Job::addSigner",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  _job.ref().addSigner(cu);
}



GWEN_DB_NODE *Job::jobData(){
  if (!_job.isValid())
    throw Error("Job::jobData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().jobData();
}


const list<MsgJob::Result> &Job::results() const {
  if (!_job.isValid())
    throw Error("Job::results",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  return _job.ref().results();

}



GWEN_DB_NODE *Job::responseData() {
  GWEN_DB_NODE *gr;

  if (!_job.isValid())
    throw Error("Job::responseData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  gr=_job.ref().jobData();
  if (!gr) {
    throw Error("Job::responseData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No data for this job",
		_name);
  }
  gr=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_DEFAULT, "response");
  assert(gr);
  return gr;
}



GWEN_DB_NODE *Job::paramData(){
  GWEN_DB_NODE *gr;

  if (!_job.isValid())
    throw Error("Job::responseData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  gr=_job.ref().jobData();
  if (!gr) {
    throw Error("Job::responseData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No data for this job",
		_name);
  }
  gr=GWEN_DB_GetGroup(gr, GWEN_DB_FLAGS_DEFAULT, "params");
  return gr;
}



GWEN_DB_NODE *Job::debugData(){
  GWEN_DB_NODE *gr;

  if (!_job.isValid())
    throw Error("Job::debugData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No MsgJob",
		_name);
  gr=_job.ref().jobData();
  if (!gr) {
    throw Error("Job::debugData",
		ERROR_LEVEL_INTERNAL,
		HBCI_ERROR_CODE_UNKNOWN,
		ERROR_ADVISE_DONTKNOW,
		"No data for this job",
		_name);
  }
  return GWEN_DB_Group_dup(gr);
}



bool Job::attachable(){
  if (!_job.isValid())
    throw Error("Job::attachable",
                ERROR_LEVEL_INTERNAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "No MsgJob",
                _name);
  return _job.ref().attachable();
}



string Job::description() const{
  if (!_job.isValid())
    throw Error("Job::description",
                ERROR_LEVEL_INTERNAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "No MsgJob",
                _name);
  return _job.ref().description();
}



bool Job::needResponse(){
  if (!_job.isValid())
    throw Error("Job::needResponse",
                ERROR_LEVEL_INTERNAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "No MsgJob",
                _name);
  return _job.ref().needResponse();
}



void Job::removeResults(unsigned int code){
  if (!_job.isValid())
    throw Error("Job::removeResults",
                ERROR_LEVEL_INTERNAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "No MsgJob",
                _name);
  _job.ref().removeResults(code);
}



void Job::addResult(unsigned int code, const string &text){
  if (!_job.isValid())
    throw Error("Job::addResult",
                ERROR_LEVEL_INTERNAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "No MsgJob",
                _name);
  return _job.ref().addResult(MsgJob::Result(code, text));
}



}; /* namespace */


const char *HBCI_Job_name(const HBCI_Job *job)
{
    assert(job);
    return job->name().c_str();
}
char *HBCI_Job_description(const HBCI_Job *job)
{
    assert(job);
    return hbci_strdup(job->description());
}
const HBCI_Customer *HBCI_Job_customer(const HBCI_Job *job)
{
    assert(job);
    return job->customer().ptr();
}
const char *HBCI_Job_accountId(const HBCI_Job *job)
{
    assert(job);
    return job->accountId().c_str();
}

const char *HBCI_Job_getProperty(HBCI_Job *job, const char *name,
				 const char *value)
{
    assert(job);
    return job->getProperty(name ? string(name) : "", 
			    value ? string(value) : "");
}
int HBCI_Job_getIntProperty(HBCI_Job *job, const char *name,
			    int defValue)
{
    assert(job);
    return job->getIntProperty(name ? string(name) : "", defValue);
}
const void *HBCI_Job_getBinProperty(HBCI_Job *job, const char *name,
				    const void *defValue, 
				    unsigned int defValueSize,
				    unsigned int *returnValueSize)
{
    assert(job);
    string d(static_cast<const char*>(defValue), defValueSize);
    string res = job->getBinProperty(name ? string(name) : "", 
				     d);
    *returnValueSize = res.size();
    return static_cast<const void*>(res.data());
}

void HBCI_Job_setIntProperty(HBCI_Job *job, const char *name, int value)
{
    assert(job);
    job->setIntProperty(name ? string(name) : "", value);
}
void HBCI_Job_setBinProperty(HBCI_Job *job, const char *name, 
			     const void *value, unsigned int dataSize)
{
    assert(job);
    string d(static_cast<const char*>(value), dataSize);
    job->setBinProperty(name ? string(name) : "", d);
}
void HBCI_Job_addRequestData(HBCI_Job *job, const char *path,
			     GWEN_DB_NODE *gr)
{
    assert(job);
    job->addRequestData(path ? string(path) : "", gr);
}

void HBCI_Job_setProperty(HBCI_Job *job, const char *name, 
			  const char *value)
{
    assert(job);
    job->setProperty(name ? string(name) : "", 
		     value ? string(value) : "");
}


void HBCI_Job_addSigner(HBCI_Job *job, const HBCI_Customer *cu)
{
    assert(job);
    assert(cu);
    HBCI::Pointer<HBCI::Customer> cp = const_cast<HBCI::Customer*>(cu);
    cp.setAutoDelete(false);
    job->addSigner(cp);
}
GWEN_DB_NODE *HBCI_Job_responseData(HBCI_Job *job)
{
    assert(job);
    return job->responseData();
}
GWEN_DB_NODE *HBCI_Job_paramData(HBCI_Job *job)
{
    assert(job);
    return job->paramData();
}
GWEN_DB_NODE *HBCI_Job_debugData(HBCI_Job *job)
{
    assert(job);
    return job->debugData();
}
const list_HBCI_MsgJob_Result *
HBCI_Job_results(const HBCI_Job *job)
{
    assert(job);
    return &(job->results());
}
