/***************************************************************************
  $RCSfile: job.h,v $
  -------------------
  cvs         : $Id: job.h,v 1.13 2004/01/19 20:20:34 aquamaniac Exp $
  begin       : Sat Jul 05 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef JOB_H
#define JOB_H

#include <openhbci2/pointer.h>

#include <openhbci2/msgjob.h>
#include <openhbci2/msgengine.h>
#include <openhbci2/customer.h>
#include <gwenhywfar/db.h>

#ifdef __cplusplus
namespace HBCI {

  /**
   * This is the main class for performing a job using OpenHBCI.
   * Usually you don't create an object of this class directly. Instead you
   * instantiate @ref OutboxJob which inherits this class.
   * After creating such a job in most cases you need to set some properties
   * appropriate for this job, add it to a job queue (called @ref Outbox in
   * OpenHBCI) and perform @ref API::executeQueue.
   * @author Martin Preuss<martin@libchipcard.de>
   */
  class Job {
    friend class API;
  private:
    Pointer<MsgJob> _job;
    Pointer<Customer> _customer;
    Pointer<MessageEngine> _msgEngine;
    string _name;
    string _account;

    /** @name methods that are called from the class @ref API */
    /*@{*/
    /**
     * This actually enqueues the job into the lowlevel queue (not to be
     * confused with the @ref Outbox, which is situated on a higher level).
     */
    Error enqueueJob(Pointer<MsgQueue> queue);
    HBCI_QUEUECHECK_RESULT checkEnqueue(Pointer<MsgQueue> queue);

    /*@}*/

  protected:
    /**
     * @internal
     * This method returns a pointer to the jobs data.
     * It is not part of the OpenHBCI API !
     */
    virtual GWEN_DB_NODE *jobData();

    /**
     * Adds a result to this job. This is needed to add results created
     * by inheriting classes.
     */
    virtual void addResult(unsigned int code, const string &text);

    /**
     * @internal
     * Removes all results with the given code
     */
    virtual void removeResults(unsigned int code);

  public:
    /** @name Constructors and Destructor
     *
     */
    /*@{*/
    /**
     * Creates a job. This job is automatically looked up by the appropriate
     * message engine. So after calling this constructor the job parameters
     * are available (such as number of signatures needed etc.).
     * Usually you need to set some properties of this job before you can
     * execute it.
     * If this job is not supported (either by OpenHBCI, the bank or your
     * account access rights) then an exception is thrown.
     * @param name name of the job as it appears in the XML file
     * @param cu customer for whom this job is to be performed
     * @param accountId the id of your account. This is used to check whether
     * the given job is allowed with your account.
     */
    Job(const string &name, Pointer<Customer> cu,
	const string &accountId="") throw(Error);
    virtual ~Job();
    /*@}*/


    /** @name Informational Methods
     *
     */
    /*@{*/
    /**
     * Returns the name of this job.
     */
    const string &name() const { return _name;};

    /**
     * Return a short description of what the job is supposed to do.
     * This information is normally taken from the XML file.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    virtual string description() const;

    /** returns true if the job may have an attachment point */
    bool attachable();

    /**
     * @internal
     * @return true if the job needs a segment result
     */
    bool needResponse();

    /**
     * Returns the customer who owns this job.
     */
    Pointer<Customer> customer() const { return _customer;};

    /**
     * Returns the id of the account for which this job is to be performed.
     */
    const string &accountId() const { return _account;};
    /*@}*/


    /** @name Property Getter And Setter
     *
     * All the data of a job (request data, parameter data and response
     * data) is stored in a GWEN_DB_NODE inside the job. The different data
     * types are stored in different subgroups of that DB node:
     * <ul>
     *  <li>
     *   request data: stores data that is needed to perform the job.
     *   For jobs that work with accounts this is e.g. the account number.
     *   (group name: "request")
     *  </li>
     *  <li>
     *   parameter data: some descriptive data for this job (called
     *   "Bankparameterdaten" in German HBCI specs). Not all jobs have such
     *   data but most jobs do.
     *   (group name: "params")
     *  </li>
     *  <li>
     *   response data: of course this is only available after a job is
     *   executed. It then contains the response data sent by the bank.
     *   (group name: "response")
     *  </li>
     * </ul>
     *
     * Please note, that all setter methods have the path "request" prepended.
     * So if you call @ref setProperty("testvar", "testvalue") then actually
     * you are setting the variable <i>request/testvar</i> instead of just
     * <i>testvar</i>.
     * This is to protect some internal values from being modified by the
     * application.
     * So if you want to access the very same variable via @ref getProperty
     * you must use @ref getProperty("request/testvar", "").
     * Job parameters are stored in "params/". Since the setter functions
     * always use the prefix "request/" you can not modify the job
     * parameters (this is not a bug, it is intended).
     * The type of the property is defined in the XML file.
     */
    /*@{*/
    /**
     * Get the value of a property or the default value, if no value is set.
     * This method works on data of the following type defined in the XML
     * file:
     * <ul>
     *  <li>an</li>
     *  <li>ascii</li>
     *  <li>float</li>
     *  <li>all other data which is not covered by @ref getIntProperty or
     *  @ref getBinProperty</li>
     * </ul>
     *
     * @return The value of the property as a C string. Sorry for
     * dealing with const char* in C++. But the underlying Gwen
     * library stores the value this way, and only passing this const
     * char* up to here we can deal with this from the C wrappers,
     * too.
     */
    const char *getProperty(const string &name,
			    const string &defValue) throw(Error);
    /**
     * Get the integer value of a property or the default value, if no value
     * is set.
     * This method works on data of the following type defined in the XML
     * file:
     * <ul>
     *  <li>num</li>
     * </ul>
     */
    int getIntProperty(const string &name,
		       int defValue) throw(Error);
    /**
     * Get the binary value of a property or the default value, if no value
     * is set.
     * This method works on data of the following type defined in a XML
     * file:
     * <ul>
     *  <li>bin</li>
     * </ul>
     */
    string getBinProperty(const string &name,
			  const string &defValue) throw(Error);

    /**
     * Set the value of a property.
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     */
    void setProperty(const string &name, const string &value) throw(Error);

    /**
     * Set the integer value of a property.
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     */
    void setIntProperty(const string &name, int value) throw(Error);

    /**
     * Set the binary value of a property.
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     */
    void setBinProperty(const string &name, const string &value) throw(Error);

    /**
     * Add the copy of a GWEN_DB_NODE object to the request path.
     * This may be used to add groups of properties which have been already
     * assembled (such as transactions etc).
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     * This functions does NOT take over the ownership of the given DB node.
     */
    void addRequestData(const string &path, GWEN_DB_NODE *gr);
    /*@}*/

    /** @name Crypto Stuff
     *
     */
    /*@{*/
    /**
     * Returns <i>true</i> if this job needs a signature.
     * The job parameter <i>minsigs</i> of most jobs tells you how many
     * signatures this job actually needs (get it using
     * @ref getIntProperty("params/minsigs",0) )
     */
    bool needsSignature() const;

    /**
     * Returns <i>true</i> if this job needs encryption. A job will usually
     * encrypted using the owner of the job queue of which this job is part.
     */
    bool needsEncryption() const;

    /**
     * Adds a signer to a given job. However, if the job needs one signature
     * and none is given by the time the job is being prepared to be send to
     * the bank, the owner of the job is added to the internal signers list
     * automatically. So usually you don't need to call this method yourself.
     * It is only needed if a job needs more than one signature, or if you
     * want to sign a job as a different customer.
     */
    void addSigner(Pointer<Customer> cu);
    /*@}*/

    /** @name Result Handling
     *
     */
    /*@{*/
    /**
     * Returns a list of job responses. All codes above (and including) 9000
     * are errors, codes starting with 3xxx are warnings and all codes
     * below 1000 are definately ok.
     */
    const list<MsgJob::Result> &results() const;

    /**
     * Returns a pointer to the response data.
     * You MUST NOT free this pointer !
     */
    GWEN_DB_NODE *responseData();

    /**
     * Returns a pointer to the parameter data.
     * If there is no parameter data for rhis job then 0 is returned.
     * You MUST NOT free this pointer !
     */
    GWEN_DB_NODE *paramData();
    /*@}*/

    /** @name Debugging
     *
     * These methods are not officially part of the API, so you should not
     * rely on them. They are used for debugging purposes while the new
     * version of OpenHBCI is not considered stable.
     */
    /*@{*/
    /**
     * This is just a debugging method, it might disappear without further
     * notice !
     * Returns a private copy of the whole job data.
     * The caller is responsible for freeing this data by calling
     * Config_free().
     */
    GWEN_DB_NODE *debugData();
    /*@}*/


  };

}; /* namespace */

typedef HBCI::Job HBCI_Job;
extern "C" {
#else /* __cplusplus */
  typedef struct HBCI_Job HBCI_Job;
#endif /* __cplusplus */

    /** @name Informational Methods
     *
     */
    /*@{*/
    /**
     * Returns the name of this job.
     */
    const char *HBCI_Job_name(const HBCI_Job *job);
    
    /**
     * Return a short description of what the job is supposed to do.
     * This information is normally taken from the XML file.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    char *HBCI_Job_description(const HBCI_Job *job);

    /**
     * Returns the customer who owns this job.
     */
    const HBCI_Customer *HBCI_Job_customer(const HBCI_Job *job);

    /**
     * Returns the id of the account for which this job is to be performed.
     */
    const char *HBCI_Job_accountId(const HBCI_Job *job);
    /*@}*/



    /**
     * Get the value of a property or the default value, if no value is set.
     * This method works on data of the following type defined in the XML
     * file:
     * <ul>
     *  <li>an</li>
     *  <li>ascii</li>
     *  <li>float</li>
     *  <li>all other data which is not covered by @ref getIntProperty or
     *  @ref getBinProperty</li>
     * </ul>
     */
    const char *HBCI_Job_getProperty(HBCI_Job *job, const char *name,
				     const char *defValue);
    /**
     * Get the integer value of a property or the default value, if no value
     * is set.
     * This method works on data of the following type defined in the XML
     * file:
     * <ul>
     *  <li>num</li>
     * </ul>
     */
    int HBCI_Job_getIntProperty(HBCI_Job *job, const char *name,
				int defValue);
    /**
     * Get the binary value of a property or the default value (whose
     * size is passed by defValueSize), if no value is set. The size
     * of the returned data is written into the variable
     * returnValueSize.
     *
     * This method works on data of the following type defined in a XML
     * file:
     * <ul>
     *  <li>bin</li>
     * </ul>
     */
    const void *HBCI_Job_getBinProperty(HBCI_Job *job, const char *name,
					const void *defValue, 
					unsigned int defValueSize,
					unsigned int *returnValueSize);
  
    /**
     * Set the value of a property.
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     */
    void HBCI_Job_setProperty(HBCI_Job *job, const char *name, 
			      const char *value);

    /**
     * Set the integer value of a property.
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     */
    void HBCI_Job_setIntProperty(HBCI_Job *job, const char *name, int value);

    /**
     * Set the binary value of a property. The size of the binary data
     * is passed by the argument dataSize.
     *
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     */
    void HBCI_Job_setBinProperty(HBCI_Job *job, const char *name, 
				 const void *value, unsigned int dataSize);

    /**
     * Add the copy of a GWEN_DB_NODE object to the request path.
     * This may be used to add groups of properties which have been already
     * assembled (such as transactions etc).
     * Please note that this function always inserts the path "request" to
     * the given name to protect itself against accidental manipulation of
     * internal data (such as job parameters etc).
     * This functions does NOT take over the ownership of the given DB node.
     */
    void HBCI_Job_addRequestData(HBCI_Job *job, const char *path,
				 GWEN_DB_NODE *gr);

    /**
     * Adds a signer to a given job. However, if the job needs one signature
     * and none is given by the time the job is being prepared to be send to
     * the bank, the owner of the job is added to the internal signers list
     * automatically. So usually you don't need to call this method yourself.
     * It is only needed if a job needs more than one signature, or if you
     * want to sign a job as a different customer.
     */
    void HBCI_Job_addSigner(HBCI_Job *job, const HBCI_Customer *cu);

    /** @name Result Handling
     *
     */
    /*@{*/
    /**
     * Returns a pointer to the response data.
     * You MUST NOT free this pointer !
     */
    GWEN_DB_NODE *HBCI_Job_responseData(HBCI_Job *job);

    /**
     * Returns a pointer to the parameter data.
     * If there is no parameter data for rhis job then 0 is returned.
     * You MUST NOT free this pointer !
     */
    GWEN_DB_NODE *HBCI_Job_paramData(HBCI_Job *job);
    /**
     * Returns a list of job responses. All codes above (and including) 9000
     * are errors, codes starting with 3xxx are warnings and all codes
     * below 1000 are definately ok.
     */
    const list_HBCI_MsgJob_Result *HBCI_Job_results(const HBCI_Job *job);

    /**
     * This is just a debugging method, it might disappear without further
     * notice !
     * Returns a private copy of the whole job data.
     * The caller is responsible for freeing this data by calling
     * Config_free().
     */
    GWEN_DB_NODE *HBCI_Job_debugData(HBCI_Job *job);

    /*@}*/


#ifdef __cplusplus 
}
#endif /* __cplusplus */

#endif /* JOB_H */


