/***************************************************************************
 $RCSfile: msgjob.h,v $
                             -------------------
    cvs         : $Id: msgjob.h,v 1.11 2004/01/19 20:20:35 aquamaniac Exp $
    begin       : Sun Jul 06 2003
    copyright   : (C) 2003 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef MSGJOB_H
#define MSGJOB_H

#ifdef __cplusplus

namespace HBCI {
  class MsgJob;
}

#include <openhbci2/customer.h>
#include <gwenhywfar/db.h>
#include <string>
#include <list>
using namespace std;


namespace HBCI {

/**
 * A message reference is a kind of unique id by which a message
 * can be referred to. It consists of the dialogId and the message number.
 * This is used for status reports send by the bank.
 */
class MessageReference {
private:
  string _dialogId;
  int _msgNumber;
public:
  MessageReference();
  /** Constructor with dialog id and message number. */
  MessageReference(const string &did, int msgnum);
  /** Constructor for a HBCI data element group (DEG). This is just
   the same string as the one returned by toString(). (FIXME: is
   this correct?) */
  MessageReference(const string &deg);
  /** Destructor */
  ~MessageReference();
  /** Returns the dialog id. */
  const string &dialogId() const { return _dialogId;};
  /** Returns the message number. */
  int messageNumber() const { return _msgNumber;};
  /** Returns the HBCI-representation of a message reference in one
   string. */
  string toString() const;
  /** Equality */
  bool operator==(const MessageReference &ref) const;
  /** Comparison. Returns -1 when this MessageReference is smaller
   than the other, returns 0 when equal, and returns 1 when this
   one is greater than the other. The dialogId are compared
   lexicographically, and the messageNumber numerically. */
  int cmp(const MessageReference &other) const;
};
} /* namespace */
typedef HBCI::MessageReference HBCI_MessageReference;
extern "C" {
#else /* __cplusplus */
  typedef struct HBCI_MessageReference HBCI_MessageReference;
#endif /* __cplusplus */


  /** @name HBCI_MessageReference */
  /*@{*/
  /** Constructor with dialog id and message number. */
  extern HBCI_MessageReference *
    HBCI_MessageReference_new(const char *did, int msgnum);
  /** Constructor for a HBCI data element group (DEG). This is just
   the same string as the one returned by toString(). (FIXME: is
   this correct?) */
  extern HBCI_MessageReference *
    HBCI_MessageReference_new_deg(const char *deg);
  /** Destructor */
  extern void HBCI_MessageReference_delete(HBCI_MessageReference *h);
  /** Returns the dialog id. */
  extern const char *
    HBCI_MessageReference_dialogId(const HBCI_MessageReference *h);
  /** Returns the message number. */
  extern int
    HBCI_MessageReference_messageNumber(const HBCI_MessageReference *h);
  /** Returns the HBCI-representation of a message reference in one
   * string.
   *
   * @note The returned char array is owned by the caller and must be
   * free'd when no longer needed */
  extern char *
    HBCI_MessageReference_toString(const HBCI_MessageReference *h);
  /** Equality */
  extern int HBCI_MessageReference_equal(const HBCI_MessageReference *h1,
					 const HBCI_MessageReference *h2);
  /** Comparison. Returns -1 when MessageReference h1 is smaller than
   h2, returns 0 when equal, and returns 1 when h1 is greater than
   h2. The dialogId are compared lexicographically, and the
   messageNumber numerically. */
  extern int HBCI_MessageReference_cmp(const HBCI_MessageReference *h1,
				       const HBCI_MessageReference *h2);
  /*@}*/

#ifdef __cplusplus
} /* extern C */
#endif



#ifdef __cplusplus

namespace HBCI {

  /**
   * @internal
   * This is an abstract base class for Jobs.
   * It is not part of the OpenHBCI API (it is part of the OpenHBCI
   * plugin interface, though)
   * You are probably looking for @ref OutboxJob.
   * @author Martin Preuss<martin@libchipcard.de>
   */
  class MsgJob {
  public:
    class Result {
    private:
      int _code;
      string _message;
      string _param;
      string _property;
      string _value;
      string _explanation;
    public:
      Result(int code=-1,
             const string &message="",
             const string &param="",
	     const string &property="",
             const string &value="",
             const string &explanation="");
      ~Result();
      int resultCode() const { return _code; };
      const string &resultMessage() const { return _message;};

      const string &param() const { return _param;};

      /**
       * Returns the name of the property which provoced the error (if
       * code is >=9000, and only if the reason was a bad value)
       */
      const string &property() const { return _property;};

      /**
       * Returns the value of the property which provoked the error (if
       * code is >=9000, and only if the reason was a bad value)
       */
      const string &value() const { return _value;};

      /**
       * Returns an explanatory string for this result.
       * Not all MessageEngines have to support this feature.
       */
      const string &explanation() const { return _explanation;};
    };

  private:
    string _name;
    string _accountnr;
    Pointer<Customer> _owner;

  protected:
    MsgJob(Pointer<Customer> cu,
	   const string &name,
	   const string &accountnr="");
  public:
    virtual ~MsgJob();

    /** returns true if the jobs wants to be alone in a queue */
    virtual bool single() { return false;};
    /** returns true if the job needs a segment result */
    virtual bool needResponse() { return 1;};
    /** returns the maximum number of jobs of this kind per message */
    virtual int jobsPerMessage() { return 1;};
    /** returns true if the job may have  an attachment point */
    virtual bool attachable() { return false;};
    virtual int signatureMinimum() { return 0;};

    virtual const string &name() const { return _name;};
    virtual string description() const { return "";};

    virtual const string &accountNumber() const { return _accountnr;};
    virtual Pointer<Customer> owner() { return _owner;};

    virtual const char *getProperty(const string &name,
				    const string &defValue)=0;
    virtual int getIntProperty(const string &name,
			       int defValue)=0;
    virtual string getBinProperty(const string &name,
				  const string &defValue)=0;
    virtual void setProperty(const string &name, const string &value)=0;
    virtual void setIntProperty(const string &name, int value)=0;
    virtual void setBinProperty(const string &name, const string &value)=0;

    /**
     * This functions does NOT take over the ownership of the given DB node.
     * @param path group where to add the given data (If empty all the data
     * inside the given group is stored in the root of the jobs request data)
     * If not empty then the given group will be created (may lead to multiple
     * groups with the same name, but in most cases this is desired).
     */
    virtual void addRequestData(const string &path, GWEN_DB_NODE *gr) =0;
    virtual bool needsSignature() const =0;
    virtual bool needsEncryption() const =0;
    virtual const list<Result> &results() const =0;
    virtual void removeResults(unsigned int code)=0;
    virtual void addResult(const Result &r)=0;
    virtual const MessageReference &messageReference() const =0;
    virtual void addSigner(Pointer<Customer> cu) =0;
    virtual const list<Pointer<Customer> > &signers() const =0;
    virtual GWEN_DB_NODE *jobData() =0;
  };

} /* namespace */
typedef HBCI::MsgJob HBCI_MsgJob;
typedef HBCI::MsgJob::Result HBCI_MsgJob_Result;
typedef std::list<HBCI::MsgJob::Result> list_HBCI_MsgJob_Result;
extern "C" {
#else /* __cplusplus */
  typedef struct HBCI_MsgJob HBCI_MsgJob;
  typedef struct HBCI_MsgJob_Result HBCI_MsgJob_Result;
  typedef struct list_HBCI_MsgJob_Result list_HBCI_MsgJob_Result;
#endif /* __cplusplus */

  int HBCI_MsgJob_Result_resultCode(const HBCI_MsgJob_Result *r);
  const char *HBCI_MsgJob_Result_resultMessage(const HBCI_MsgJob_Result *r);
  const char *HBCI_MsgJob_Result_param(const HBCI_MsgJob_Result *r);

  /**
   * Returns the name of the property which provoced the error (if
   * code is >=9000, and only if the reason was a bad value)
   */
  const char *HBCI_MsgJob_Result_property(const HBCI_MsgJob_Result *r);

  /**
   * Returns the value of the property which provoked the error (if
   * code is >=9000, and only if the reason was a bad value)
   */
  const char *HBCI_MsgJob_Result_value(const HBCI_MsgJob_Result *r);

  /**
   * Returns an explanatory string for this result.
   * Not all MessageEngines have to support this feature.
   */
  const char *HBCI_MsgJob_Result_explanation(const HBCI_MsgJob_Result *r);

  /** @name list of HBCI_MsgJob_Result */
  /*@{*/
  extern int list_HBCI_MsgJob_Result_size(const list_HBCI_MsgJob_Result *l);
  /** Callback function for @ref list_HBCI_MsgJob_Result_foreach. */
  typedef void *(*list_HBCI_MsgJob_Result_cb) (const HBCI_MsgJob_Result *res,
					       void *user_data);
  /** Traverses the list, calling the callback function 'func' on
   * each list element.  Traversal will stop when 'func' returns a
   * non-NULL value, and the routine will return with that
   * value. Otherwise the routine will return NULL.
   *
   * @param list The list to traverse.
   * @param func The function to be called with each list element.
   * @param user_data A pointer passed on to the function 'func'.
   * @return The non-NULL pointer returned by 'func' as soon as it
   * returns one. Otherwise (i.e. 'func' always returns NULL)
   * returns NULL.
   * @author Christian Stimming <stimming@tuhh.de> */
  extern void *
  list_HBCI_MsgJob_Result_foreach (const list_HBCI_MsgJob_Result *list,
				   list_HBCI_MsgJob_Result_cb func,
				   void *user_data);
  /*@}*/

#ifdef __cplusplus 
}
#endif /* __cplusplus */

#endif



