/***************************************************************************
  $RCSfile: pluginmanager.cpp,v $
  -------------------
  cvs         : $Id: pluginmanager.cpp,v 1.5 2004/01/20 03:40:14 aquamaniac Exp $
  begin       : Sat Jul 05 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "dllimport.h"
#include "pluginmanager.h"
#include "hbcistring.h"
#include "hbci.h"
#include <gwenhywfar/debug.h>


namespace HBCI {

  PluginManager::PluginManager(Hbci *hbci)
    :_hbci(hbci)
  {
  }


  PluginManager::~PluginManager(){
    _plugins.clear();
    _files.clear();
  }


  string PluginManager::_getLibraryPrefix(const string &path) {
    string prefix;
    unsigned int pp;

    // get basename
    pp=path.rfind("/");
    if (pp==string::npos)
      prefix=path;
    else
      prefix=path.substr(pp+1);
    pp=prefix.rfind(PLUGIN_EXT);
    if (pp!=string::npos)
      prefix=prefix.substr(0,pp);

    // convert to lower
    for (pp=0; pp<prefix.length(); pp++)
      prefix[pp]=tolower(prefix[pp]);

    return prefix;
  }

  string PluginManager::_getTypeSubdir(PluginType t) {
    string s;

    switch(t) {
    case PluginTypeMedium:    s="media"; break;
    case PluginTypeMessage:   s="message"; break;
    case PluginTypeTransport: s="transport"; break;
    case PluginTypeParser:    s="parser"; break;
    default: s="";
    } // switch
    return s;
  }


  Error PluginManager::_loadPlugin(const string &path, PluginType pt) {
    Pointer<PluginFile> f;
    Error err;

    try {
      f=new PluginFile(path, _getLibraryPrefix(path));
    }
    catch(HBCI::Error xerr) {
      return Error("PluginManager::loadPlugin",xerr);
    }

    err=f.ref().registerPlugin(_hbci, this, pt);
    if (!err.isOk())
      return Error("PluginManager::loadPlugin",err);

    _files.push_back(f);
    if (Hbci::debugLevel()>5)
      fprintf(stderr," Loaded plugin \"%s\"\n", path.c_str());
    return Error();
  }


  Pointer<Plugin> PluginManager::findPlugin(const string &name,
					    PluginType pt) {
    list<Pointer<Plugin> >::iterator it;

    for (it=_plugins.begin();
	 it!=_plugins.end();
	 it++)
      if ((*it).ref().pluginType()==pt)
	if (strcasecmp((*it).ref().pluginName().c_str(), name.c_str())==0)
	  return *it;

    return 0;
  }


  Error PluginManager::_tryToLoadPlugin(const list<string> &dnames,
					PluginType pt,
					const string &mname,
					unsigned int wantsoversion,
					unsigned int wantsoage) {
    int soversion;
    list<string>::const_iterator sit;
    Error err;
    string t;
    unsigned int i;
    string mtypedir;

    mtypedir=_getTypeSubdir(pt);
    t=mname;
    for (i=0; i<t.length(); i++)
      t[i]=tolower(t[i]);

    for (soversion=(int)wantsoversion;
         soversion>=(int)(wantsoversion-wantsoage);
         soversion--) {
      if (soversion<0) {
        DBG_INFO(0, "No version left to check. Wanted version %d, age is %d",
                 wantsoversion,
                 wantsoage);
        break;
      }
      for (sit=dnames.begin();
	   sit!=dnames.end();
	   sit++) {
	// try to load the plugin
	string mpath;

	mpath=*sit;
	mpath+="/";
	mpath+=String::num2string(soversion);
	if (!mtypedir.empty()) {
	  mpath+="/";
	  mpath+=mtypedir;
	  mpath+="/";
	}
	mpath+=t;
	mpath+=PLUGIN_EXT;
        if (Hbci::debugLevel()>10)
          fprintf(stderr, "Checking directory \"%s\"\n", mpath.c_str());
	err=_loadPlugin(mpath, pt);
	if (err.isOk()) {
	  return Error();
	}
	else {
	  if (Hbci::debugLevel()>5) {
	    fprintf(stderr,
		    "PluginManager::_tryToLoadPlugin: "
		    "Error loading plugin (%s)\n",
		    err.errorString().c_str());
	  }
	}
      } // for
    } // for

    return Error("PluginManager::_tryToLoadPlugin",
		 ERROR_LEVEL_NORMAL,
		 HBCI_ERROR_CODE_INVALID,
		 ERROR_ADVISE_DONTKNOW,
		 "plugin type not found");
  }


  Pointer<Plugin> PluginManager::ensurePlugin(const string &mname,
					      PluginType pt) throw(Error){
    Pointer<Plugin> tmp;
    string mtypedir;

    mtypedir=_getTypeSubdir(pt);

    tmp=findPlugin(mname, pt);
    if (!tmp.isValid()) {
#ifdef PLUGINLOADING_SUPPORTED
      Error err;
      list<string> dnames;

      dnames.push_back(PLUGIN_PATH);
      dnames.push_back("/usr/share/openhbci2/plugins");
      dnames.push_back("/usr/local/share/openhbci2/plugins");

      err=_tryToLoadPlugin(dnames,
			   pt,
			   mname,
			   OPENHBCI_SO_VERSION,
			   OPENHBCI_SO_AGE);
      if (!err.isOk()) {
	throw Error("PluginManager::ensurePlugin",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_INVALID,
		    ERROR_ADVISE_DONTKNOW,
		    "plugin type not found",
		    mname);
      }

      tmp=findPlugin(mname, pt);
      if (!tmp.isValid()) {
	throw Error("PluginManager::ensurePlugin",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_INVALID,
		    ERROR_ADVISE_ABORT,
		    "plugin type not found after loading it",
		    mname);
      }
#else  // ifdef PLUGINLOADING_SUPPORTED
      throw Error("PluginManager::ensurePlugin",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_INVALID,
		  ERROR_ADVISE_DONTKNOW,
		  "plugin type not found",
		  mname);
#endif // ifdef PLUGINLOADING_SUPPORTED
    }

    return tmp;
  }

  Error PluginManager::loadPlugin(const string &path,
				  PluginType pt,
				  Pointer<PluginFile> &pf,
				  Pointer<Plugin> &pl) {
    Error err;

    try {
      pf=new PluginFile(path, _getLibraryPrefix(path));
    }
    catch(HBCI::Error xerr) {
      return Error("PluginManager::loadPlugin",xerr);
    }

    try {
      pl=pf.ref().createPlugin(_hbci, pt);
    }
    catch (Error xerr) {
      return Error("PluginManager::loadPlugin",err);
    }

    if (Hbci::debugLevel()>5)
      fprintf(stderr," Loaded plugin \"%s\"\n", path.c_str());
    return Error();
  }

    Error PluginManager::registerPlugin(Pointer<Plugin> p) {
      list<Pointer<Plugin> >::iterator it;

      if (Hbci::debugLevel()>5)
	fprintf(stderr, "PluginManager: Registering plugin \"%s\"\n",
		p.ref().pluginName().c_str());
      for (it=_plugins.begin(); it!=_plugins.end(); it++) {
	if ((*it).ref().pluginType()==p.ref().pluginType())
	  if (strcasecmp((*it).ref().pluginName().c_str(),
			 p.ref().pluginName().c_str())==0) {
	    return Error("PluginManager::registerPlugin",
			 ERROR_LEVEL_INTERNAL,
			 HBCI_ERROR_CODE_INVALID,
			 ERROR_ADVISE_ABORT,
			 "plugin already exists",
			 p.ref().pluginName());
	  }
      }
      p.setDescription("Member of PluginManager::_plugins");
      if (Hbci::debugLevel()>5)
	fprintf(stderr, "PluginManager: Registered plugin \"%s\"\n",
		p.ref().pluginName().c_str());
      _plugins.push_back(p);

      return Error();
    }


} /* namespace HBCI */

