/* outboxjob.cpp

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "assert.h"

#include "outbox.h"
#include "outboxjob.h"
#include <gwenhywfar/debug.h>


namespace HBCI {

  OutboxJob::OutboxJob(const string &name, Pointer<Customer> c,
                       const string &account)
    :Job(name, c, account)
    ,_status(HBCI_JOB_STATUS_TODO)
    ,_id(0)
    ,_cust(c)
    ,_result(HBCI_JOB_RESULT_NONE)
{
  _cust.setDescription("OutboxJob::_cust");
  _bank.setDescription("OutboxJob::_bank");
  if (!c.isValid())
    fprintf(stderr,"OutboxJob: invalid customer pointer.\n");
  else {
    _bank=c.ref().user().ref().bank();
    /* preset some properties to reasonable values
     * these values may be overwritten later by the application
     */
    setProperty("accountid", account);
    setIntProperty("country", _bank.ref().country());
    setProperty("bankcode", _bank.ref().bankCode());
  }

}


OutboxJob::~OutboxJob(){
}


void OutboxJob::addSigner(Pointer<Customer> c){
  c.setDescription("Member of OutboxJob::_signers");
  _signers.push_back(c);
}


list<int> OutboxJob::resultCodes() const{
  return list<int>();
}


Error OutboxJob::evaluate(Pointer<Outbox> outbox){
  bool foundOkCode;

  DBG_NOTICE(0, "Evaluating job");
  if (results().empty()) {
    list<MsgQueue::Result>::const_iterator rit;

    DBG_INFO(0, "No results for job \"%s\", checking queue results ...",
             name().c_str());
    setStatus(HBCI_JOB_STATUS_DONE);
    if (outbox.ref().results().empty()) {
      DBG_WARN(0, "Really no results for job \"%s\"", name().c_str());
      setStatus(HBCI_JOB_STATUS_DONE);
      setResult(HBCI_JOB_RESULT_FAILED);
      return Error("OutboxJob::evaluate",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "no results");
    }

    // check every queue result
    foundOkCode=false;
    for (rit=outbox.ref().results().begin();
         rit!=outbox.ref().results().end();
         rit++) {
      if ((*rit).resultCode()>=9000) {
        DBG_INFO(0,
                 "Found an error code in result for \"%s\" (%d)",
                 name().c_str(),
                 (*rit).resultCode());
        setStatus(HBCI_JOB_STATUS_DONE);
        setResult(HBCI_JOB_RESULT_FAILED);
      }
      else if ((*rit).resultCode()>=3000 &&
               (*rit).resultCode()<4000) {
        DBG_INFO(0, "Found a warning for job \"%s\" (%d)",
                 name().c_str(), (*rit).resultCode());
        foundOkCode=true;
      }
      else if ((*rit).resultCode()<1000){
	DBG_NOTICE(0, "Found success indicator for job \"%s\" (%d)",
		   name().c_str(), (*rit).resultCode());
        foundOkCode=true;
      }
      else {
	DBG_NOTICE(0, "Found unknown (but nice) code for job \"%s\" (%d)",
		   name().c_str(), (*rit).resultCode());
        foundOkCode=true;
      }
    } // for
  } // if no job results
  else {
    list<MsgJob::Result>::const_iterator rit;

    // check every result
    foundOkCode=false;
    for (rit=results().begin();
         rit!=results().end();
         rit++) {
      DBG_NOTICE(0, "Checking result for job %s", name().c_str());
      if (attachable()) {
        DBG_INFO(0, "Attachable job");
        if ((*rit).resultCode()==3040) {
          // found an attachpoint
	  if ((*rit).param().empty()) {
            DBG_ERROR(0,
                      "Result code suggests attachpoint, but there's none");
            setStatus(HBCI_JOB_STATUS_DONE);
            setResult(HBCI_JOB_RESULT_FAILED);
          }
          else {
	    string oldAttach;

	    oldAttach=getProperty("request/attach", "");
	    if (!oldAttach.empty() && oldAttach==(*rit).param()) {
	      DBG_WARN(0,
		       "Server is sending the same attach "
		       "point again, ignoring to prevent endless loop.");
	      addResult(3041, "Same attach point, ignored");
	      setStatus(HBCI_JOB_STATUS_DONE);
	      setResult(HBCI_JOB_RESULT_SUCCESS);
	    }
	    else {
	      DBG_NOTICE(0, "Setting attach point");
	      setProperty("attach", (*rit).param());
	    }
	    if (1) {
	      list<MsgJob::Result>::const_iterator lrit;

	      DBG_NOTICE(0, "Results before removing 3040:");
	      for (lrit=results().begin();
		   lrit!=results().end();
		   lrit++) {
		DBG_NOTICE(0, "Result code: %d", (*lrit).resultCode());
		DBG_NOTICE(0, "Parameter: \"%s\"", (*lrit).param().c_str());
	      } // for
	    }
	    removeResults(3040);
	    if (1) {
	      list<MsgJob::Result>::const_iterator lrit;

	      DBG_NOTICE(0, "Results after removing 3040:");
	      for (lrit=results().begin();
		   lrit!=results().end();
		   lrit++) {
		DBG_NOTICE(0, "Result code: %d", (*lrit).resultCode());
		DBG_NOTICE(0, "Parameter: \"%s\"", (*lrit).param().c_str());
	      } // for
	    }
	    return Error();
	  }
	} // if code 3040
      } // if attachable
      if ((*rit).resultCode()>=9000) {
        DBG_INFO(0,
                 "Found an error code in result for \"%s\" (%d)",
                 name().c_str(),
                 (*rit).resultCode());
        setStatus(HBCI_JOB_STATUS_DONE);
        setResult(HBCI_JOB_RESULT_FAILED);
      }
      else if ((*rit).resultCode()>=3000 &&
               (*rit).resultCode()<4000) {
        DBG_INFO(0, "Found a warning for job \"%s\" (%d)",
                 name().c_str(), (*rit).resultCode());
        foundOkCode=true;
      }
      else if ((*rit).resultCode()<1000){
        DBG_INFO(0, "Found success indicator for job \"%s\" (%d)",
                 name().c_str(), (*rit).resultCode());
        foundOkCode=true;
      }
      else {
        DBG_INFO(0, "Found unknown (but nice) code for job \"%s\" (%d)",
                 name().c_str(), (*rit).resultCode());
        foundOkCode=true;
      }
    } // for
  } // if job results
  // if we also found a good result flag it...
  if (result()==HBCI_JOB_RESULT_FAILED) {
    if (foundOkCode) {
      DBG_INFO(0, "Job \"%s\" succeeded partly", name().c_str());
      setResult(HBCI_JOB_RESULT_PARTLY_SUCCESS);
      return Error();
    }
    return Error("OutboxJob::evaluate",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "Job failed");
  }

  // mark job as done and ok
  setStatus(HBCI_JOB_STATUS_DONE);
  setResult(HBCI_JOB_RESULT_SUCCESS);
  return Error();
}





} // namespace  HBCI

HBCI_OutboxJob *HBCI_OutboxJob_new(const char *name, 
				   HBCI_Customer *c,
				   const char *account)
{
    assert(c);
    HBCI::Pointer<HBCI::Customer> cp = const_cast<HBCI::Customer*>(c);
    cp.setAutoDelete(false);

    try {
	
	return new HBCI::OutboxJob(name ? string(name) : "", 
				   cp, 
				   account ? string(account) : "");
    }
    catch(HBCI_Error err) {
	fprintf(stderr,"HBCI_OutboxJob_new: Caught error: %s\n",
		err.errorString().c_str());
	return NULL;
    }
    catch (...) {
	fprintf(stderr, "HBCI_OutboxJob_new: Caught unknown exception.\n");
	return NULL;
    }
}
void HBCI_OutboxJob_delete(HBCI_OutboxJob *j)
{
    delete j;
}
HBCI_Job *HBCI_OutboxJob_Job(HBCI_OutboxJob *j) {return j;}
const HBCI_Job *HBCI_OutboxJob_Job_const(const HBCI_OutboxJob *j)  {return j;}

void HBCI_OutboxJob_addSigner(HBCI_OutboxJob *j, const HBCI_Customer *c)
{
    assert(j);
    assert(c);
    HBCI::Pointer<HBCI::Customer> cp = const_cast<HBCI::Customer*>(c);
    cp.setAutoDelete(false);
    j->addSigner(cp);
}
OutboxJob_Status HBCI_OutboxJob_status(const HBCI_OutboxJob *j)
{
    assert(j);
    return j->status();
}
OutboxJob_Result HBCI_OutboxJob_result(const HBCI_OutboxJob *j)
{
    assert(j);
    return j->result();
}
int HBCI_OutboxJob_id(const HBCI_OutboxJob *j)
{
    assert(j);
    return j->id();
}
list_int *HBCI_OutboxJob_resultCodes(const HBCI_OutboxJob *j)
{
    assert(j);
    return new list_int(j->resultCodes());
}





