/* outboxjob.h
 
 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef OUTBOXJOB_H
#define OUTBOXJOB_H

/** @file outboxjob.h
 *
 * @short Abstract base class HBCI::OutboxJob and its C wrapper @ref
 * HBCI_OutboxJob functions. */

#ifdef __cplusplus
#include <string>
#include <list>
#endif /* __cplusplus */

#include <openhbci2/dllimport.h>
#include <openhbci2/pointer.h>
#include <openhbci2/customer.h>
#include <openhbci2/user.h>
#include <openhbci2/bank.h>
#include <openhbci2/job.h>
#include <openhbci2/listwrappers.h>
#include <openhbci2/msgjob.h>


/** Status of an OutboxJob. */
typedef enum OutboxJob_Status {
    /** No status, i.e. for freshly created jobs */
    HBCI_JOB_STATUS_NONE = 0,
    /** Job is to be executed */
    HBCI_JOB_STATUS_TODO,
    /** Job was executed */
    HBCI_JOB_STATUS_DONE
} OutboxJob_Status;
/** Result of an OutboxJob. */
typedef enum OutboxJob_Result {
    /** Job was not executed */
    HBCI_JOB_RESULT_NONE = 0,
    /** Job was successfully performed */
    HBCI_JOB_RESULT_SUCCESS,
    /** Job was partly successfully performed */
    HBCI_JOB_RESULT_PARTLY_SUCCESS,
    /** Job was executed and failed */
    HBCI_JOB_RESULT_FAILED
} OutboxJob_Result;

#ifdef __cplusplus
#include <openhbci2/msgqueue.h>
/* Forward declaration */
namespace HBCI {
  class OutboxJob;
};
/** @ingroup HBCI_OutboxJobg
 * @short C accessible type of HBCI::OutboxJob */
typedef HBCI::OutboxJob HBCI_OutboxJob;

namespace HBCI {

/**
 * @short Abstract base class for all jobs to be used with HBCIAPI
 * 
 * This is an abstract base class for the jobs of the OpenHBCI-Wrapper
 * API. Every job to be used in the @ref Outbox is derived from this. 
 *
 * @author Martin Preuss<martin@libchipcard.de>
 */
  class DLLIMPORT OutboxJob: public Job {
    friend class API;
    friend class Outbox;
    friend class customerQueue;
  private:
    OutboxJob_Status _status;
    int _id;
    list<Pointer<Customer> > _signers;
  protected:
    Pointer<Customer> _cust;
    Pointer<Bank> _bank;
    OutboxJob_Result _result;

  public:
    /** @name Constructors And Destructor
     *
     */
    /*@{*/
    /**
     *  Default Constructor.
     */
    OutboxJob(const string &name, Pointer<Customer> c,
              const string &account="");
    /**
     *  Default Destructor.
     */
    virtual ~OutboxJob();
    /*@}*/


    /** @name Manage Customer 
     *
     * Methods to manage the Customer's involved in this OutboxJob. */
    /*@{*/
    /**
     *  Returns a pointer to the Customer owning this job.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    Pointer<Customer> customer() const { return _cust;};

    /**
     *  Adds a signer.
     *
     * Here you can add a signer. In most cases you don't need this
     * method, since the first signer is already given in the
     * constructor. But you need this method if this job needs
     * multiple signers for your bank.
     *
     * @author Martin Preuss<martin@libchipcard.de> */
    void addSigner(Pointer<Customer> c);
    /*@}*/


    /** @name Status Information 
     *
     * Methods to retrieve information about this OutboxJob's status
     * and result. */
    /*@{*/
    /**
     *  Returns the status of this job.
     *
     * Returns the status of this job. It can be one of the following:
     * <ul>
     * <li>@ref HBCI_JOB_STATUS_NONE (no status, e.g. for freshly created jobs)
     * </li>
     * <li>@ref HBCI_JOB_STATUS_TODO (job is to be executed)</li>
     * <li>@ref HBCI_JOB_STATUS_DONE (job was executed) </li>
     * </ul>
     * @author Martin Preuss<martin@libchipcard.de>
     */
    OutboxJob_Status status() const { return _status;};

    /**
     *  Return the result code of the job. 
     *
     * Return the result code of the job. This of course is only usefull
     * AFTER the job has been performed ;-)
     * The result can be one of these:
     * <ul>
     * <li>@ref HBCI_JOB_RESULT_NONE (job was not executed)</li>
     * <li>@ref HBCI_JOB_RESULT_SUCCESS (job was successfully performed) </li>
     * <li>@ref HBCI_JOB_RESULT_PARTLY_SUCCESS
     *     (job was partly successfull)</li>
     * <li>@ref HBCI_JOB_RESULT_FAILED (job was executed and failed) </li>
     * </ul>
     * @author Martin Preuss<martin@libchipcard.de>
     */
    OutboxJob_Result result() const { return _result;};

    /**
     *  Return the id of this job. 
     *
     * Return the id of this job. Each enqueued job has a unique
     * id. This can be used from an application to identify specific
     * jobs after they have been added to the outbox queue.
     *
     * @author Martin Preuss<martin@libchipcard.de> */
    int id() const { return _id;};

    /**  Returns a list of result codes of this job. 
     *
     * This can be used to get more detailed information if the result
     * of this job was HBCI_JOB_RESULT_FAILED. In that case, you can
     * traverse this list, checking for resultcodes >= 9000. Each of
     * those can be the source of the job failure, so you can react
     * accordingly. %HBCI spec suggests that the most important of
     * them can be found at the front() already, but that depends on
     * your bank. 
     *
     * Note: A resultcode >= 9000 does not automatically mean an
     * error, see MessageQueue::getResult().
     *
     * @see MessageQueue::getResult(), Job::errorcodeIsLibraryBug() */
    virtual list<int> resultCodes() const;

    /*@}*/

  protected:

    /** @name Manipulation from HBCIAPI
     *
     * Methods to be called only from HBCIAPI or Outbox. These methods
     * manipulate this OutboxJob.  */
    /*@{*/
    /**
     *  Set id.
     *
     * This is called by the Outbox to assign a unique id to all
     * enqueued jobs.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    void setId(int i) { _id=i;};

    /**  Set status of this job. */
    void setStatus(OutboxJob_Status s) { _status=s;};

    /**  Set result of this job. */
    void setResult(OutboxJob_Result r) { _result=r;};

    /**
     *  Let the job check the result and set the flags. 
     *
     * Let the job check the result. This means checking the results
     * of all Jobs involved (which otherwise is not done
     * automatically). This method sets its status and result
     * accordingly -- otherwise the status and result flags might not
     * be up-to-date.
     *
     * @author Martin Preuss<martin@libchipcard.de>
     * @return error object
     */
    virtual Error evaluate(Pointer<Outbox> outbox);
    /*@}*/
};

} /* namespace HBCI */

extern "C" {
#else /* __cplusplus */
    typedef struct HBCI_OutboxJob HBCI_OutboxJob;
#endif /* __cplusplus */
    /** @defgroup HBCI_OutboxJobg HBCI_OutboxJob and derivations */
    /*@{*/
    /** @name HBCI_OutboxJob */
    /*@{*/
    /**
     *  Default Constructor.
     *
     * @param name The name of this job. You can use the tool @c
     * hbcixml to find out about the name for your desired job.
     *
     * @param c The Customer that shall own this job.
     * @param account The account id this job works on. This may be left empty ("").
     */
    extern HBCI_OutboxJob *HBCI_OutboxJob_new(const char *name, 
					      HBCI_Customer *c,
					      const char *account);
    /** Destructor */
    extern void HBCI_OutboxJob_delete(HBCI_OutboxJob *j);
    /** Upcast. */
    extern HBCI_Job *HBCI_OutboxJob_Job(HBCI_OutboxJob *j);
    /** Const upcast. */
    extern const HBCI_Job *HBCI_OutboxJob_Job_const(const HBCI_OutboxJob *j);

    /**
     *  Adds a signer.
     *
     * Here you can add a signer. In most cases you don't need this
     * method, since the first signer is already given in the
     * constructor. But you need this method if this job needs
     * multiple signers for your bank.
     */
    extern void HBCI_OutboxJob_addSigner(HBCI_OutboxJob *j, 
					 const HBCI_Customer *c);
    /**
     *  Returns the status of this job.
     */
    extern OutboxJob_Status HBCI_OutboxJob_status(const HBCI_OutboxJob *j);

    /** Returns the result of this job. */
    extern OutboxJob_Result HBCI_OutboxJob_result(const HBCI_OutboxJob *j);

    /**
     *  Return the id of this job. 
     *
     * Return the id of this job. Each enqueued job has a unique
     * id. This can be used from an application to identify specific
     * jobs after they have been added to the outbox queue.
     */
    extern int HBCI_OutboxJob_id(const HBCI_OutboxJob *j);

    /**  Returns a list of result codes of this job. 
     *
     * This can be used to get more detailed information if the result
     * of this job was HBCI_JOB_RESULT_FAILED. In that case, you can
     * traverse this list, checking for resultcodes >= 9000. Each of
     * those can be the source of the job failure, so you can react
     * accordingly. %HBCI spec suggests that the most important of
     * them can be found at the front() already, but that depends on
     * your bank. 
     *
     * Note: A resultcode >= 9000 does not automatically mean an
     * error, see MessageQueue::getResult().
     *
     * @see MessageQueue::getResult(), Job::errorcodeIsLibraryBug() */
    extern list_int *HBCI_OutboxJob_resultCodes(const HBCI_OutboxJob *j);

    /*@}*/
    /*@}*/

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif


