/***************************************************************************
 $RCSfile: transaction.h,v $
                             -------------------
    cvs         : $Id: transaction.h,v 1.8 2004/01/21 16:46:43 cstim Exp $
    begin       : Sat Jun 08 2002
    copyright   : (C) 2002 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HBCITRANSACTION_H
#define HBCITRANSACTION_H

/** @file transaction.h
 *
 * @short HBCI::Transaction and its C wrapper @ref HBCI_Transaction. */

#include <openhbci2/dllimport.h>
#include <gwenhywfar/db.h>
#include <openhbci2/date.h>
#include <openhbci2/value.h>
#include <openhbci2/hbcistringlist.h>

#ifdef __cplusplus
#include <string>
#include <list>

namespace HBCI {

/**
 * @short Convenience wrapper for all the data of a transaction in %HBCI.
 *
 * This class represents a transaction. However, in OpenHBCI-TNG the
 * data of one transaction is hold in a GWEN_DB_NODE (@ref
 * GWEN_DB_NODE). This class merely represents a wrapper class to
 * access the data in the GWEN_DB_NODE.
 *
 * In %HBCI, transactions appear in three places, and unfortunately
 * all cases use a different data format: Firstly, in a
 * JobGetTransactions, where the data format is specified by SWIFT
 * MT940. Secondly, in a JobSingleTransfer, where the data format is
 * defined in %HBCI and is almost similar to DTAUS0. Thirdly, in the
 * DTAUS-data of a JobMultiTransfer, where the data format is
 * specified by DTAUS.
 *
 * This wrapper class intends to support only those data members that
 * are actually supported and/or required in all three cases. In other
 * words, each of the abovementioned data formats still has additional
 * data fields. If the application needs those, it can always access
 * the GWEN_DB_NODE directly.
 *
 * Previous documentation said here: It was intended to document each
 * member of this class with a reference to the corresponding fields
 * in those two specifications. FIXME: Unfortunately so far only the
 * references to the SWIFT MT940 fields have been added, but DTAUS0
 * comments will be added soon.
 *
 * The data getters/setters are ordered in the order in which they
 * appear in the SWIFT MT940 Specification (%HBCI Spec. chapter
 * IX.2.8). That specification is therefore the authoritative source
 * of what precisely each data element means. (Unfortunately sometimes
 * the spec is not so precise...)
 *
 * Note that a whole lot of important information wasn't specified in
 * the MT940 itself but only as an additional structure of the
 * "misc. field" Number 86. Those data fields are specified by the ZKA
 * but not SWIFT itself, and therefore it is noted where that field
 * No. 86 comes into play.
 *
 * @author Christian Stimming <stimming@tuhh.de> */
class DLLIMPORT Transaction {
 private:
  GWEN_DB_NODE *_n;

  // These are all populated on-the-fly, but their existence makes C
  // wrappers immensly easier.
  mutable Value tmp_value;
  mutable Date tmp_date;
  mutable Date tmp_valutadate;
  mutable list<string> tmp_purpose;
  mutable list<string> tmp_othername;
  mutable Value tmp_origValue;
  mutable Value tmp_charges;
  
  const char *get(const char *key) const
    {
      return GWEN_DB_GetCharValue(const_cast<GWEN_DB_NODE*>(_n), key, 0, "");
    }
  const int getInt(const char *key) const
    {
      return GWEN_DB_GetIntValue(const_cast<GWEN_DB_NODE*>(_n), key, 0, -1);
    }
  void set(const char *key, const char *value)
    { 
      GWEN_DB_SetCharValue(_n, GWEN_DB_FLAGS_OVERWRITE_VARS,
			   key, value);
    }
  void set(const char *key, int value)
    { 
      GWEN_DB_SetIntValue(_n, GWEN_DB_FLAGS_OVERWRITE_VARS,
			   key, value);
    }
      
 public:

  /** Constructor for this wrapper class around the GWEN_DB for a
   * %HBCI transaction.
   *
   * @param n The GWEN_DB_NODE that is used as underlying storage
   * facility for this data, and which will be returned by
   * node(). This class will @e not take over ownership of this node,
   * so it will continue to be owned by the caller.
   */
  Transaction(GWEN_DB_NODE *n);

  /** Destructor. The underlying GWEN_DB_NODE will be untouched by the
      destructor. */
  ~Transaction();

  /** Returns the underlying GWEN_DB_NODE that was used for storing
      the data. This is the same object that was specified in the
      constructor and this class did not take over ownership of the
      node. Therefore the caller can use this returned value for
      whatever he is going to do. */
  GWEN_DB_NODE *node() 
      {
	  return _n;
      }
  

  /** @name Our account */
  /*@{*/
  /* country code of your institute -- not in MT940 */
  /*int ourCountryCode() const { return 0; };*/
  /* FIXME - where is country code? */
  /* country code of your institute -- not in MT940 */
  /*void setOurCountryCode(int i) { };*/

  /** bank code of your institute (BLZ or S.W.I.F.T. code) -- BLZ in
   :25: */
  const char *ourBankCode() const 
    {
      return get("ourAccount/BankCode");
    }
  
  /** bank code of your institute (BLZ or S.W.I.F.T. code) -- BLZ in
   :25: */
  void setOurBankCode(const char *code)
    {
      set("ourAccount/BankCode", code); 
    }

  /** Your account number; up to 23 characters. Note: possibly
   * including the currency (yes, that's what MT940 says) --
   * Kontonummer in :25: */
  const char *ourAccountId() const
    { return get("ourAccount/AccountId"); }

  /** Your account number; up to 23 characters. Note: possibly
   * including the currency (yes, that's what MT940 says) --
   * Kontonummer in :25: */
  void setOurAccountId(const char *id)
    { set("ourAccount/AccountId", id); }

  /** If your account has subaccounts which are distinguished by
   * different suffixes, then this is that suffix. Otherwise it's
   * empty. -- not in MT940 */
  //const char *ourSuffix() const { return ""; /* FIXME: no suffix yet */ };
  /** If your account has subaccounts which are distinguished by
   * different suffixes, then this is that suffix. Otherwise it's
   * empty. -- not in MT940 */
  //void setOurSuffix(const char *id) { };
  /*@}*/

  /** @name Date */
  /*@{*/
  /** The date "Valuta" when the transaction was really executed --
   * Datum Valuta/Wertstellung in :61: */
  const Date& valutaDate() const;
  /** The date "Valuta" when the transaction was really executed --
   * Datum Valuta/Wertstellung in :61: */
  void setValutaDate(const Date& d)
    { set("valutaDate", d.toString().c_str()); }

  /** The date when the transaction was booked (but usually it is
   * unused). In MT940 spec this has only four characters (no year),
   * so we take the year from the valutaDate(). -- Buchungsdatum in
   * :61: */
  const Date& date() const;
  /** The date when the transaction was booked (but usually it is
   * unused). In MT940 spec this has only four characters (no year),
   * so we take the year from the valutaDate(). -- Buchungsdatum in
   * :61: */
  void setDate(const Date& d) { set("date", d.toString().c_str());};
  /*@}*/

  /** @name Value */
  /*@{*/
  /** The value of this transaction in the account's currency. --
   * Betrag in :61: */
  const Value& value() const;
  /** The value of this transaction in the account's currency. --
   * Betrag in :61:
   * @return The numeric value (without the currency part)
   */
  const char *value_value() const { return get("value/value");};
  /** The value of this transaction in the account's currency. --
   * Betrag in :61:
   * @return The currency of the value (without the numeric part)
   */
  const char *value_currency() const 
    { return get("value/currency"); }
  
  /** The value of this transaction in the account's currency. --
   * Betrag in :61: */
  void setValue(const Value& v);
  /** The value of this transaction in the account's currency. --
   * Betrag in :61:
   * @param v The numeric value (without the currency part)
   */
  void setValue_value(const char *v) { set("value/value", v); };
  /** The value of this transaction in the account's currency. --
   * Betrag in :61:
   * @param c The currency of the value (without the numeric part)
   */
  void setValue_currency(const char *c) 
    {
      set("value/currency", c);
    }
  /*@}*/

  /** @name Type of Transaction */
  /*@{*/
  /** this is the transaction id that tells you more about the type of
   * transaction (3 character code) -- Buchungsschluessel in
   * :61:. Only in SWIFT! */
  //const char *transactionKey() const { return get("key");};
  /** this is the transaction id that tells you more about the type of
   * transaction (3 character code) -- Buchungsschluessel in
   * :61:. Only in SWIFT! */
  //void setTransactionKey(const char *s) { set("key", s); };
  /*@}*/

  /** @name Reference String */
  /*@{*/
  /** Reference string, if the customer (you) has specified
   * one. (E.g. the cheque number.) Otherwise "NONREF". --
   * Kundenreferenz in :61: */
  const char *customerReference() const { return get("custref"); };
  /** Reference string, if the customer (you) has specified
   * one. (E.g. the cheque number.) Otherwise "NONREF". --
   * Kundenreferenz in :61: */
  void setCustomerReference(const char *s) { set("custref", s);};

  /** Reference string for this transaction given by the bank, if it
   * has given one. Otherwise empty. -- Bankreferenz in :61: */
  const char *bankReference() const { return get("bankref"); };
  /** Reference string for this transaction given by the bank, if it
   * has given one. Otherwise empty. -- Bankreferenz in :61: */
  void setBankReference(const char *s) { set("bankref", s);};
  /*@}*/

  /** @name Currency exchange */
  /*@{*/
  /** If this transaction was in a different currency than your
   * account is in, then this is the transaction value in the original
   * currency.  -- Ursprungsbetrag in :61: */
  const Value &originalValue() const;
  const char *originalValue_value() const { return get("origvalue/value"); };
  const char *originalValue_currency() const
    { return get("origvalue/currency"); };
  /** NOT IMPLEMENTED. If this transaction was in a different
   * currency than your account is in, then this is the transaction
   * value in the original currency. Otherwise zero. Currently NOT
   * IMPLEMENTED in the MT940 parser. -- Ursprungsbetrag in :61: */
  //void setOriginalValue(const Value &v) { _originalValue=v;};

  /** If this transaction was in a different currency than your
   * account is in, then this is the extra amount charged for currency
   * exchange.  */
  const Value &charge() const;
  const char *charge_currency() const 
    { return get("charges/currency"); };
  const char *charge_value() const
    { return get("charges/value"); };
  /** NOT IMPLEMENTED. If this transaction was in a different
   * currency than your account is in, then this is the extra amount
   * charged for currency exchange. Currently NOT IMPLEMENTED in the
   * MT940 parser. -- Gebhrenbetrag (charges) in :61: */
  //void setCharge(const Value &v) { _charge=v;};
  /*@}*/

  /* Now follows the obnoxious "Mehrzweckfeld :86:" */

  /** @name Transaction Description */
  /*@{*/
  /** A 3 digit numerical transaction code, defined for all kinds of
   * different HBCI actions ("textkey"). -- Geschaeftsvorfallcode,
   * first field in field :86:
   *
   * For a normal transfer you should set it to 51. For debit notes
   * the values 04 or 05 (Watch out: 06 was WRONG!) may be used. For
   * other values please refer to your credit institute.
   */
  int transactionCode() const { return getInt("textkey");};
  /** A 3 digit numerical transaction code, defined for all kinds of
   * different HBCI actions. -- Geschaeftsvorfallcode, first field
   * in field :86:
   *
   * For a normal transfer you should set it to 51. For debit notes
   * the values 04 or 05 (Watch out: 06 was WRONG!) may be used. For
   * other values please refer to your credit institute.
   */
  void setTransactionCode(int c) { set("textkey", c);};

  /** Transaction text (e.g. STANDING ORDER) -- Buchungstext; field
   * 00 in field :86:. Only in SWIFT! */
  //const char *transactionText() const { return get("text"); };
  /** Transaction text (e.g. STANDING ORDER) -- Buchungstext; field
   * 00 in field :86:. Only in SWIFT! */
  //void setTransactionText(const char *s) { set("text", s);};

  /** Primanoten-Number (no idea what this means) --
   * Primanoten-Nummer; field 10 in field :86:  */
  //const char *primanota() const { return get("primanota");};
  /** Primanoten-Number (no idea what this means) --
   * Primanoten-Nummer; field 10 in field :86:  */
  //void setPrimanota(const char *s) { set("primanota", s);};

  /** Description of what this transaction is for, and how the Bank
   * describes this transaction.
   *
   * According to HBCI Spec chapter IX.2.8, somewhere in these lines
   * this might contain the equivalent transaction value in DEM
   * vs. EUR or vice versa. Also, if the originalValue() and
   * charge() above was unused but this in fact had a original value
   * and charges, then that information might be contained somewhere
   * in these description lines. (Good luck.)
   *
   * -- Verwendungszweck; fields 20-29 and 60-63 in field :86:  */
  const list<string>& description() const;

  /** Description of what this transaction is for, and how the Bank
   * describes this transaction.
   *
   * Note: Every single string represents one line, and each line is
   * going to be truncated to have at most 27 characters.
   *
   * According to HBCI Spec chapter IX.2.8, somewhere in these lines
   * this might contain the equivalent transaction value in DEM
   * vs. EUR or vice versa. Also, if the originalValue() and
   * charge() above was unused but this in fact had a original value
   * and charges, then that information might be contained somewhere
   * in these description lines. (Good luck.)
   *
   * -- Verwendungszweck; fields 20-29 and 60-63 in field :86:  */
  void addDescription(const char *s) { 
    GWEN_DB_SetCharValue(_n, GWEN_DB_FLAGS_DEFAULT,
			 "purpose", s);
  };
  /*@}*/

  /** @name Other account */
  /*@{*/
  /** Country code (BLZ) of the other party's institute -- not in
   * MT940/field :86: */
  //int otherCountryCode() const { return _otherCountryCode;};
  /** Country code (BLZ) of the other party's institute -- not in
   * MT940/field :86: */
  //void setOtherCountryCode(int i) { _otherCountryCode=i;};

  /** Institute code (BLZ) of the other party's institute -- BLZ;
   * field 30 in field :86: */
  const char *otherBankCode() const { return get("otherAccount/BankCode");};
  /** Institute code (BLZ) of the other party's institute -- BLZ;
   * field 30 in field :86: */
  void setOtherBankCode(const char *s) { set("otherAccount/BankCode", s); };

  /** Account number of the other party -- Kontonummer; field 31 in
   * field :86: */
  const char *otherAccountId() const { return get("otherAccount/AccountId");};
  /** Account number of the other party -- Kontonummer; field 31 in
   * field :86: */
  void setOtherAccountId(const char *s) { set("otherAccount/AccountId", s);};

  /** If the account of the other party has subaccounts which are
   * distinguished by different suffixes, then this is that suffix
   * of the other party. Otherwise it's empty. -- *not* in MT940 /
   * field :86: */
  //const char *otherSuffix() const { return _otherSuffix;};
  /** If the account of the other party has subaccounts which are
   * distinguished by different suffixes, then this is that suffix
   * of the other party. Otherwise it's empty. -- *not* in MT940 /
   * field :86: */
  //void setOtherSuffix(const char *id) { _otherSuffix=id;};

  /** Name of the other party -- fields 32 and 33 in field :86: */
  const list<string>& otherName() const;
  
  /** Name of the other party -- fields 32 and 33 in field :86:
   * Each single string (there are two strings allowed) is going
   * to be truncated to have at most 27 characters. */
  void addOtherName(const char *s) { 
    GWEN_DB_SetCharValue(_n, GWEN_DB_FLAGS_DEFAULT,
			 "othername", s);
  };
  /*@}*/

  /**
   * Compares this Transaction to the given one.
   * @author Martin Preuss<openhbci@aquamaniac.de>
   */
  //bool operator==(const Transaction &tda) const;

};

} /* namespace HBCI */

/** @ingroup HBCI_Transactiong
 * @short C accessibly type of HBCI::Transaction */
typedef struct HBCI::Transaction HBCI_Transaction;
/* for C wrapper functions */

extern "C" {    
#else /* __cplusplus */
typedef struct HBCI_Transaction HBCI_Transaction;
#endif /* __cplusplus */

/*     typedef struct _list_HBCI_Transaction list_HBCI_Transaction; */

    /** @defgroup HBCI_Transactiong HBCI_Transaction functions */
    /*@{*/
    /** Constructor for this wrapper class around the GWEN_DB for a
     * %HBCI transaction.
     *
     * @param n The GWEN_DB_NODE that is used as underlying storage
     * facility for this data, and which will be returned by
     * node(). This class will @e not take over ownership of this node,
     * so it will continue to be owned by the caller.
     */
    extern HBCI_Transaction * HBCI_Transaction_new(GWEN_DB_NODE *n);
    /** Destructor. The underlying GWEN_DB_NODE will be untouched by the
	destructor. */
    extern void HBCI_Transaction_delete(HBCI_Transaction *t);
    /** Returns the underlying GWEN_DB_NODE that was used for storing
    the data. This is the same object that was specified in the
    constructor and this class did not take over ownership of the
    node. Therefore the caller can use this returned value for
    whatever he is going to do. */
    extern GWEN_DB_NODE *HBCI_Transaction_node(HBCI_Transaction *t);

    /** @name Our account */
    /*@{*/
    /** country code of your institute -- not in MT940. */
    /*extern int HBCI_Transaction_ourCountryCode(const HBCI_Transaction *t);
    extern void 
    HBCI_Transaction_setOurCountryCode(HBCI_Transaction *t, int i);*/

    /** bank code of your institute (BLZ or S.W.I.F.T. code) -- BLZ in
      :25:
      *
      * The const char* pointer of this and all the following
      * functions points to the c_str() result of the underlying
      * std::string object. It is therefore valid as long as the
      * Transaction object wasn't changed, and it should NOT be
      * free'd by the caller.  */
    extern const char * 
    HBCI_Transaction_ourBankCode(const HBCI_Transaction *t);
    /** bank code of your institute (BLZ or S.W.I.F.T. code) -- BLZ in
	:25: */
    extern void 
    HBCI_Transaction_setOurBankCode(HBCI_Transaction *t, const char *code);

    /** Your account number; up to 23 characters. Note: possibly
     * including the currency (yes, that's what MT940 says) --
     * Kontonummer in :25: */
    extern const char * 
    HBCI_Transaction_ourAccountId(const HBCI_Transaction *t);
    /** Your account number; up to 23 characters. Note: possibly
     * including the currency (yes, that's what MT940 says) --
     * Kontonummer in :25: */
    extern void 
    HBCI_Transaction_setOurAccountId(HBCI_Transaction *t, const char *id);

    /** If your account has subaccounts which are distinguished by
     * different suffixes, then this is that suffix. Otherwise it's
     * empty. -- not in MT940 */
    /*extern const char * 
      HBCI_Transaction_ourSuffix(const HBCI_Transaction *t);
      extern void 
      HBCI_Transaction_setOurSuffix(HBCI_Transaction *t, const char *id);*/
    /*@}*/
    
    /** @name Date */
    /*@{*/
    /** The date "Valuta" when the transaction was really executed --
     * Datum Valuta/Wertstellung in :61: */
    extern const HBCI_Date* 
    HBCI_Transaction_valutaDate(const HBCI_Transaction *t);
    /** The date "Valuta" when the transaction was really executed --
     * Datum Valuta/Wertstellung in :61: 
     *
     * @param t The object to perform this action on.
     * @param d the date. a deep copy of its data will be stored, so the
     * pointer lifecycle doesn't affect this transaction. */
    extern void 
    HBCI_Transaction_setValutaDate(HBCI_Transaction *t, const HBCI_Date *d);

    /** The date when the transaction was booked (but usually it is
     * unused). In MT940 spec this has only four characters (no year),
     * so we take the year from the valutaDate(). -- Buchungsdatum in
     * :61: */
    extern const HBCI_Date* 
    HBCI_Transaction_date(const HBCI_Transaction *t);
    /** 
     * @param t The object to perform this action on.
     * @param d The date. A deep copy of its data will be stored, so the
     * pointer lifecycle doesn't affect this transaction. */
    extern void 
    HBCI_Transaction_setDate(HBCI_Transaction *t, const HBCI_Date *d);
    /*@}*/

    /** @name Value */
    /*@{*/
    /** The value of this transaction in the account's currency. --
    Betrag in :61: */
    extern const HBCI_Value* 
    HBCI_Transaction_value(const HBCI_Transaction *t);

    /** The value of this transaction in the account's currency. --
    Betrag in :61: 
     * @param t The object to perform this action on.
     * @param v the value. a deep copy of its data will be stored, so the
     * pointer lifecycle doesn't affect this transaction. */
    extern void 
    HBCI_Transaction_setValue(HBCI_Transaction *t, const HBCI_Value *v);
    /*@}*/

    /** @name Type of Transaction */
    /*@{*/
    /** this is the transaction id that tells you more about the type
     * of transaction (3 character code) -- Buchungsschluessel in :61: */
    /*extern const char * 
      HBCI_Transaction_transactionKey(const HBCI_Transaction *t);*/
    /** this is the transaction id that tells you more about the type
     * of transaction (3 character code) -- Buchungsschluessel in :61: */
    /*extern void 
      HBCI_Transaction_setTransactionKey(HBCI_Transaction *t, const char *s);*/
    /*@}*/

    /** @name Reference String */
    /*@{*/
    /* Reference char *, if the customer (you) has specified
     * one. (E.g. the cheque number.) Otherwise "NONREF". --
     * Kundenreferenz in :61: */
    extern const char * 
    HBCI_Transaction_customerReference(const HBCI_Transaction *t);
    /* Reference char *, if the customer (you) has specified
     * one. (E.g. the cheque number.) Otherwise "NONREF". --
     * Kundenreferenz in :61: */
    extern void 
    HBCI_Transaction_setCustomerReference(HBCI_Transaction *t, const char *s);
 
    /* Reference string for this transaction given by the bank, if it
     * has given one. Otherwise empty. -- Bankreferenz in :61: */
    extern const char * 
    HBCI_Transaction_bankReference(const HBCI_Transaction *t);
    /* Reference string for this transaction given by the bank, if it
     * has given one. Otherwise empty. -- Bankreferenz in :61: */
    extern void 
    HBCI_Transaction_setBankReference(HBCI_Transaction *t, const char *s);
    /*@}*/

    /* Now follows the obnoxious "Mehrzweckfeld :86:" */

    /** @name Transaction Description */
    /*@{*/
    /** A 3 digit numerical transaction code, defined for all kinds of
     * different HBCI actions ("textkey"). -- Geschaeftsvorfallcode,
     * first field in field :86:
     *
     * For a normal transfer you should set it to 51. For debit notes
     * the values 04 or 05 (Watch out: 06 was WRONG!) may be used. For
     * other values please refer to your credit institute.
     */
    extern int HBCI_Transaction_transactionCode(const HBCI_Transaction *t);
    /** A 3 digit numerical transaction code, defined for all kinds of
     * different HBCI actions ("textkey"). -- Geschaeftsvorfallcode,
     * first field in field :86:
     *
     * For a normal transfer you should set it to 51. For debit notes
     * the values 04 or 05 (Watch out: 06 was WRONG!) may be used. For
     * other values please refer to your credit institute.
     */
    extern void 
    HBCI_Transaction_setTransactionCode(HBCI_Transaction *t, int c);

    /** HBCI_Transaction text (e.g. STANDING ORDER); field
     * 00 in field :86: */
    /*extern const char * 
      HBCI_Transaction_transactionText(const HBCI_Transaction *t);*/
    /** HBCI_Transaction text (e.g. STANDING ORDER); field
     * 00 in field :86: */
    /*extern void 
      HBCI_Transaction_setTransactionText(HBCI_Transaction *t, const char *s);*/

    /** Primanoten-Number (no idea what this means) --
     * Primanoten-Nummer; field 10 in field :86:  */
    /*extern const char * 
      HBCI_Transaction_primanota(const HBCI_Transaction *t);*/
    /** Primanoten-Number (no idea what this means) --
     * Primanoten-Nummer; field 10 in field :86:  */
    /*extern void 
      HBCI_Transaction_setPrimanota(HBCI_Transaction *t, const char *s);*/

    /** Description of what this transaction is for, and how the Bank
     * describes this transaction.  -- Verwendungszweck; fields 20-29
     * and 60-63 in field :86:  */
    extern const list_string* 
    HBCI_Transaction_description(const HBCI_Transaction *t);
    /** Description of what this transaction is for, and how the Bank
     * describes this transaction.  -- Verwendungszweck; fields 20-29
     * and 60-63 in field :86:  */
    extern void 
    HBCI_Transaction_addDescription(HBCI_Transaction *t, const char *s);
    /*@}*/

    /** @name Other account */
    /*@{*/
    /** Country code (BLZ) of the other party's institute -- not in
     * MT940/field :86: */
    /* extern int 
    HBCI_Transaction_otherCountryCode(const HBCI_Transaction *t);
    extern void 
    HBCI_Transaction_setOtherCountryCode(HBCI_Transaction *t, int i);*/

    /** Institute code (BLZ);
     * field 30 in field :86: */
    extern const char * 
    HBCI_Transaction_otherBankCode(const HBCI_Transaction *t);
    /** Institute code (BLZ);
     * field 30 in field :86: */
    extern void 
    HBCI_Transaction_setOtherBankCode(HBCI_Transaction *t, const char *s);

    /** Account number of the other party -- Kontonummer; field 31 in
     * field :86: */
    extern const char * 
    HBCI_Transaction_otherAccountId(const HBCI_Transaction *t);
    /** Account number of the other party -- Kontonummer; field 31 in
     * field :86: */
    extern void 
    HBCI_Transaction_setOtherAccountId(HBCI_Transaction *t, const char *s);

    /** If the account of the other party has subaccounts which are
     * distinguished by different suffixes, then this is that suffix
     * of the other party. Otherwise it's empty. -- *not* in MT940 /
     * field :86: */
    /*extern const char * 
      HBCI_Transaction_otherSuffix(const HBCI_Transaction *t);
      extern void 
      HBCI_Transaction_setOtherSuffix(HBCI_Transaction *t, const char *id);*/

    /** Name of the other party -- fields 32 and 33 in field :86: */
    extern const list_string* 
    HBCI_Transaction_otherName(const HBCI_Transaction *t);
    /** Name of the other party -- fields 32 and 33 in field :86: */
    extern void 
    HBCI_Transaction_addOtherName(HBCI_Transaction *t, const char *s);
    /*@}*/

    /** If this transaction was in a different currency than your
     * account is in, then this is the transaction value in the original
     * currency.  -- Ursprungsbetrag in :61: */
    extern const HBCI_Value *HBCI_Transaction_originalValue(const HBCI_Transaction *t);
    /** If this transaction was in a different currency than your
     * account is in, then this is the extra amount charged for currency
     * exchange.  */
    extern const HBCI_Value *HBCI_Transaction_charge(const HBCI_Transaction *t);

    /** @name List of HBCI_Transaction */
    /*@{*/
/*     extern list_HBCI_Transaction *list_HBCI_Transaction_new(); */
    /* FIXME: Can the GWEN_DB_NODE of this list_HBCI_Transaction
     * possibly be used for JobMultiTransfer? No it cannot. In theory
     * if yes, then the Application needs to have a function for
     * adding more transactions to this list. On the other hand, if
     * such a list_HBCI_Transaction only occurs as a return value then
     * of course the App doesn't need to add transactions to it and
     * this function is not needed.
     *
     * extern void list_HBCI_Transaction_push_back(list_HBCI_Transaction *l, 
     * const HBCI_Transaction *trans);*/
/*     extern GWEN_DB_NODE *list_HBCI_Transaction_node(list_HBCI_Transaction *l); */
/*     extern void list_HBCI_Transaction_delete(list_HBCI_Transaction *l); */
/*     extern int list_HBCI_Transaction_size(const list_HBCI_Transaction *l); */
    /** Callback function for @ref list_HBCI_Transaction_foreach . */
/*     typedef void *(*list_HBCI_Transaction_cb) (const HBCI_Transaction *trans,  */
/* 					       void *user_data); */
    /** Traverses the list, calling the callback function 'func' on
     * each list element.  Traversal will stop when 'func' returns a
     * non-NULL value, and the routine will return with that
     * value. Otherwise the routine will return NULL. 
     * @param list The list to traverse.
     * @param func The function to be called with each list element.
     * @param user_data A pointer passed on to the function 'func'.
     * @return The non-NULL pointer returned by 'func' as soon as it
     * returns one. Otherwise (i.e. 'func' always returns NULL)
     * returns NULL.
     * @author Christian Stimming <stimming@tuhh.de> */
/*     extern void * */
/*     list_HBCI_Transaction_foreach (const list_HBCI_Transaction *list,  */
/* 				   list_HBCI_Transaction_cb func, */
/* 				   void *user_data); */
    /*@}*/
    /*@}*/
#ifdef __cplusplus
}

#endif


#endif



