/***************************************************************************
  $RCSfile: msgxml.cpp,v $
  -------------------
  cvs         : $Id: msgxml.cpp,v 1.39 2004/06/04 11:36:20 aquamaniac Exp $
  begin       : Mon Jul 07 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "misc.h"
#include "msgxml.h"
#include "msgxmlqueue.h"
#include "msgxmljob.h"
#include "msgenginehbci.h"
#include <openhbci2/datetime.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/md.h>
#include <gwenhywfar/debug.h>
#include <gwenhywfar/gwenhywfar.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>



namespace HBCI {
  MessageEngineXML::XMLFile::XMLFile(const string &path, GWEN_XMLNODE *n)
  :_fpath(path)
  ,_node(n)
  ,_counter(0){

  }


  MessageEngineXML::XMLFile::~XMLFile(){
    GWEN_XMLNode_free(_node);
  }

  void MessageEngineXML::XMLFile::attach(){
    _counter++;
  }


  void MessageEngineXML::XMLFile::detach(){
    if (_counter)
      _counter--;
  }


  list<Pointer<MessageEngineXML::XMLFile> > MessageEngineXML::_xmlFiles;



  MessageEngineXML::MessageEngineXML(Hbci *hbci,
				     int country,
				     const string &bankCode,
				     int hversion)
  :MessageEngine(hbci, country, bankCode)
  ,_engine(0)
  ,_msgTimeout(60)
  ,_hversion(hversion)
  ,_dialogOpen(false){
    Error err;

    err=_loadConfig(country, bankCode, hversion, "hbci.xml");
    if (!err.isOk())
      throw Error("MessageEngineXML::MessageEngineXML", err);
    _config=GWEN_DB_Group_new("MessageEngineXML::_config");
  }


  MessageEngineXML::MessageEngineXML(Hbci *hbci,
				     int country,
				     const string &bankCode,
                                     int hversion,
				     GWEN_DB_NODE *cfg)
  :MessageEngine(hbci, country, bankCode)
  ,_engine(0)
  ,_msgTimeout(60)
  ,_hversion(hversion)
  ,_dialogOpen(false){
    Error err;

    err=_loadConfig(country, bankCode, hversion, "hbci.xml");
    if (!err.isOk())
      throw Error("MessageEngineXML::MessageEngineXML", err);
    _config=GWEN_DB_Group_dup(cfg);
  }


  Error MessageEngineXML::_loadConfig(int country,
				      const string &bankCode,
				      int hversion,
				      const string &fname) {
    GWEN_XMLNODE *defs;
    string fpath;
    list<Pointer<XMLFile> >::iterator it;

    if (_engine)
      // free current engine
      MsgEngineHBCI_free(_engine);
    // create new engine for this object
    _engine=MsgEngineHBCI_new(this);
    GWEN_MsgEngine_SetProtocolVersion(_engine, _hversion);

    // first try local file for bank
    fpath=hbci()->applicationDataDir();
    fpath+="/";
    fpath+=String::num2string(country, true, 3);
    fpath+="/";
    fpath+=bankCode;
    fpath+="/";
    fpath+=fname;
    if (access(fpath.c_str(), F_OK)) {
      // try local file for user
      fpath=hbci()->applicationDataDir();
      fpath+="/";
      fpath+=fname;
      if (access(fpath.c_str(), F_OK)) {
	// try global file
	fpath=XMLDATA_DIR;
	fpath+="/";
	fpath+=fname;
	if (access(fpath.c_str(), F_OK)) {
	  throw Error("MessageEngineXML::MessageEngineXML",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_DONTKNOW,
		      "No XML file found");
	}
      }
    }

    // check whether this file has already been loaded
    defs=0;
    for (it=_xmlFiles.begin(); it!=_xmlFiles.end(); it++) {
      if ((*it).ref().fileName()==fpath) {
	// found, attach to it
	(*it).ref().attach();
	defs=(*it).ref().node();
	DBG_INFO(0, "Sharing HBCI file (%s)", fpath.c_str());
        break;
      }
    }

    if (!defs) {
      // file not loaded, so do it now
      Pointer<XMLFile> xf;

      // load file
      defs=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag,"defs");
      // load the HBCI description file
      DBG_INFO(0, "Loading HBCI file (%s)", fpath.c_str());
      if (GWEN_XML_ReadFile(defs, fpath.c_str(), GWEN_XML_FLAGS_DEFAULT)) {
        GWEN_XMLNode_free(defs);
	return Error("MessageEngineXML::MessageEngineXML",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_UNKNOWN,
		     ERROR_ADVISE_DONTKNOW,
		     "Error loading XML file",
		     fpath);
      }
      xf=new XMLFile(fpath, defs);
      xf.ref().attach();
      _xmlFiles.push_back(xf);
    }

    // do NOT take over ownership, this is left to XMLFile
    GWEN_MsgEngine_SetDefinitions(_engine, defs, 0);

    DBG_INFO(0, "Config loaded.");
    return Error();
  }


  MessageEngineXML::~MessageEngineXML(){
    list<Pointer<XMLFile> >::iterator it;

    for (it=_xmlFiles.begin(); it!=_xmlFiles.end(); it++) {
      if ((*it).ref().node()==GWEN_MsgEngine_GetDefinitions(_engine)) {
	// found, detach from it
	DBG_INFO(0, "Detaching from  XMLFile");
	(*it).ref().detach();
	if ((*it).ref().usedBy()==0) {
	  DBG_INFO(0, "Erasing XMLFile");
	  _xmlFiles.erase(it);
	}
	break;
      }
    }
    MsgEngineHBCI_free(_engine);
    GWEN_DB_Group_free(_config);
  }


  void MessageEngineXML::setHbciVersion(int hversion) {
    if (_hversion!=hversion) {
      _hversion=hversion;
    }
  }


  Error MessageEngineXML::executeQueue(Pointer<MsgQueue> mq) {
    Error err;

    if (mq.ref().empty()) {
      DBG_NOTICE(0, "Empty queue");
    }
    else {
      Pointer<MsgQueueXML> xq;

      xq=mq.cast<MsgQueueXML>();
      err=_handleQueue(xq);
      if (!err.isOk())
	return Error("MessageEngineXML::executeQueue", err);
    }
    return Error();
  }


  Pointer<MsgJob> MessageEngineXML::jobFactory(const string &name,
					       Pointer<Customer> cu,
					       const string &account){
    Pointer<MsgJobXML> job;

    job=_jobFactory(name, cu, account);
    return job.cast<MsgJob>();
  }


  Pointer<MsgJobXML> MessageEngineXML::_jobFactory(const string &name,
						   Pointer<Customer> cu,
						   const string &account){
    GWEN_XMLNODE *node;
    GWEN_DB_NODE *jobData;
    GWEN_DB_NODE *jobUPD;
    GWEN_DB_NODE *jobBPD;
    GWEN_DB_NODE *updgrp;
    GWEN_DB_NODE *tmpcfg;
    const char *segCode;
    const char *paramName;
    bool needsBPD;
    bool dlg;
    Pointer<MsgJobXML> job;
    Pointer<User> u;
    Pointer<Medium> m;
    string secmode;
    string systemId;

    u=cu.ref().user();
    if (!u.isValid()) {
      throw Error("MessageEngineXML::_jobFactory",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_JOB_NOT_SUPPORTED_BY_LIB,
		  ERROR_ADVISE_DONTKNOW,
		  "No user in customer",
		  cu.ref().custId());
    }
    m=u.ref().medium();
    if (!m.isValid()) {
      throw Error("MessageEngineXML::_jobFactory",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_JOB_NOT_SUPPORTED_BY_LIB,
		  ERROR_ADVISE_DONTKNOW,
		  "No medium in user",
		  u.ref().userId());
    }
    switch(m.ref().securityMode()) {
    case HBCI_SECURITY_DDV:  secmode="DDV"; break;
    case HBCI_SECURITY_RDH:
      secmode="RDH";
      systemId=u.ref().systemId();
      break;
    default:
      throw Error("MessageEngineXML::_jobFactory",
		  ERROR_LEVEL_INTERNAL,
		  HBCI_ERROR_CODE_JOB_NOT_SUPPORTED_BY_LIB,
		  ERROR_ADVISE_DONTKNOW,
		  "Unsupported security mode for user",
		  u.ref().userId());
      break;
    } // switch

    // this is needed to select the correct nodes from within the XML file
    GWEN_MsgEngine_SetMode(_engine, secmode.c_str());

    // first select any version, we simply need to know the BPD job name
    node=GWEN_MsgEngine_FindNodeByProperty(_engine,
                                           "JOB",
                                           "id",
                                           0,
                                           name.c_str());
    if (!node) {
      throw Error("MessageEngineXML::_jobFactory",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_JOB_NOT_SUPPORTED_BY_LIB,
		  ERROR_ADVISE_DONTKNOW,
		  "Unsupported job (not in XML files)",
		  name);
    }
    needsBPD=(atoi(GWEN_XMLNode_GetProperty(node, "needbpd", "0"))!=0);
    paramName=GWEN_XMLNode_GetProperty(node, "params", "");
    segCode=GWEN_XMLNode_GetProperty(node, "code", "");
    dlg=(atoi(GWEN_XMLNode_GetProperty(node, "dlg", "0"))!=0);

    updgrp=GWEN_DB_GetGroup(_config,
                            GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                            cu.ref().custId().c_str());

    // create new config for the job
    jobData=GWEN_DB_Group_new("MessageEngineXML::_jobFactory::jobData");

    if (needsBPD) {
      int version;

      DBG_INFO(0, "Job \"%s\" needs BPD \"%s\"",
               name.c_str(),
               paramName);
      // check/create BPD
      jobBPD=GWEN_DB_GetGroup(_config,
			      GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			      "bpd/jobs");
      if (jobBPD) {
	// get BPD
	jobBPD=GWEN_DB_GetGroup(jobBPD,
				GWEN_PATH_FLAGS_NAMEMUSTEXIST,
				paramName);
	if (jobBPD) {
	  // children are one group per version
	  jobBPD=GWEN_DB_GetFirstGroup(jobBPD);
	}
      }

      node=0;
      // check for a job for which we have a BPD
      while(jobBPD) {
	// get version from BPD
	version=atoi(GWEN_DB_GroupName(jobBPD));

        // now get the correct version of the JOB
        DBG_INFO(0, "Checking for Job %s (%d)",
                 name.c_str(), version);
	node=GWEN_MsgEngine_FindNodeByProperty(_engine,
					       "JOB",
					       "id",
					       version,
					       name.c_str());
	if (node)
	  break;
        jobBPD=GWEN_DB_GetNextGroup(jobBPD);
      } // while

      if (!jobBPD) {
	// no BPD when needed, error
        GWEN_DB_Group_free(jobData);
        DBG_ERROR(0,"Job \"%s\" not supported by your bank", name.c_str());
	throw Error("MessageEngineXML::_jobFactory",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_JOB_NOT_SUPPORTED_BY_BANK,
		    ERROR_ADVISE_DONTKNOW,
		    "job not supported by your bank",
		    name);
      }

      // copy job parameters
      DBG_INFO(0, "Copying BPD description");
      tmpcfg=GWEN_DB_GetGroup(jobData,
                              GWEN_DB_FLAGS_DEFAULT,
                              "params");
      GWEN_DB_AddGroupChildren(tmpcfg, jobBPD);

    } // if needsBPD

    // check for UPD
    jobUPD=0;
    if (!cu.ref().custId().empty() && *segCode) {
      bool ignoreUPDJobs;
      bool hasUPDJob;

      ignoreUPDJobs=true;
      hasUPDJob=false;
      jobUPD=updgrp;
      if (jobUPD && !account.empty()) {
        jobUPD=GWEN_DB_GetGroup(jobUPD,
                                GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                                account.c_str());
        if (jobUPD) {
          ignoreUPDJobs=GWEN_DB_GetIntValue(jobUPD, "ignoreUPDJobs", 0, 1);
          jobUPD=GWEN_DB_GetFirstGroup(jobUPD);
          while(jobUPD) {
            if (strcasecmp(GWEN_DB_GroupName(jobUPD), "updjob")==0) {
              const char *pn;

              pn=GWEN_DB_GetCharValue(jobUPD, "job", 0, 0);
              if (pn) {
                // found a job
                if (strcasecmp(pn, segCode)==0) {
                  // found THE job description
                  DBG_INFO(0, "Copying UPD job data");
                  tmpcfg=GWEN_DB_GetGroup(jobData,
                                          GWEN_DB_FLAGS_DEFAULT,
                                          "params");
                  GWEN_DB_AddGroupChildren(tmpcfg, jobUPD);
                  hasUPDJob=true;
                  break;
                }
              } // if pn
            } // if "updjob"
            jobUPD=GWEN_DB_GetNextGroup(jobUPD);
          } // while
	} // if jobUPD
      } // if account group found

      // now check whether we have an UPD job description
      if (!hasUPDJob) {
        /* no UPDJob for this job, but the user knows about UPDJobs, so
         * this job is not supported for this customer
	 */
	if (!ignoreUPDJobs) {
	  GWEN_DB_Group_free(jobData);
	  throw Error("MessageEngineXML::_jobFactory",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_JOB_NOT_SUPPORTED_FOR_CUSTOMER,
		      ERROR_ADVISE_DONTKNOW,
		      "job not supported by your bank for this customer",
		      name);
	}
      }
    } // if customer Id and paramName are given

    // cleanup config
    _cleanUpConfig(jobData, "head");

    // finally create the job
    DBG_INFO(0, "Creating job with these settings:");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo)
      GWEN_DB_Dump(jobData, stderr, 1);
    job=new MsgJobXML(this,
		      cu,
		      name, account,
		      secmode,
		      jobData, node);
    return job;
  }


  Pointer<MsgQueue> MessageEngineXML::queueFactory(Pointer<Customer> cu){
    Pointer<MsgQueue> p;
    int maxdiffjobs;

    maxdiffjobs=GWEN_DB_GetIntValue(_config,
                                "bpd/jobtypespermsg", 0,
                                1);
    p=new MsgQueueXML(cu, maxdiffjobs);
    p.setObjectDescription("MsgQueueXML (MessageEngineXML::queueFactory)");
    p.setDescription("MsgQueueXML");
    return p;
  }


  void MessageEngineXML::_logMessage(const char *msg,
                                     unsigned int msgSize,
                                     bool crypt,
                                     bool bank) {
    FILE *f;
    int major, minor, patchlevel, build;

    f=fopen(_logFile.c_str(),"a+b");
    if (!f) {
      DBG_ERROR(0, "fopen(%s): %s", _logFile.c_str(), strerror(errno));
      return;
    }

    fprintf(f, "size: %d\n", msgSize);
    fprintf(f, "mode: %s\n", GWEN_MsgEngine_GetMode(_engine));
    fprintf(f, "crypt: ");
    fprintf(f, crypt?"yes":"no");
    fprintf(f, "\n");
    fprintf(f, "sender: ");
    fprintf(f, bank?"bank":"user");
    fprintf(f, "\n");
    GWEN_Version(&major,
		 &minor,
		 &patchlevel,
		 &build);
    fprintf(f, "gwenhywfar: %d.%d.%d.%d\n", major, minor, patchlevel, build);
    Hbci::libraryVersion(major, minor, patchlevel, build);
    fprintf(f, "openhbci: %d.%d.%d.%d\n", major, minor, patchlevel, build);
    fprintf(f, "application: %s\n", hbci()->applicationName().c_str());
    // append empty line as end of header
    fprintf(f, "\n");
    if (msgSize) {
      fwrite(msg, msgSize, 1, f);
    }
    // append CR for better readability
    fprintf(f, "\n");
    if (fclose(f)) {
      DBG_ERROR(0, "fclose(%s): %s", _logFile.c_str(), strerror(errno));
      return;
    }
  }


  GWEN_DB_NODE*
  MessageEngineXML::_exchangeMessages(Pointer<Transporter> tx,
                                      Pointer<Customer> owner,
				      list<Pointer<Customer> > signers,
				      char *msg,
				      unsigned int msgSize,
				      bool sign,
				      bool crypt) {
    string completeMsg;
    string result;
    Error err;
    GWEN_DB_NODE *cfg;
    const char *dlgId;
    int mn;

    DBG_DEBUG(0, "Will send this message:\n");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
      GWEN_Text_DumpString(msg, msgSize, stderr, 1);

    hbci()->interactor().ref().actionStarted(ACT_ENCODEMSG,
                                             "Encoding message");
    completeMsg=_encode(msg, msgSize, owner, signers, sign, crypt);

    DBG_DEBUG(0, "Encoded message:\n");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
      GWEN_Text_DumpString(completeMsg.data(), completeMsg.length(),
                           stderr, 1);

    // send message
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo) {
      DBG_INFO(0, "Hash of message to write (%d bytes) is: %s",
               completeMsg.length(),
               _hash(completeMsg).c_str());
    }
    hbci()->interactor().ref().actionStarted(ACT_LOW_SENDMSG,
					     "Sending message");
    err=tx.ref().writeBytes(completeMsg);
    if (!err.isOk()) {
      DBG_INFO(0, "Error writing bytes");
      throw Error("MessageEngineXML::_exchangeMessages", err);
    }

    err=Error();

    hbci()->interactor().ref().actionStarted(ACT_LOW_RECEIVEMSG,
					     "Receiving message");
    result=_readMessage(tx);

    DBG_DEBUG(0, "Received message:\n");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
      GWEN_Text_DumpString(result.data(), result.length(),
                           stderr, 1);

    // decode the received message
    hbci()->interactor().ref().actionStarted(ACT_DECODEMSG,
					     "Decoding message");
    cfg=_decode(owner, result);

    // extract dialogId, if necessary
    dlgId=GWEN_MsgEngine_GetValue(_engine, "DialogId", "0");
    if (strcmp(dlgId, "0")==0 || *dlgId==0) {
      DBG_INFO(0, "Setting DialogId to \"%s\"",
               GWEN_DB_GetCharValue(cfg,
                                    "MsgHead/dialogId",
                                    0,
                                    "0"));
      GWEN_MsgEngine_SetValue(_engine, "DialogId",
                              GWEN_DB_GetCharValue(cfg,
                                                   "MsgHead/dialogId",
                                                   0,
                                                   "0"));
    }

    // increment MessageNumber
    mn=GWEN_MsgEngine_GetIntValue(_engine, "MessageNumber",0);
    if (mn<0 || mn>1000) {
      DBG_ERROR(0,"Hey, what is that for a message number (%d)", mn);
    }
    mn++;
    GWEN_MsgEngine_SetIntValue(_engine, "MessageNumber", mn);

    // show message
    //fprintf(stderr, "Broken down message:\n");
    //Config_DumpGroup(stderr, cfg, 1);
    return cfg;
  }


  void MessageEngineXML::_cleanUpConfig(GWEN_DB_NODE *cfg,
					const char *name) {
    GWEN_DB_NODE *rungr;

    rungr=GWEN_DB_GetFirstGroup(cfg);
    while(rungr) {
      GWEN_DB_NODE *next;

      next=GWEN_DB_GetNextGroup(rungr);
      if (strcasecmp(GWEN_DB_GroupName(rungr), name)==0) {
        GWEN_DB_UnlinkGroup(rungr);
        GWEN_DB_Group_free(rungr);
      }
      else
        _cleanUpConfig(rungr, name);
      rungr=next;
    }
  }


  Error MessageEngineXML::_processResponse(Pointer<MsgQueueXML> mq,
					   list<Pointer<MsgJobXML> > &jobs,
					   GWEN_DB_NODE *cfg) {
    GWEN_DB_NODE *gr;
    string userid;
    MessageReference msgRef;
    list<Pointer<MsgJobXML> >::iterator jit;
    Pointer<User> user;

    DBG_INFO(0, "Processing this response:");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo)
      GWEN_DB_Dump(cfg, stderr, 1);

    user=mq.ref().owner().ref().user();
    userid=user.ref().userId();

    /* cleanup config (remove all "segment" groups, we dont need them any
     * longer */
    _cleanUpConfig(cfg, "segment");

    gr=GWEN_DB_GetFirstGroup(cfg);
    while(gr) {
      list<Pointer<MsgJobXML> >::iterator it;
      int ref;
      bool handled;
      GWEN_DB_NODE *nextgr;

      DBG_INFO(0, "Handling response group %s",
               GWEN_DB_GroupName(gr));
      nextgr=GWEN_DB_GetNextGroup(gr);
      handled=false;

      if (strcasecmp(GWEN_DB_GroupName(gr), "bpd")==0) {
	GWEN_DB_NODE *bpdgr;

        // save BankParameterData
        DBG_INFO(0, "Saving BPD");
        bpdgr=GWEN_DB_GetGroup(_config,
                               GWEN_DB_FLAGS_OVERWRITE_GROUPS, "bpd");
        assert(bpdgr);
        _cleanUpConfig(gr, "head");
	// save a copy of this group as new BPD
        GWEN_DB_AddGroupChildren(bpdgr, gr);
      }
      else if (strcasecmp(GWEN_DB_GroupName(gr), "bankmsg")==0) {
        // store adressee of the message
        GWEN_DB_SetCharValue(gr,
                             GWEN_DB_FLAGS_DEFAULT,
                             "customer",
                             mq.ref().owner().ref().custId().c_str());
      }
      else if (strcasecmp(GWEN_DB_GroupName(gr), "userdata")==0) {
        const char *p;
	GWEN_DB_NODE *updgr;

	p=GWEN_DB_GetCharValue(gr, "userid", 0,0);
	if (p) {
          if (strcasecmp(userid.c_str(), p)!=0) {
            if (strcmp(p, "9999999999")==0) {
              DBG_NOTICE(0,"Anonymous user");
            }
            else {
              DBG_ERROR(0,"Bad user id ! (%s != %s)",
                        userid.c_str(), p);
              GWEN_DB_Group_free(cfg);
              return Error("MessageEngineXML::_processResponse",
                           ERROR_LEVEL_NORMAL,
                           HBCI_ERROR_CODE_UNKNOWN,
                           ERROR_ADVISE_DONTKNOW,
                           "Bad segment: Bad customer id !");
            }
          }
	  updgr=GWEN_DB_GetGroup(_config,
				 GWEN_DB_FLAGS_OVERWRITE_GROUPS, p);
	  assert(updgr);
	  _cleanUpConfig(gr, "head");
	  GWEN_DB_AddGroupChildren(updgr, gr);
	}
	else {
	  DBG_ERROR(0,"No user id !");
	  GWEN_DB_Group_free(cfg);
	  return Error("MessageEngineXML::_processResponse",
		       ERROR_LEVEL_NORMAL,
		       HBCI_ERROR_CODE_UNKNOWN,
		       ERROR_ADVISE_DONTKNOW,
		       "Bad segment: No user id !");
	}
      }
      else if (strcasecmp(GWEN_DB_GroupName(gr), "accountdata")==0) {
	GWEN_DB_NODE *updgr;
	const char *accountid;

        _cleanUpConfig(gr, "head");
        if (userid.empty()) {
	  DBG_ERROR(0,"accountdata for unknown user!");
	  GWEN_DB_Group_free(cfg);
	  return Error("MessageEngineXML::_processResponse",
		       ERROR_LEVEL_NORMAL,
		       HBCI_ERROR_CODE_UNKNOWN,
		       ERROR_ADVISE_DONTKNOW,
		       "AccountData for unknown user");
	}
	else {
	  DBG_INFO(0, "Saving UPD");
	  if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo)
	    GWEN_DB_Dump(gr, stderr, 1);
	  accountid=GWEN_DB_GetCharValue(gr, "accountid", 0, 0);
	  DBG_INFO(0, "AccountID is: %s", accountid);
	  if (accountid) {
	    const char *customerid;

	    customerid=GWEN_DB_GetCharValue(gr, "customer", 0, 0);
	    if (customerid) {
	      Pointer<Customer> newCust;

	      /* check whether we need to add this customer to the user */
	      newCust=user.ref().findCustomer(customerid);
	      if (!newCust.isValid()) {
		DBG_NOTICE(0,
			   "Adding customer \"%s\" for user \"%s\"",
			   customerid, userid.c_str());
		newCust=new Customer(user, customerid,
				     GWEN_DB_GetCharValue(gr,
							  "name1",
							  0, ""));
		user.ref().addCustomer(newCust);
	      }
	      else {
		DBG_NOTICE(0, "Customer \"%s\" for user \"%s\" exists.",
			   customerid, userid.c_str());
	      }
	    }
	    else {
              DBG_NOTICE(0, "No customer in accountdata");
	    }
	    DBG_INFO(0, "We have an account id (%s/%s)",
		     userid.c_str(), accountid);
	    updgr=GWEN_DB_GetGroup(_config,
				   GWEN_DB_FLAGS_DEFAULT,
				   userid.c_str());
	    assert(updgr);
	    updgr=GWEN_DB_GetGroup(updgr,
				   GWEN_DB_FLAGS_OVERWRITE_GROUPS,
				   accountid);
	    assert(updgr);
            // save a copy of this group as new UPD
	    _cleanUpConfig(gr, "head");
	    GWEN_DB_AddGroupChildren(updgr, gr);
	  } // if accountid
	}
      } // if accountdata
      else if (strcasecmp(GWEN_DB_GroupName(gr), "msghead")==0) {
	GWEN_DB_NODE *msggr;

        msggr=GWEN_DB_GetGroup(gr, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "msgref");
	if (msggr) {
	  string dialogId;
	  int msgNum;

          dialogId=GWEN_DB_GetCharValue(gr, "dialogId", 0, "");
          msgNum=GWEN_DB_GetIntValue(gr, "msgnum", 0, 0);
	  msgRef=MessageReference(dialogId, msgNum);
          DBG_INFO(0, "Found message reference %s:%d",
                   dialogId.c_str(), msgNum);
        }
      }
      else if (strcasecmp(GWEN_DB_GroupName(gr), "CryptHead")==0 ||
	       strcasecmp(GWEN_DB_GroupName(gr), "CryptTail")==0 ||
	       strcasecmp(GWEN_DB_GroupName(gr), "CryptData")==0 ||
	       strcasecmp(GWEN_DB_GroupName(gr), "MsgTail")==0 ||
	       strcasecmp(GWEN_DB_GroupName(gr), "SigTail")==0
	      ) {
	// remove segment
        GWEN_DB_UnlinkGroup(gr);
        GWEN_DB_Group_free(gr);
        gr=0;
      }
      else {
	/* check for BPD job */
	GWEN_XMLNODE *bpdn;
        int segver;

        DBG_INFO(0, "Checking whether \"%s\" is a BPD job",
                 GWEN_DB_GroupName(gr));
        segver=GWEN_DB_GetIntValue(gr, "head/version", 0, 0);
	// get segment description (first try id, then code)
	bpdn=GWEN_MsgEngine_FindNodeByProperty(_engine,
					       "SEG",
					       "id",
					       segver,
                                               GWEN_DB_GroupName(gr));
	if (!bpdn)
	  bpdn=GWEN_MsgEngine_FindNodeByProperty(_engine,
						 "SEG",
						 "code",
						 segver,
                                                 GWEN_DB_GroupName(gr));

        if (bpdn) {
          DBG_DEBUG(0, "Found a candidate");
            if (atoi(GWEN_XMLNode_GetProperty(bpdn, "isbpdjob", "0"))) {
            /* segment contains a BPD job, move contents to
             * "bpd/jobs/<CODE>/<VERSION> in global config of this object */
            GWEN_DB_NODE *bn;
            char numbuffer[32];

            DBG_INFO(0, "Found BPD job \"%s\"",GWEN_DB_GroupName(gr));
            bn=GWEN_DB_GetGroup(_config, GWEN_DB_FLAGS_DEFAULT, "bpd/jobs");
            assert(bn);
            bn=GWEN_DB_GetGroup(bn, GWEN_DB_FLAGS_DEFAULT,
                                GWEN_DB_GroupName(gr));
            assert(bn);

            if (GWEN_Text_NumToString(segver,
                                      numbuffer,
                                      sizeof(numbuffer)-1,
                                      0)<1) {
              return Error("MessageEngineXML::_processResponse",
                           ERROR_LEVEL_NORMAL,
                           HBCI_ERROR_CODE_UNKNOWN,
                           ERROR_ADVISE_DONTKNOW,
                           "could not convert num to string");
            }
            bn=GWEN_DB_GetGroup(bn, GWEN_DB_FLAGS_OVERWRITE_GROUPS,
                                numbuffer);
            assert(bn);
            // remove "head" group
            _cleanUpConfig(gr, "head");
            if (GWEN_DB_AddGroupChildren(bn, gr)) {
              return Error("MessageEngineXML::_processResponse",
                           ERROR_LEVEL_NORMAL,
                           HBCI_ERROR_CODE_UNKNOWN,
                           ERROR_ADVISE_DONTKNOW,
                           "could not add group children");
            }
            // remove segment
            GWEN_DB_UnlinkGroup(gr);
            GWEN_DB_Group_free(gr);
            gr=0;
          } // if isbpdjob
          else {
            DBG_INFO(0, "Segment is known but not as a BPD job");
          }
        } // if segment found
        else {
          DBG_WARN(0, "Did not find segment \"%s\" (%d) ignoring",
                   GWEN_DB_GroupName(gr), segver);
        }
      }

      if (gr) {
        ref=GWEN_DB_GetIntValue(gr, "head/ref", 0, 0);
	if (ref) {
          for (it=jobs.begin(); it!=jobs.end(); it++) {
            if ((*it).ref().hasSegment(ref)) {
	      GWEN_DB_NODE *jobgr;

	      if (strcasecmp(GWEN_DB_GroupName(gr), "SegResult")==0) {
		int resultCode;
                string resultText;
                string resultParam;
                MsgJob::Result res;
                GWEN_DB_NODE *rn;

                rn=GWEN_DB_GetFirstGroup(gr);
                while (rn) {
                  if (strcasecmp(GWEN_DB_GroupName(rn), "Result")==0) {
                    resultCode=GWEN_DB_GetIntValue(rn, "resultcode", 0,0);
                    resultText=GWEN_DB_GetCharValue(rn, "text", 0, "");
                    resultParam=GWEN_DB_GetCharValue(rn, "param", 0, "");
                    // store reference segment name, for debugging purposes
                    GWEN_DB_SetIntValue(gr,
                                        GWEN_DB_FLAGS_DEFAULT,
                                        "reference", ref);
                    res=MsgJob::Result(resultCode,
                                       resultText,
                                       resultParam);
                    // found a result
                    DBG_INFO(0, "Adding result to job \"%s\"",
                             (*it).ref().name().c_str());
                    // create result code and add it
                    (*it).ref().addResult(res);
                  } // if result
                  rn=GWEN_DB_GetNextGroup(rn);
                } // while results
              }

              // response belongs to this job
              DBG_INFO(0, "Adding response \"%s\" to job \"%s\"",
                       GWEN_DB_GroupName(gr),
                       (*it).ref().name().c_str());
              jobgr=(*it).ref().jobData();
	      assert(jobgr);
              jobgr=GWEN_DB_GetGroup(jobgr, GWEN_DB_FLAGS_DEFAULT, "response");
              assert(jobgr);
	      // move group to the job's config
              _cleanUpConfig(gr, "head");
              GWEN_DB_UnlinkGroup(gr);
              GWEN_DB_AddGroup(jobgr, gr);
	      gr=0;
	      handled=true;
	      break;
	    } // if hasSegment
	    else {
              DBG_INFO(0, "Job %s does not have segment %d",
                       (*it).ref().name().c_str(), ref);
	    }
	  } // for
	} // if ref
      } // if gr
      if (!handled && gr) {
	GWEN_DB_NODE *qgr;

        _cleanUpConfig(gr, "head");

        if (strcasecmp(GWEN_DB_GroupName(gr), "MsgResult")==0) {
	  int resultCode;
	  string resultText;
	  MsgQueue::Result res;
          GWEN_DB_NODE *rn;

          rn=GWEN_DB_GetFirstGroup(gr);
          while (rn) {
            if (strcasecmp(GWEN_DB_GroupName(rn), "Result")==0) {
              resultCode=GWEN_DB_GetIntValue(rn, "resultcode", 0,0);
              resultText=GWEN_DB_GetCharValue(rn, "text", 0, "");

              res=MsgQueue::Result(resultCode,
                                   resultText);
              // found a result
              DBG_INFO(0, "Adding result to queue");
              // create result code and add it
              mq.ref().addResult(res);
            } // if result
            rn=GWEN_DB_GetNextGroup(rn);
          } // while results
        } // if MsgResult

        // MsgResult and all unhandled groups go to the queue
        DBG_INFO(0, "Adding response \"%s\" to queue",
                 GWEN_DB_GroupName(gr));
        qgr=mq.ref().queueData();
        assert(qgr);
        GWEN_DB_UnlinkGroup(gr);
        GWEN_DB_AddGroup(qgr, gr);
        gr=0;
        handled=true;
      } // if !handled

      if (!handled && gr) {
	// if still not handled discard this group
        GWEN_DB_UnlinkGroup(gr);
	GWEN_DB_Group_free(gr);
      }
      gr=nextgr;
    } //while

    // set message reference in all jobs
    for (jit=jobs.begin(); jit!=jobs.end(); jit++) {
      (*jit).ref().setMessageReference(msgRef);
    }

    if (GWEN_DB_GetFirstGroup(cfg)!=0) {
      DBG_ERROR(0,"Group is not empty !");
      GWEN_DB_Dump(cfg, stderr, 1);
    }
    GWEN_DB_Group_free(cfg);

    return Error();
  }


  string MessageEngineXML::_getBinProperty(GWEN_DB_NODE *cfg,
					   const string &name,
					   const string &defValue){
    string s;
    const void *d;
    unsigned int size;

    size=0;
    d=GWEN_DB_GetBinValue(cfg, name.c_str(), 0, 0, 0, &size);
    if (!d || !size)
      return defValue;

    return string((char*)d, size);
  }


  void MessageEngineXML::_setBinProperty(GWEN_DB_NODE *cfg,
					 const string &name,
					 const string &value){
    if (GWEN_DB_SetBinValue(cfg,
                            GWEN_DB_FLAGS_DEFAULT |
                            GWEN_DB_FLAGS_OVERWRITE_VARS,
                            name.c_str(),
                            value.c_str(),
                            value.length())) {
      throw Error("MessageEngineXML::setBinPropertyt()",
                  ERROR_LEVEL_NORMAL,
                  HBCI_ERROR_CODE_UNKNOWN,
		  ERROR_ADVISE_DONTKNOW,
                  "Error setting binary value.",
                  name);
    }
  }


  Error MessageEngineXML::_mountCustomersMedium(Pointer<Customer> cust) {
    Error err;

    if (!cust.isValid()) {
      return Error("MessageEngineXML::_mountCustomersMedium",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Invalid owner.");
    }

    try {
      _medium=hbci()->mountCustomersMedium(country(), bankCode(), cust);
    }
    catch(Error xerr) {
      return Error("MessageEngineXML::_mountCustomersMedium()",xerr);
    }
    return Error();
  }


  string
  MessageEngineXML::_encode(char *msg,
			    unsigned int msgSize,
			    Pointer<Customer> owner,
			    const list<Pointer<Customer> > &signers,
			    bool needSign,
			    bool needCrypt){
    list<Pointer<Customer> >::const_iterator it;
    string result;
    Error err;
    unsigned int pos;
    unsigned int firstSegNum;
    unsigned int lastSegNum;
    GWEN_BUFFER *gbuf;

    result=string(msg, msgSize);
    gbuf=GWEN_Buffer_new(0, MSGENGINE_HBCI_MAXSIZE, 0, 1);
    firstSegNum=2;
    lastSegNum=GWEN_MsgEngine_GetIntValue(_engine, "SegmentNumber",
					  0);
    if (lastSegNum<firstSegNum) {
      GWEN_Buffer_free(gbuf);
      throw Error("MessageEngineXML::_encode",
		  ERROR_LEVEL_INTERNAL,
		  HBCI_ERROR_CODE_UNKNOWN,
		  ERROR_ADVISE_ABORT,
		  "lastSegNum<firstSegNum");
    }

    if (needSign) {
      GWEN_XMLNODE *sigHeadNode;
      GWEN_XMLNODE *sigTailNode;
      string pureMessage;
      GWEN_DB_NODE *cfg;

      if (signers.empty()) {
        DBG_ERROR(0, "No signers !");
        throw Error("MessageEngineXML::_encode",
                    ERROR_LEVEL_INTERNAL,
                    HBCI_ERROR_CODE_UNKNOWN,
                    ERROR_ADVISE_ABORT,
                    "No signers");
      }
      DBG_INFO(0, "Signing message");
      lastSegNum=GWEN_MsgEngine_GetIntValue(_engine, "SegmentNumber",
					    signers.size()+1);
      firstSegNum=signers.size()+1; // not -1 !
      sigHeadNode=GWEN_MsgEngine_FindNodeByProperty(_engine,
						    "SEG",
						    "id",
						    0,
						    "SigHead");
      if (!sigHeadNode) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not get node for SigHead");
      }

      sigTailNode=GWEN_MsgEngine_FindNodeByProperty(_engine,
						    "SEG",
						    "id",
						    0,
						    "SigTail");
      if (!sigTailNode) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not get node for SigTail");
      }

      pureMessage=result;
      for (it=signers.begin();
	   it!=signers.end();
	   it++) {
	string controlref;
	string sign;
	string toSign;
	string seg;
	int rv;
	Pointer<User> u;

	DBG_INFO(0, "Letting customer \"%s\" sign",
		 (*it).ref().custId().c_str());
	u=(*it).ref().user();
	if (!u.isValid()) {
	  GWEN_Buffer_free(gbuf);
	  throw Error("MessageEngineXML::_encode",
		      ERROR_LEVEL_INTERNAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_ABORT,
		      "Could not get user from customer");
	}

        DBG_INFO(0, "Mounting medium for signing");
	err=_mountCustomersMedium(*it);
	if (!err.isOk()) {
	  DBG_ERROR(0,"Could not mount medium of \"%s\"",
		    (*it).ref().custId().c_str());
	  GWEN_Buffer_free(gbuf);
	  throw Error("MessageEngineXML::_letAllSign", err);
	}

	controlref=String::date2string()+String::time2string();
	controlref=controlref.substr(0,14);

	cfg=GWEN_DB_Group_new("signhead");
	GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "head/seq",
			    firstSegNum);
	GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			     "ctrlref", controlref.c_str());
	// set medium data
	GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "SecDetails/dir", "1");
        if (_medium.ref().securityMode()==HBCI_SECURITY_DDV)
          _setBinProperty(cfg, "SecDetails/SecId",
                          _medium.ref().mediumId());
        else
          GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                               "SecDetails/SecId",
                               _medium.ref().mediumId().c_str());

	GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "signseq",
			    _medium.ref().nextSEQ());
	GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "key/keynum",
			    _medium.ref().signKeyNumber());
	GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "key/keyversion",
			    _medium.ref().signKeyVersion());
	// set user/bank/customer data
        GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "key/userid", (*it).ref().user().ref().userId().c_str());
	GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "key/country", country());
	GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			     "key/bankcode", bankCode().c_str());

	GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			     "SecStamp/date",
			     String::date2string().c_str());
	GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			     "SecStamp/time",
			     String::time2string().c_str());

	pos=0;
        GWEN_Buffer_Reset(gbuf);
	rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
						sigHeadNode,
                                                gbuf,
						cfg);
	GWEN_DB_Group_free(cfg);
	if (rv) {
	  GWEN_Buffer_free(gbuf);
	  throw Error("MessageEngineXML::_letAllSign",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_DONTKNOW,
		      "Error creating signature head");
	}
	seg=string(GWEN_Buffer_GetStart(gbuf),
		   GWEN_Buffer_GetUsedBytes(gbuf));
	toSign=seg+pureMessage;

        hbci()->interactor().ref().actionStarted(ACT_SIGNMSG,
						 "Signing message");
        sign=_medium.ref().sign(toSign);
	result=seg+result;

	// create signature tail
	cfg=GWEN_DB_Group_new("sigtail");
	GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "ctrlref", controlref.c_str());
	_setBinProperty(cfg, "signature", sign);
	GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "head/seq",
			    lastSegNum);

	pos=0;
        GWEN_Buffer_Reset(gbuf);
	rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
						sigTailNode,
                                                gbuf,
						cfg);
	GWEN_DB_Group_free(cfg);
	if (rv) {
	  GWEN_Buffer_free(gbuf);
	  throw Error("MessageEngineXML::_letAllSign",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_DONTKNOW,
		      "Error creating signature tail");
	}
	seg=string(GWEN_Buffer_GetStart(gbuf),
		   GWEN_Buffer_GetUsedBytes(gbuf));
	result+=seg;
	firstSegNum--;
	lastSegNum++;
      } // for
    } // needsign

    // log message
    _logMessage(result.data(), result.length(), false, false);

    // result contains message so far
    DBG_DEBUG(0, "Signed message:");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
      GWEN_Text_DumpString(result.c_str(), result.length(), stderr, 1);

    if (needCrypt) {
      GWEN_XMLNODE *cryptHeadNode;
      GWEN_XMLNODE *cryptTailNode;
      GWEN_DB_NODE *cfg;
      string msgkey;
      string cryptedkey;
      string crypteddata;
      string seg;
      int rv;
      Pointer<User> user;

      DBG_INFO(0, "Encrypting message");
      user=owner.ref().user();
      cryptHeadNode=GWEN_MsgEngine_FindNodeByProperty(_engine,
						      "SEG",
						      "id",
						      0,
						      "CryptHead");
      if (!cryptHeadNode) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not get node for CryptHead");
      }

      cryptTailNode=GWEN_MsgEngine_FindNodeByProperty(_engine,
						      "SEG",
						      "id",
						      0,
						      "CryptData");
      if (!cryptTailNode) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not get node for CryptTail");
      }

      DBG_INFO(0, "Mounting medium for encrypting");
      err=_mountCustomersMedium(owner);
      if (!err.isOk()) {
	DBG_ERROR(0,"Could not mount medium of \"%s\"",
		  (*it).ref().custId().c_str());
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode", err);
      }

      hbci()->interactor().ref().actionStarted(ACT_ENCRYPTMSG,
					       "Encrypting message");
      // create message key
      msgkey=_medium.ref().createMessageKey();
      // encrypt message
      crypteddata=_medium.ref().encrypt(result,msgkey);
      // encrypt message key
      cryptedkey=_medium.ref().encryptKey(msgkey);

      cfg=GWEN_DB_Group_new("crypthead");
      // set message data
      _setBinProperty(cfg, "CryptAlgo/MsgKey", cryptedkey);
      // set medium data
      GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			    "SecDetails/dir", "1");
      if (_medium.ref().securityMode()==HBCI_SECURITY_DDV) {
        _setBinProperty(cfg, "SecDetails/SecId",
                        _medium.ref().mediumId());
        GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "key/userid",
                             user.ref().userId().c_str());
      }
      else {
        Pointer<MediumRDHBase> mrdh;

        mrdh=_medium.cast<MediumRDHBase>();
        GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                             "SecDetails/SecId",
                             _medium.ref().mediumId().c_str());
        if (mrdh.ref().cryptKeyOwner().empty()) {
          GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                               "key/userid",
                               user.ref().userId().c_str());
        }
        else {
          GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
                               "key/userid",
                               mrdh.ref().cryptKeyOwner().c_str());
        }
      }
      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "key/keynum",
                          _medium.ref().cryptKeyNumber());
      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			 "key/keyversion",
                          _medium.ref().cryptKeyVersion());
      // set user/bank/customer data
      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "key/country", country());
      GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			   "key/bankcode", bankCode().c_str());

      GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			   "SecStamp/date",
			   String::date2string().c_str());
      GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			   "SecStamp/time",
			   String::time2string().c_str());
      pos=0;
      GWEN_Buffer_Reset(gbuf);
      rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
					      cryptHeadNode,
                                              gbuf,
					      cfg);
      GWEN_DB_Group_free(cfg);
      if (rv) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error creating crypthead");
      }
      result=string(GWEN_Buffer_GetStart(gbuf),
                    GWEN_Buffer_GetUsedBytes(gbuf));

      // create crypt data segment
      cfg=GWEN_DB_Group_new("cryptdata");
      // set message data
      _setBinProperty(cfg, "cryptdata", crypteddata);
      pos=0;
      GWEN_Buffer_Reset(gbuf);
      rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
					      cryptTailNode,
                                              gbuf,
                                              cfg);
      GWEN_DB_Group_free(cfg);
      if (rv) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error creating encrypted data");
      }
      seg=string(GWEN_Buffer_GetStart(gbuf),
		 GWEN_Buffer_GetUsedBytes(gbuf));
      result+=seg;

      DBG_DEBUG(0, "Encrypted message:");
      if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
        GWEN_Text_DumpString(result.c_str(), result.length(), stderr, 1);
    } // needcrypt


    // add MsgTail
    if (1) { // just to have local vars
      GWEN_XMLNODE *msgHeadNode;
      GWEN_XMLNODE *msgTailNode;
      GWEN_DB_NODE *cfg;
      string seg;
      int rv;
      int msize;

      DBG_INFO(0, "Adding message tail");
      msgHeadNode=GWEN_MsgEngine_FindNodeByProperty(_engine,
						    "SEG",
						    "id",
						    0,
						    "MsgHead");
      if (!msgHeadNode) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not get node for MsgHead");
      }

      msgTailNode=GWEN_MsgEngine_FindNodeByProperty(_engine,
						    "SEG",
						    "id",
						    0,
						    "MsgTail");
      if (!msgTailNode) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not get node for MsgTail");
      }

      // create msg tail segment
      cfg=GWEN_DB_Group_new("msgtail");

      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			 "head/seq",
			 lastSegNum);
      pos=0;
      GWEN_Buffer_Reset(gbuf);
      rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
					      msgTailNode,
                                              gbuf,
					      cfg);
      GWEN_DB_Group_free(cfg);
      if (rv) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error creating message tail");
      }
      seg=string(GWEN_Buffer_GetStart(gbuf),
		 GWEN_Buffer_GetUsedBytes(gbuf));
      result+=seg;

      // create msg head segment (premature)
      cfg=GWEN_DB_Group_new("msghead");

      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "size", 1);
      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "head/seq", 1);

      pos=0;
      GWEN_Buffer_Reset(gbuf);
      rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
					      msgHeadNode,
					      gbuf,
					      cfg);
      if (rv) {
	GWEN_DB_Group_free(cfg);
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error creating message tail");
      }

      // set real size
      msize=result.length()+GWEN_Buffer_GetUsedBytes(gbuf);
      DBG_INFO(0, "Message size is: %d", msize);
      GWEN_DB_SetIntValue(cfg,
			  GWEN_DB_FLAGS_OVERWRITE_VARS,
			  "size", msize);
      GWEN_DB_SetIntValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "head/seq", 1);
      // create again, now with correct size
      pos=0;
      GWEN_Buffer_Reset(gbuf);
      rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
					      msgHeadNode,
					      gbuf,
					      cfg);
      GWEN_DB_Group_free(cfg);
      if (rv) {
	GWEN_Buffer_free(gbuf);
	throw Error("MessageEngineXML::_encode",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error creating message tail");
      }

      seg=string(GWEN_Buffer_GetStart(gbuf),
		 GWEN_Buffer_GetUsedBytes(gbuf));
      result=seg+result;
    }

    // log message
    _logMessage(result.data(), result.length(), needCrypt, false);

    GWEN_Buffer_free(gbuf);
    return result;
  }


  string MessageEngineXML::_hash(const string &s) {
    char buffer[64];
    char hashbuffer[128];
    unsigned int bsize;

    bsize=sizeof(buffer);
    if (GWEN_MD_Hash("RMD160",
		     s.data(),
		     s.length(),
		     buffer,
		     &bsize)) {
      DBG_ERROR(0, "Could not hash data (%d bytes)", s.length());
    }
    GWEN_Text_ToHex(buffer, bsize, hashbuffer, sizeof(hashbuffer));
    return string(hashbuffer);
  }


  string MessageEngineXML::_readMessage(Pointer<Transporter> tx) {
    GWEN_XMLNODE *node;
    int size;
    Pointer<Interactor> interactor;
    string result;
    string tmp;
    Error err;
    time_t startTime;
    GWEN_DB_NODE *cfg;
    GWEN_BUFFER *gbuf;
    int rv;

    node=GWEN_MsgEngine_FindNodeByProperty(_engine,
                                           "SEG",
                                           "id",
                                           0,
                                           "MsgHeadShort");
    if (!node) {
      throw Error("MessageEngineXML::_readMessage",
                  ERROR_LEVEL_INTERNAL,
                  HBCI_ERROR_CODE_UNKNOWN,
                  ERROR_ADVISE_ABORT,
                  "Could not get node for MsgHeadShort");
    }

    // get start time for timeout calculation
    startTime=time(0);

    if (Hbci::debugLevel()>15)
      fprintf(stderr,"MessageEngineXML::_readMessage\n");
    interactor=hbci()->interactor();


    // first read up to 30 bytes
    size=30;
    while (interactor.ref().keepAlive() && size>0) {
      tmp.erase();
      err=Error();
      try {
	tmp=tx.ref().readBytes(size);
      }
      catch (Error xerr) {
	DBG_DEBUG(0, "Exception: %s", xerr.errorString().c_str());
	err=xerr;
      }
      if (!err.isOk()) {
	if (err.code()!=HBCI_ERROR_CODE_SOCKET_ERROR_TIMEOUT) {
	  if (Hbci::debugLevel()>1)
	    fprintf(stderr,"Error %d(%s)\n",
		    err.code(),
		    err.errorString().c_str());
          DBG_ERROR(0, "Error is fatal");
	  throw Error("MessageEngineXML::_readMessage", err);
	}
      } // if error
      else {
	if (tmp.length()==0) {
	  // broken pipe
	  DBG_ERROR(0, "Broken pipe");
	  throw Error("MessageEngineXML::_readMessage",
		      ERROR_LEVEL_INTERNAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_ABORT,
		      "Connection lost");
	}
	result+=tmp;
	size-=tmp.length();
      }
      if (difftime(time(0), startTime)>_msgTimeout) {
	if (Hbci::debugLevel()>0)
          fprintf(stderr,
                  "Timeout while reading message header\n");
        throw Error("MessageEngineXML::_readMessage",
                    ERROR_LEVEL_INTERNAL,
                    HBCI_ERROR_CODE_UNKNOWN,
                    ERROR_ADVISE_ABORT,
                    "Message timeout (while receiving header)");
      }
    } // while
    if (size>0) {
      DBG_ERROR(0, "Incomplete message (%d bytes left)", size);
      throw Error("MessageEngineXML::_readMessage",
		  ERROR_LEVEL_INTERNAL,
		  HBCI_ERROR_CODE_UNKNOWN,
		  ERROR_ADVISE_ABORT,
		  "Incomplete message received, user aborted");
    }
    cfg=GWEN_DB_Group_new("MessageEngineXML::_readMessage::cfg");
    gbuf=GWEN_Buffer_new((char*)result.data(),
                         result.length(),
                         result.length(),
                         0);
    /* temporarily disable protocol version dependant parsing
     * This allows parsing even segments with version which normally are not
     * used with the current HBCI protocol version.
     * It is allowed for banks to use segments from older HBCI versions
     * with newer versions so we must take this into account.
     */
    GWEN_MsgEngine_SetProtocolVersion(_engine, 0);
    rv=GWEN_MsgEngine_ParseMessage(_engine,
                                   node,
                                   gbuf,
                                   cfg,
                                   GWEN_MSGENGINE_READ_FLAGS_DEFAULT);
    GWEN_MsgEngine_SetProtocolVersion(_engine, _hversion);
    if (rv!=0) {
      GWEN_DB_Group_free(cfg);
      GWEN_Buffer_free(gbuf);
      DBG_ERROR(0, "Error parsing message");
      throw Error("MessageEngineXML::_readMessage",
		  ERROR_LEVEL_INTERNAL,
		  HBCI_ERROR_CODE_UNKNOWN,
		  ERROR_ADVISE_ABORT,
		  "Could not parse message");
    }

    size=GWEN_DB_GetIntValue(cfg, "size", 0,0);
    DBG_DEBUG(0, "Message size is: %d", size);
    GWEN_DB_Group_free(cfg);
    GWEN_Buffer_free(gbuf);
    if (size<(int)result.length()) {
      throw Error("MessageEngineXML::_readMessage",
                  ERROR_LEVEL_INTERNAL,
                  HBCI_ERROR_CODE_UNKNOWN,
                  ERROR_ADVISE_ABORT,
                  "Bad message size");
    }
    size-=result.length();

    // now read the rest of the message
    while (interactor.ref().keepAlive() && size>0) {
      DBG_DEBUG(0, "In data loop (%d, %d)", size, result.length());
      tmp.erase();
      err=Error();
      try {
	tmp=tx.ref().readBytes(size);
      }
      catch (Error xerr) {
	err=xerr;
      }
      if (!err.isOk()) {
	if (err.code()!=HBCI_ERROR_CODE_SOCKET_ERROR_TIMEOUT) {
	  if (Hbci::debugLevel()>1)
	    fprintf(stderr,"Error %d(%s)\n",
		    err.code(),
		    err.errorString().c_str());
	  throw Error("MessageEngineXML::_readMessage", err);
	}
      } // if error
      else {
	if (tmp.length()==0)
	  // broken pipe
	  throw Error("MessageEngineXML::_readMessage",
		      ERROR_LEVEL_INTERNAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_ABORT,
		      "Connection lost");
	result+=tmp;
	size-=tmp.length();
      }
      if (difftime(time(0), startTime)>_msgTimeout) {
	if (Hbci::debugLevel()>0)
          fprintf(stderr,
                  "Timeout after reading message header\n");
        throw Error("MessageEngineXML::_readMessage",
                    ERROR_LEVEL_INTERNAL,
                    HBCI_ERROR_CODE_UNKNOWN,
                    ERROR_ADVISE_ABORT,
                    "Message timeout (after receiving header)");
      }
    } // while
    if (size>0)
      throw Error("MessageEngineXML::_readMessage",
		  ERROR_LEVEL_INTERNAL,
		  HBCI_ERROR_CODE_UNKNOWN,
		  ERROR_ADVISE_ABORT,
		  "Incomplete message received, user aborted");

    DBG_DEBUG(0, "Returned message size is: %d", result.length());
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo) {
      DBG_INFO(0, "Hash of message read (%d bytes) is: %s",
               result.length(),
               _hash(result).c_str());
    }
    return result;
  }


  Error MessageEngineXML::_prepareSignatures(GWEN_DB_NODE *cfg) {
    GWEN_DB_NODE *gr;
    unsigned int dataStart;
    unsigned int dataLength;

    // find first signature head
    DBG_DEBUG(0, "Looking for first signature head");
    gr=GWEN_DB_GetFirstGroup(cfg);
    while (gr) {
      if (strcasecmp(GWEN_DB_GroupName(gr), "SigHead")==0)
        break;
      gr=GWEN_DB_GetNextGroup(gr);
    }
    if (!gr) {
      // no signatures
      DBG_DEBUG(0, "No signatures found");
      return Error();
    }

    // find first non-signhead
    DBG_DEBUG(0, "Looking for first non-signature-head");
    gr=GWEN_DB_GetNextGroup(gr);
    while (gr) {
      if (strcasecmp(GWEN_DB_GroupName(gr), "SigHead")!=0)
        break;
      gr=GWEN_DB_GetNextGroup(gr);
    }
    if (!gr) {
      // all-signatures, error
      DBG_DEBUG(0, "Only signature heads ?!");
      return Error("MessageEngineXML::_prepareSignatures",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_BAD_MESSAGE,
                   ERROR_ADVISE_DONTKNOW,
                   "only signature heads");
    }
    // found begin of signed data
    dataStart=GWEN_DB_GetIntValue(gr, "segment/pos",0,0);
    dataLength=0;

    DBG_DEBUG(0, "Found start of signed data at %d", dataStart);

    // find first signature tail
    DBG_DEBUG(0, "Looking for first signature tail");
    while (gr) {
      if (strcasecmp(GWEN_DB_GroupName(gr), "SigTail")==0)
        break;
      gr=GWEN_DB_GetNextGroup(gr);
    }
    if (!gr) {
      DBG_DEBUG(0, "No signature tail ?");
      return Error("MessageEngineXML::_prepareSignatures",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_BAD_MESSAGE,
                   ERROR_ADVISE_DONTKNOW,
                   "no signature tail");
    }

    // store data area in segment heads
    dataLength=GWEN_DB_GetIntValue(gr, "segment/pos",0,0)-dataStart;
    DBG_DEBUG(0, "Storing signature data pos in signature heads (length=%d)",
             dataLength);
    gr=GWEN_DB_GetFirstGroup(cfg);
    while (gr) {
      if (strcasecmp(GWEN_DB_GroupName(gr), "SigHead")==0) {
        DBG_DEBUG(0, "Storing...");
        GWEN_DB_SetIntValue(gr, GWEN_DB_FLAGS_DEFAULT,
                            "datastart",
                            dataStart);
        GWEN_DB_SetIntValue(gr, GWEN_DB_FLAGS_DEFAULT,
                            "datalength",
                            dataLength);
      } /* if sighead */
      gr=GWEN_DB_GetNextGroup(gr);
    }
    DBG_DEBUG(0, "Stored signature data pos in signature heads");

    return Error();
  }


  GWEN_DB_NODE *MessageEngineXML::_decode(Pointer<Customer> owner,
                                          string data) {
    unsigned int pos;
    GWEN_DB_NODE *cfg;
    GWEN_DB_NODE *gr;
    int rv;
    Error err;
    string origdata;

    origdata=data;
    cfg=GWEN_DB_Group_new("MessageEngineXML::_decode:cfg");

    if (data.length()) {
      GWEN_BUFFER *gbuf;
      gbuf=GWEN_Buffer_new((char*)data.data(),
                           data.length(),
                           data.length(),
                           0);

      rv=GWEN_MsgEngine_ReadMessage(_engine,
                                    "SEG",
                                    gbuf,
                                    cfg,
                                    GWEN_MSGENGINE_READ_FLAGS_DEFAULT);
      GWEN_Buffer_free(gbuf);
      if (rv) {
        GWEN_DB_Group_free(cfg);
        throw Error("MessageEngineXML::_decode",
                    ERROR_LEVEL_INTERNAL,
                    HBCI_ERROR_CODE_UNKNOWN,
                    ERROR_ADVISE_ABORT,
                    "Could not parse message");
      }
    }

    // check for encryption
    if (GWEN_DB_GetGroup(cfg, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "CryptHead") &&
	GWEN_DB_GetGroup(cfg, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "CryptData")) {
      string crypteddata;
      string cryptedkey;
      string msgkey;
      GWEN_BUFFER *gbuf;

      // we found encrypted data
      hbci()->interactor().ref().actionStarted(ACT_DECRYPTMSG,
					       "Decrypting message");
      DBG_INFO(0, "Found encrypted data");
      _logMessage(origdata.data(), origdata.length(), true, true);
      cryptedkey=_getBinProperty(cfg, "CryptHead/CryptAlgo/MsgKey", "");
      if (cryptedkey.empty()) {
	GWEN_DB_Group_free(cfg);
	throw Error("MessageEngineXML::_decode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "No message key found");
      }
      crypteddata=_getBinProperty(cfg, "CryptData/CryptData", "");
      if (crypteddata.empty()) {
	GWEN_DB_Group_free(cfg);
	throw Error("MessageEngineXML::_decode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "No crypt data found");
      }

      // mount corresponding medium
      DBG_INFO(0, "Mounting medium for decrypting");
      err=_mountCustomersMedium(owner);
      if (!err.isOk())
	throw Error("MessageEngineXML::_decode", err);

      // decrypt the key and store it for later use
      msgkey=_medium.ref().decryptKey(cryptedkey);

      // decrypt message
      data=_medium.ref().decrypt(crypteddata,msgkey);

      _logMessage(data.data(), data.length(), false, true);

      // decode the encrypted part
      pos=0;
      gbuf=GWEN_Buffer_new((char*)data.data(),
                           data.length(),
                           data.length(),
                           0);

      rv=GWEN_MsgEngine_ReadMessage(_engine,
                                    "SEG",
                                    gbuf,
                                    cfg,
                                    GWEN_MSGENGINE_READ_FLAGS_DEFAULT);
      GWEN_Buffer_free(gbuf);
      if (rv) {
	GWEN_DB_Group_free(cfg);
	throw Error("MessageEngineXML::_decode",
		    ERROR_LEVEL_INTERNAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_ABORT,
		    "Could not parse decrypted message");
      }
    } // if crypted head/tail found
    else {
      _logMessage(origdata.data(), origdata.length(), false, true);
    }
    // prepare signature check
    err=_prepareSignatures(cfg);
    if (!err.isOk())
      throw Error("MessageEngineXML::_decode", err);

    // check signatures
    gr=GWEN_DB_GetFirstGroup(cfg);
    while (gr) {
      if (strcasecmp(GWEN_DB_GroupName(gr), "SigHead")==0) {
	string toVerify;
	string sdata;
	string signature;
	string ctrlref;
	GWEN_DB_NODE *sigTail;

        // TODO: check signature type (only current mode is supported!)
	DBG_INFO(0,"Verifying a signature (%s)", GWEN_DB_GroupName(gr));
        ctrlref=GWEN_DB_GetCharValue(gr, "ctrlref",0, "");
        sigTail=GWEN_DB_GetFirstGroup(cfg);
        DBG_INFO(0, "Searching for SigTail");
	while(sigTail) {
          if (strcasecmp(GWEN_DB_GroupName(sigTail), "SigTail")==0) {
            // have a signature tail, compare ctrlref
            if (GWEN_DB_GetCharValue(sigTail, "ctrlref",0, "")==ctrlref)
              break;
          }
          sigTail=GWEN_DB_GetNextGroup(sigTail);
        } // while

	if (!sigTail) {
	  DBG_ERROR(0,"Found a signature head without tail !");
	  throw Error("MessageEngineXML::_decode",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_ABORT,
		      "No signature tail found");
	}
        DBG_INFO(0, "Found SigTail");

	// signature head found, tail is available, too
	signature=_getBinProperty(sigTail, "signature","");
	if (signature.empty()) {
	  DBG_ERROR(0,"Signature missing");
	  throw Error("MessageEngineXML::_decode",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_ABORT,
		      "Signature missing");
	}

	// ok, we have all, now get the data to verify
        DBG_INFO(0, "Compiling data for signature check");
	toVerify=data.substr(GWEN_DB_GetIntValue(gr, "segment/pos",0,0),
			     GWEN_DB_GetIntValue(gr, "segment/length",0,0));
        assert(!toVerify.empty());
	sdata=data.substr(GWEN_DB_GetIntValue(gr, "datastart",0,0),
                          GWEN_DB_GetIntValue(gr, "datalength",0,0));
        assert(!sdata.empty());
        toVerify+=sdata;

	/* if this is RDH-mode, we might have recieved a new key with
	 * this message. if this is the first time, we can overstep
	 * the verification because it does not make much sense;-)
	 * the verification is done by hand with the ini-letter
	 * otherwise, we perform the verification! */
        DBG_INFO(0, "Checking if signature check is to be performed");
        DBG_INFO(0, "Mounting medium for signature verification");
        err=_mountCustomersMedium(owner);
        if (!err.isOk())
          throw Error("MessageEngineXML::_decode", err);

	if ((HBCI_SECURITY_RDH ==_medium.ref().securityMode()) &&
	    !(_medium.cast<MediumRDHBase>().ref().hasInstSignKey())) {
	  DBG_WARN(0, "Not performing verification, we have no sign key");
	}
	else {
	  // do the verification
          DBG_INFO(0, "Really verifying signature");
          hbci()->interactor().ref().actionStarted(ACT_VERIFYMSG,
						   "Verifying message");
	  err=_medium.ref().verify(toVerify,signature);
	  if (!err.isOk()) {
            DBG_ERROR(0,"Signature is invalid");
	    throw Error("MessageEngineXML::_decode", err);
	  }
          DBG_INFO(0, "Signature is valid");
	} // should we do the verification?
      } // if SigHead
      gr=GWEN_DB_GetNextGroup(gr);
    } // while

    return cfg;
  }



  GWEN_DB_NODE *MessageEngineXML::_openDialog(){
    Pointer<MsgJobXML> j;
    Pointer<User> u;
    int rv;
    GWEN_XMLNODE *node;
    GWEN_DB_NODE *jobData;
    Error err;
    list<Pointer<Customer> > signers;
    GWEN_DB_NODE *response;
    GWEN_DB_NODE *custcfg;
    GWEN_BUFFER *gbuf;

    DBG_INFO(0, "Opening dialog for \"%s\"",
             _sessionOwner.ref().custId().c_str());
    GWEN_MsgEngine_SetValue(_engine,"DialogId", "0");
    custcfg=GWEN_DB_GetGroup(_config, GWEN_DB_FLAGS_DEFAULT,
                             _sessionOwner.ref().custId().c_str());
    DBG_INFO(0, "Creating JobDialogInit");
    u=_sessionOwner.ref().user();
    j=_jobFactory("JobDialogInit",_sessionOwner);
    GWEN_MsgEngine_SetValue(_engine,"product",
                            hbci()->systemName().c_str());
    GWEN_MsgEngine_SetValue(_engine,"productVersion",
                            hbci()->systemVersion().c_str());
    signers.push_back(_sessionOwner);

    GWEN_MsgEngine_SetIntValue(_engine, "MessageNumber", 1);
    GWEN_MsgEngine_SetIntValue(_engine, "SegmentNumber", signers.size()+2);

    DBG_INFO(0, "Opening dialog");
    node=j.ref().node();
    jobData=j.ref().jobData();
    GWEN_MsgEngine_SetMode(_engine, j.ref().securityMode().c_str());
    j.ref().setFirstSegment(1);
    gbuf=GWEN_Buffer_new(0, MSGXML_MAXMSGSIZE, 0, 1);
    hbci()->interactor().ref().actionStarted(ACT_CREATEJOB,
					     "Creating job");
    rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
					    node,
					    gbuf,
					    jobData);
    if (rv) {
      GWEN_Buffer_free(gbuf);
      throw Error("MsgJobXML::_openDialog",
                  ERROR_LEVEL_NORMAL,
                  HBCI_ERROR_CODE_UNKNOWN,
                  ERROR_ADVISE_DONTKNOW,
		  "Error creating message",
		  j.ref().name());
    }
    j.ref().setLastSegment(GWEN_MsgEngine_GetIntValue(_engine,
                                                      "SegmentNumber",
                                                      1)-1);
    try {
      response=_exchangeMessages(_sessionTransporter, _sessionOwner, signers,
                                 GWEN_Buffer_GetStart(gbuf),
                                 GWEN_Buffer_GetUsedBytes(gbuf),
                                 true, true);
    }
    catch(Error xerr) {
      DBG_ERROR(0,"Could not handle jobs");
      throw Error("MessageEngineXML::_openDialog", xerr);
    }
    GWEN_Buffer_free(gbuf);
    return response;
  }


  Error MessageEngineXML::_closeDialog(){
    Pointer<MsgJobXML> j;
    Pointer<User> u;
    int rv;
    GWEN_XMLNODE *node;
    GWEN_DB_NODE *jobData;
    Error err;
    list<Pointer<Customer> > signers;
    GWEN_DB_NODE *response;
    GWEN_BUFFER *gbuf;

    DBG_INFO(0, "Closing dialog for \"%s\"",
             _sessionOwner.ref().custId().c_str());
    DBG_INFO(0, "Creating JobDialogEnd");
    u=_sessionOwner.ref().user();
    j=_jobFactory("JobDialogEnd",_sessionOwner);
    signers.push_back(_sessionOwner);

    GWEN_MsgEngine_SetIntValue(_engine, "SegmentNumber", signers.size()+2);

    DBG_INFO(0, "Closing dialog");
    node=j.ref().node();
    jobData=j.ref().jobData();
    GWEN_MsgEngine_SetMode(_engine, j.ref().securityMode().c_str());
    j.ref().setFirstSegment(1);
    gbuf=GWEN_Buffer_new(0, MSGXML_MAXMSGSIZE, 0, 1);
    hbci()->interactor().ref().actionStarted(ACT_CREATEJOB,
					     "Creating job");
    rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
                                            node,
                                            gbuf,
                                            jobData);
    if (rv) {
      GWEN_Buffer_free(gbuf);
      return Error("MsgJobXML::_closeDialog",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error creating message",
		   j.ref().name());
    }
    j.ref().setLastSegment(GWEN_MsgEngine_GetIntValue(_engine,
                                                      "SegmentNumber",
                                                      1)-1);
    try {
      response=_exchangeMessages(_sessionTransporter, _sessionOwner, signers,
                                 GWEN_Buffer_GetStart(gbuf),
                                 GWEN_Buffer_GetUsedBytes(gbuf),
                                 true, true);
    }
    catch (Error xerr) {
      GWEN_Buffer_free(gbuf);
      return Error("MsgJobXML::_closeDialog", xerr);
    }

    // TODO: Evaluate and commit to bank

    GWEN_MsgEngine_SetValue(_engine,"DialogId", "0");
    GWEN_DB_Group_free(response);
    GWEN_Buffer_free(gbuf);
    return err;
  }


  GWEN_DB_NODE *
  MessageEngineXML::_handleJobs(Pointer<Transporter> tx,
				Pointer<Customer> owner,
				list<Pointer<Customer> > signers,
				list<Pointer<MsgJobXML> > jobs) {
    list<Pointer<MsgJobXML> >::const_iterator jit;
    Error err;
    bool needCrypt;
    bool needSign;
    GWEN_DB_NODE *response;
    GWEN_BUFFER *gbuf;

    needCrypt=false;
    needSign=false;

    // check whether these jobs need to be signed and/or crypted
    for (jit=jobs.begin();
	 jit!=jobs.end();
	 jit++) {
      needCrypt|=(*jit).ref().needsEncryption();
      needSign|=(*jit).ref().needsSignature();
    } // for

    if (needSign)
      GWEN_MsgEngine_SetIntValue(_engine, "SegmentNumber", signers.size()+2);
    else
      GWEN_MsgEngine_SetIntValue(_engine, "SegmentNumber", 2);

    DBG_INFO(0, "Handling %d jobs", jobs.size());

    gbuf=GWEN_Buffer_new(0, MSGXML_MAXMSGSIZE, 0, 1);
    for (jit=jobs.begin();
	 jit!=jobs.end();
	 jit++) {
      int rv;
      GWEN_XMLNODE *node;
      GWEN_DB_NODE *jobData;
      Pointer<MsgJobXML> job;

      job=*jit;

      DBG_INFO(0, "Handling job \"%s\"", job.ref().name().c_str());
      hbci()->interactor().ref().actionStarted(ACT_CREATEJOB,
					       "Creating job");
      node=job.ref().node();
      jobData=job.ref().jobData();
      assert(jobData);
      jobData=GWEN_DB_GetGroup(jobData, GWEN_DB_FLAGS_DEFAULT, "request");
      GWEN_MsgEngine_SetMode(_engine, job.ref().securityMode().c_str());
      job.ref().setFirstSegment(GWEN_MsgEngine_GetIntValue(_engine,
                                                           "SegmentNumber",
                                                           1));
      rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
                                              node,
                                              gbuf,
                                              jobData);
      if (rv) {
        GWEN_Buffer_free(gbuf);
	throw Error("MsgJobXML::_handleJobs",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error creating message",
		    job.ref().name());
      }
      job.ref().setLastSegment(GWEN_MsgEngine_GetIntValue(_engine,
                                                          "SegmentNumber",
							  1)-1);
    } // for

    try {
      response=_exchangeMessages(tx, owner, signers,
                                 GWEN_Buffer_GetStart(gbuf),
                                 GWEN_Buffer_GetUsedBytes(gbuf),
                                 needSign, needCrypt);
    }
    catch(Error xerr) {
      DBG_ERROR(0,"Could not handle jobs");
      throw Error("MessageEngineXML::_handleJobs", xerr);
    }
    GWEN_Buffer_free(gbuf);
    return response;
  }



  GWEN_DB_NODE *MessageEngineXML::_handleDlgJob(Pointer<MsgJobXML> job,
                                                GWEN_XMLNODE *node,
                                                GWEN_DB_NODE *jobData) {
    /* this job has its own dialog */
    GWEN_XMLNODE *msgtag;
    list<Pointer<Customer> > signers;
    Error err;
    GWEN_DB_NODE *response;
    GWEN_BUFFER *gbuf;
    int msgs;
    Pointer<User> u;

    // DEBUG
    //DBG_ERROR(0, "Handling a dialog job");
    //GWEN_DB_Dump(jobData, stderr, 2);

    if (_dialogOpen) {
      // first close dialog
      hbci()->interactor().ref().actionStarted(ACT_CLOSEDIALOG,
					       "Closing dialog");
      err=_closeDialog();
      if (!err.isOk()) {
	DBG_WARN(0,"Could not close dialog (%s)",
		 err.errorString().c_str());
      }
      _dialogOpen=false;
    }
    u=_sessionOwner.ref().user();

    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo)
      GWEN_DB_Dump(jobData, stderr, 2);
    jobData=GWEN_DB_GetGroup(jobData, GWEN_DB_FLAGS_DEFAULT, "request");
    GWEN_MsgEngine_SetValue(_engine,"DialogId", "0");
    GWEN_MsgEngine_SetIntValue(_engine, "MessageNumber", 1);
    GWEN_MsgEngine_SetValue(_engine,"product", hbci()->systemName().c_str());
    GWEN_MsgEngine_SetValue(_engine,"productVersion",
                            hbci()->systemVersion().c_str());

    signers=job.ref().signers();
    if (signers.empty())
      signers.push_back(_sessionOwner);

    msgtag=GWEN_XMLNode_GetChild(node);
    gbuf=GWEN_Buffer_new(0, MSGXML_MAXMSGSIZE, 0, 1);
    response=GWEN_DB_Group_new("response");
    msgs=0;
    while(msgtag) {
      /* this job has its own messages, handle them */
      const char *pname;
      pname=GWEN_XMLNode_GetData(msgtag);
      if (pname) {
        if (strcasecmp(pname, "MESSAGE")==0) {
          /* found a message */
          int rv;
          GWEN_DB_NODE *lresponse;
          GWEN_DB_NODE *msgData;
          bool needCrypt;
          bool needSign;
          string msgName;

	  msgName=GWEN_XMLNode_GetProperty(msgtag, "name", "");
          if (msgName.empty())
            msgData=jobData;
          else {
            msgData=GWEN_DB_GetGroup(jobData, GWEN_DB_FLAGS_DEFAULT,
                                     msgName.c_str());
          }

          // start new message
          needCrypt=atoi(GWEN_XMLNode_GetProperty(msgtag, "crypt", "1"));
          needSign=atoi(GWEN_XMLNode_GetProperty(msgtag, "sign", "1"));
          GWEN_Buffer_Reset(gbuf);
          if (needSign) {
            GWEN_MsgEngine_SetIntValue(_engine,
                                       "SegmentNumber",
                                       signers.size()+2);
          }
          else
            GWEN_MsgEngine_SetIntValue(_engine, "SegmentNumber", 2);

          msgs++;
          DBG_INFO(0, "Handling message %d", msgs);

          job.ref().
            setFirstSegment(GWEN_MsgEngine_GetIntValue(_engine,
                                                       "SegmentNumber",
                                                       0));
          hbci()->interactor().ref().actionStarted(ACT_CREATEJOB,
						   "Creating job");
          rv=GWEN_MsgEngine_CreateMessageFromNode(_engine,
                                                  msgtag,
                                                  gbuf,
                                                  msgData);
	  if (rv) {
	    GWEN_DB_Group_free(response);
	    GWEN_Buffer_free(gbuf);
            throw Error("MsgJobXML::_handleDlgJob",
                        ERROR_LEVEL_NORMAL,
                        HBCI_ERROR_CODE_UNKNOWN,
                        ERROR_ADVISE_DONTKNOW,
                        "Error creating message",
                        job.ref().name());
            job.ref().
              setLastSegment(GWEN_MsgEngine_GetIntValue(_engine,
                                                        "SegmentNumber",
                                                        1)-1);
          }
          DBG_DEBUG(0, "Buffer is now:");
          if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
            GWEN_Buffer_Dump(gbuf, stderr, 2);

          // flush msg
          try {
            lresponse=_exchangeMessages(_sessionTransporter,
                                        _sessionOwner,
                                        signers,
                                        GWEN_Buffer_GetStart(gbuf),
                                        GWEN_Buffer_GetUsedBytes(gbuf),
                                        needSign,
                                        needCrypt);
          }
          catch(Error xerr) {
            DBG_ERROR(0,"Could not handle jobs");
            throw Error("MessageEngineXML::_handleDlgJob", xerr);
          }
          GWEN_DB_AddGroupChildren(response, lresponse);
          GWEN_DB_Group_free(lresponse);
        } // if <MESSAGE>
      }
      msgtag=GWEN_XMLNode_Next(msgtag);
    } /* while child */
    if (!msgs) {
      DBG_ERROR(0, "No messages for this job in XML file");
      return 0;
    }
    GWEN_Buffer_free(gbuf);
    GWEN_MsgEngine_SetValue(_engine,"DialogId", "0");
    return response;
  }


  bool
  MessageEngineXML::_compareSignatures(const list<Pointer<Customer> > &s,
				       const list<Pointer<Customer> > &s2){
    list<Pointer<Customer> >::const_iterator it1;
    list<Pointer<Customer> >::const_iterator it2;
    unsigned int matches;

    // s=signers
    // s2=job-signers

    if (s.size()!=s2.size())
      return false;

    matches=0;
    for (it1=s2.begin();
	 it1!=s2.end();
	 it1++) {
      for (it2=s.begin();
	   it2!=s.end();
	   it2++) {
	if (*it1==*it2) {
	  // job signer found in signers
	  matches++;
	  break;
	}
      } // for signers
    } // for job-signers
    return (matches==s.size());
  }


  Error
  MessageEngineXML::_openSendClose(Pointer<MsgQueueXML> q,
				   const list<Pointer<Customer> > &signers,
				   list<Pointer<MsgJobXML> > &jobs) {
    GWEN_DB_NODE *response;
    Error err;
    list<Pointer<MsgJobXML> > tmplist;

    if (!_dialogOpen) {
      // really open dialog
      hbci()->interactor().ref().actionStarted(ACT_OPENDIALOG,
					       "Opening dialog");
      DBG_INFO(0, "Opening dialog");
      try {
	response=_openDialog();
	_dialogOpen=true;
      }
      catch (Error xerr) {
        DBG_ERROR(0,"Could not open dialog");
        return Error("MessageEngineXML::_openSendClose", xerr);
      }

      // handle responses from dialog opening
      hbci()->interactor().ref().actionStarted(ACT_PROCESSMSG,
					       "Processing response");
      err=_processResponse(q, tmplist, response);
      if (!err.isOk()) {
        DBG_ERROR(0,"Could not handle response to open dialog");
        return Error("MessageEngineXML::_openSendClose", err);
      }
      if (q.ref().results().empty()) {
        DBG_ERROR(0,"No result in queue");
        return Error("MessageEngineXML::openSendClose",
                     ERROR_LEVEL_NORMAL,
                     HBCI_ERROR_CODE_UNKNOWN,
                     ERROR_ADVISE_DONTKNOW,
                     "No result in queue");
      }
      else if (q.ref().results().front().resultCode()>=9000) {
        DBG_ERROR(0,"Error flagged !");
        return Error("MessageEngineXML::openSendClose",
                     ERROR_LEVEL_NORMAL,
                     HBCI_ERROR_CODE_UNKNOWN,
                     ERROR_ADVISE_DONTKNOW,
                     "Error in queue");
      }
      else {
        DBG_INFO(0,"Result is: %d",
                 q.ref().results().front().resultCode());
      }
    }
    else {
      DBG_INFO(0, "Dialog already open");
    }

    // handle jobs so far
    hbci()->interactor().ref().actionStarted(ACT_HANDLEJOBS,
					     "Processing response");
    try {
      DBG_INFO(0, "Handling jobs");
      response=_handleJobs(_sessionTransporter, _sessionOwner, signers, jobs);
    }
    catch(Error xerr) {
      DBG_ERROR(0,"Could not handle jobs");
      return Error("MessageEngineXML::_openSendClose", xerr);
    }

    // handle responses from jobs
    hbci()->interactor().ref().actionStarted(ACT_PROCESSMSG,
					     "Processing response");
    err=_processResponse(q, jobs, response);
    if (!err.isOk()) {
      DBG_ERROR(0,"Could not handle response to jobs");
      return Error("MessageEngineXML::_openSendClose", err);
    }

    DBG_INFO(0, "Joblist handled");
    return Error();
  }


  Error MessageEngineXML::_handleQueue(Pointer<MsgQueueXML> q) {
    list<Pointer<MsgJob> >::const_iterator it;
    Error err;
    list<Pointer<MsgJobXML> > jobs;
    list<Pointer<Customer> > signers;
    bool first;

    try {
      it=q.ref().jobs().begin();
      first=true;
      while(it!=q.ref().jobs().end()) {
        GWEN_XMLNODE *node;
        GWEN_DB_NODE *jobData;
        Pointer<MsgJobXML> job;
        bool isDlgJob;
        bool flush;
        bool next;

        isDlgJob=false;
        flush=false;
        next=false;

        /* FIXME: switching between multiple secmodes should not be done.
         * At least we should not intermix jobs which need different secmodes
         * (secmode="DDV" or "RDH")
         */
        job=(*it).cast<MsgJobXML>();
        DBG_INFO(0, "Handling job \"%s\"", job.ref().name().c_str());
        GWEN_MsgEngine_SetMode(_engine, job.ref().securityMode().c_str());
        node=job.ref().node();
        if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo)
          GWEN_XMLNode_Dump(node, stderr, 1);
        jobData=job.ref().jobData();

        if (atoi(GWEN_XMLNode_GetProperty(node, "dlg", "0"))!=0) {
          DBG_INFO(0, "Is a dialog job");
          isDlgJob=true;
          flush=true;
          next=false;
        }
        else {
          list<Pointer<Customer> > jobSigners;

          DBG_INFO(0, "Is not a dialog job");
          // not a dlg job, check whether the current job fits into the list
          isDlgJob=false;

          jobSigners=job.ref().signers();
          if (jobSigners.empty()) {
            // no job signer, so add the owner of the msg queue
            DBG_INFO(0, "Adding queue owner as signer tojob \"%s\"",
                     job.ref().name().c_str());
            jobSigners.push_back(q.ref().owner());
          }

          // check signatures
          if (first) {
            list<Pointer<Customer> >::iterator it3;

            DBG_INFO(0, "First job \"%s\"", job.ref().name().c_str());
            // empty signer list, so add those from the current job
            for (it3=jobSigners.begin();
                 it3!=jobSigners.end();
                 it3++)
              signers.push_back(*it3);
            DBG_INFO(0, "Adding first job \"%s\"", job.ref().name().c_str());
            jobs.push_back(job);
            next=true;
          } // if signers empty
          else {
            /* there already are signatures, so check whether they are the same
             * as those provided with this job. This is to avoid letting a
             * secondary signer sign other jobs than that they really
             * wanted to sign
             */
            DBG_INFO(0, "Have signers for job \"%s\"", job.ref().name().c_str());
            if (_compareSignatures(signers, jobSigners)) {
              // add job
              DBG_INFO(0, "Adding job \"%s\"", job.ref().name().c_str());
              jobs.push_back(job);
              next=true;
            }
            else {
              /* other signatures, so flush the current list and add this job
               * to the next list */
              flush=true;
              next=false;
            }
          } // if signers !empty
        } // if not a dialog job

        if (!isDlgJob) {
          if (flush) {
            err=_openSendClose(q, signers, jobs);
            if (!err.isOk()) {
              DBG_ERROR(0,"Could not send jobs");
              return Error("MessageEngineXML::_handleQueue", err);
            }

            signers.clear();
            jobs.clear();
            first=true;
          }
          else {
            first=false;
          }
        }
        else {
          // handle DLG job
          list<Pointer<MsgJobXML> > tmpl;
          GWEN_DB_NODE *response;

          response=_handleDlgJob(job, node, jobData);
          tmpl.push_back(job);
          hbci()->interactor().ref().actionStarted(ACT_PROCESSMSG,
						   "Processing response");
	  err=_processResponse(q, tmpl, response);
          if (!err.isOk()) {
            DBG_ERROR(0,"Could not process response for job");
            return Error("MessageEngineXML::_handleQueue", err);
          }
          next=true;
        } // if dlg job

        if (next)
          it++;
      } // while

      if (!jobs.empty()) {
        // handle remaining jobs
        DBG_INFO(0, "Handling remaining %d jobs", jobs.size());
        err=_openSendClose(q, signers, jobs);
        if (!err.isOk()) {
          DBG_ERROR(0,"Could not send jobs");
          return Error("MessageEngineXML::_handleQueue", err);
        }
      }
    } // try
    catch (Error xerr) {
      DBG_ERROR(0,"Could not handle queue");
      return Error("MessageEngineXML::_handleQueue", xerr);
    }
    return Error();
  }






  int MessageEngineXML::_toDTA(int c) {
    if (isdigit(c))
      return c;
    if (c>='A' && c<='Z')
      return c;
    if (c>='a' && c<='z')
      return toupper(c);
    if (strchr(" .,&-+*%/$", c))
      return c;
    switch(c) {
    case '': return 0x5b;
    case '': return 0x5c;
    case '': return 0x5d;
    case '': return 0x7e;
    default:
      break;
    } /* switch */
    return 0;
  }


  int MessageEngineXML::_fromDTA(int c) {
    if (isdigit(c))
      return c;
    if (c>='A' && c<='Z')
      return c;
    if (c>='a' && c<='z')
      return toupper(c);
    if (strchr(" .,&-+*%/$", c))
      return c;
    switch(c) {
    case 0x5b: return '';
    case 0x5c: return '';
    case 0x5d: return '';
    case 0x7e: return '';
    default:
      break;
    } /* switch */
    return 0;
  }



  int MessageEngineXML::typeRead(GWEN_BUFFER *msgbuf,
                                 GWEN_XMLNODE *node,
                                 GWEN_BUFFER *vbuf,
                                 char escapeChar,
                                 const char *delimiters){
    const char *type;
    int v;

    type=GWEN_XMLNode_GetProperty(node, "type","");
    if (strcasecmp(type, "date")==0) {
      char buffer[9];
      unsigned int i;

      DBG_DEBUG(0, "Supporting type \"date\"");
      memset(buffer,0, sizeof(buffer));
      if (GWEN_Buffer_GetBytesLeft(msgbuf)<8) {
        DBG_ERROR(0, "Too few bytes for date (%d<8)",
                  GWEN_Buffer_GetBytesLeft(msgbuf));
        return -1;
      }
      for (i=0; i<8; i++) {
        int c;

        c=GWEN_Buffer_ReadByte(msgbuf);
        if (c==-1)
          return -1;
        if (!isdigit(c)) {
          DBG_ERROR(0, "Non-digit found in date");
          return -1;
        }
        buffer[i]=c;
        if (i==3) {
          /* year is full, check it */
          v=atoi(buffer);
          if (v<1970 || v>2100) {
            DBG_ERROR(0, "Year out of range (%d)", v);
            return -1;
          }
        } /* if i==3 */
        else if (i==5) {
          /* month is full, check it */
          v=atoi(buffer+4);
          if (v<1 || v>12) {
            DBG_ERROR(0, "Month out of range (%d)", v);
            return -1;
          }
        } /* if i==6 */
      }
      v=atoi(buffer+6);
      if (v<1 || v>31) {
        DBG_ERROR(0, "Day out of range (%d)", v);
        return -1;
      }
      /* date read, now store it */
      if (GWEN_Buffer_AppendBytes(vbuf, buffer, 8)) {
        DBG_INFO(0, "called from here");
        return -1;
      }
      return 0;
    } // if date

    else if (strcasecmp(type, "time")==0) {
      char buffer[7];
      unsigned int i;

      DBG_INFO(0, "Supporting type \"time\"");
      memset(buffer,0, sizeof(buffer));
      if (GWEN_Buffer_GetBytesLeft(msgbuf)<6) {
        DBG_ERROR(0, "Too few bytes for time (%d<6)",
                  GWEN_Buffer_GetBytesLeft(msgbuf));
        return -1;
      }
      for (i=0; i<6; i++) {
        int c;

        c=GWEN_Buffer_ReadByte(msgbuf);
        if (c==-1)
          return -1;
        if (!isdigit(c)) {
          DBG_ERROR(0, "Non-digit found in date");
          return -1;
        }
        buffer[i]=c;
        if (i==1) {
          /* hour is full, check it */
          v=atoi(buffer);
          if (v>23) {
            DBG_ERROR(0, "Hour out of range (%d)", v);
            return -1;
          }
        } /* if i==1 */
        else if (i==3) {
          /* minutes full, check */
          v=atoi(buffer+2);
          if (v>59) {
            DBG_ERROR(0, "Minutes out of range (%d)", v);
            return -1;
          }
        } /* if i==3 */
      }
      v=atoi(buffer+4);
      if (v>59) {
        DBG_ERROR(0, "Seconds out of range (%d)", v);
        return -1;
      }
      /* date read, now store it */
      if (GWEN_Buffer_AppendBytes(vbuf, buffer, 6)) {
        DBG_INFO(0, "called from here");
        return -1;
      }
      return 0;
    }
    else if (strcasecmp(type, "dta")==0) {
      int lastWasEscape;
      int isEscaped;

      DBG_INFO(0, "Supporting type \"dta\"");
      isEscaped=0;
      lastWasEscape=0;

      while(GWEN_Buffer_GetBytesLeft(msgbuf)) {
        int c;

        c=GWEN_Buffer_ReadByte(msgbuf);
        if (lastWasEscape) {
          lastWasEscape=0;
          isEscaped=1;
        }
        else {
          isEscaped=0;
          if (c==escapeChar) {
            lastWasEscape=1;
            c=-1;
          }
        }
        if (c!=-1) {
          if (!isEscaped && strchr(delimiters, c)!=0) {
            /* delimiter found, step back */
            GWEN_Buffer_DecrementPos(msgbuf,1);
            break;
          }
          else {
            c=_fromDTA(c);
            if (!c) {
              DBG_ERROR(0, "Found a non-DTA character");
              return -1;
            }
            if (GWEN_Buffer_AppendByte(vbuf, c)) {
              DBG_DEBUG(0, "called from here");
              return -1;
            }
          }
        }
      } /* while */

      /* append a 0 character, but do not count it. This is just for my
       * conceniance, so that I may fprintf the result */
      if (GWEN_Buffer_AllocRoom(vbuf,1)) {
	DBG_INFO(0, "Value buffer full.");
	return -1;
      }
      *(GWEN_Buffer_GetPosPointer(vbuf))=0;
      return 0;
    }
    else {
      DBG_DEBUG(0, "Type \"%s\" not supported by MsgEngineHBCI", type);
      return 1;
    }

    /* I know, we'll never get here ;-) */
    return 1;
  }



  int MessageEngineXML::typeWrite(GWEN_BUFFER *gbuf,
                                  GWEN_BUFFER *data,
                                  GWEN_XMLNODE *node){
    const char *type;
    int v;

    DBG_VERBOUS(0, "MsgEngineHBCI_TypeWrite");

    type=GWEN_XMLNode_GetProperty(node, "type","");
    if (strcasecmp(type, "date")==0) {
      char buffer[9];
      unsigned int i;

      DBG_DEBUG(0, "Supporting type \"date\"");
      memset(buffer,0, sizeof(buffer));
      if (GWEN_Buffer_GetBytesLeft(data)<8) {
        DBG_ERROR(0, "Too few bytes for date (%d<8)",
                  GWEN_Buffer_GetBytesLeft(data));
        return -1;
      }
      for (i=0; i<8; i++) {
        int c;

        c=GWEN_Buffer_ReadByte(data);
        if (c==-1)
          return -1;
        if (!isdigit(c)) {
          DBG_ERROR(0, "Non-digit found in date");
          return -1;
        }
        buffer[i]=c;
        if (i==3) {
          /* year is full, check it */
          v=atoi(buffer);
          if (v<1970 || v>2100) {
            DBG_ERROR(0, "Year out of range (%d)", v);
            return -1;
          }
        } /* if i==3 */
        else if (i==5) {
          /* month is full, check it */
          v=atoi(buffer+4);
          if (v<1 || v>12) {
            DBG_ERROR(0, "Month out of range (%d)", v);
            return -1;
          }
        } /* if i==5 */
      }
      v=atoi(buffer+6);
      if (v<1 || v>31) {
        DBG_ERROR(0, "Day out of range (%d)", v);
        return -1;
      }
      /* date read, now store it */
      if (GWEN_Buffer_AppendBytes(gbuf, buffer, 8)) {
        DBG_INFO(0, "called from here");
        return -1;
      }
      return 0;
    } // if date

    else if (strcasecmp(type, "time")==0) {
      char buffer[7];
      unsigned int i;

      DBG_DEBUG(0, "Supporting type \"time\"");
      memset(buffer,0, sizeof(buffer));
      if (GWEN_Buffer_GetBytesLeft(data)<6) {
        DBG_ERROR(0, "Too few bytes for time (%d<6)",
                  GWEN_Buffer_GetBytesLeft(data));
        return -1;
      }
      for (i=0; i<6; i++) {
        int c;

        c=GWEN_Buffer_ReadByte(data);
        if (c==-1)
          return -1;
        if (!isdigit(c)) {
          DBG_ERROR(0, "Non-digit found in date");
          return -1;
        }
        buffer[i]=c;
        if (i==1) {
          /* hour is full, check it */
          v=atoi(buffer);
          if (v>23) {
            DBG_ERROR(0, "Hour out of range (%d)", v);
            return -1;
          }
        } /* if i==1 */
        else if (i==3) {
          /* minutes full, check */
          v=atoi(buffer+2);
          if (v>59) {
            DBG_ERROR(0, "Minutes out of range (%d)", v);
            return -1;
          }
        } /* if i==3 */
      }
      v=atoi(buffer+4);
      if (v>59) {
        DBG_ERROR(0, "Seconds out of range (%d)", v);
        return -1;
      }
      /* date read, now store it */
      if (GWEN_Buffer_AppendBytes(gbuf, buffer, 6)) {
        DBG_INFO(0, "called from here");
        return -1;
      }
      return 0;
    }
    else if (strcasecmp(type, "dta")==0) {
      int lastWasEscape;
      int isEscaped;

      DBG_DEBUG(0, "Supporting type \"dta\"");
      isEscaped=0;
      lastWasEscape=0;

      while(GWEN_Buffer_GetBytesLeft(data)) {
        int c;
        const char *charsToEscape;

        charsToEscape=GWEN_MsgEngine_GetCharsToEscape(_engine);
        c=GWEN_Buffer_ReadByte(data);
        c=_toDTA(c);
        if (!c) {
          DBG_ERROR(0, "Found a non-DTA character");
          return -1;
        }
        if (strchr(charsToEscape, c)) {
          /* need to escape this char */
          if (GWEN_Buffer_AppendByte(gbuf,
                                     GWEN_MsgEngine_GetEscapeChar(_engine))){
            DBG_DEBUG(0, "called from here");
            return -1;
          }
        }
        if (GWEN_Buffer_AppendByte(gbuf, c)) {
          DBG_DEBUG(0, "called from here");
          return -1;
        }
      } /* while */

      /* append a 0 character, but do not count it. This is just for my
       * conceniance, so that I may fprintf the result */
      if (GWEN_Buffer_AllocRoom(gbuf, 1)) {
        DBG_INFO(0, "Message buffer full.");
        return -1;
      }
      *(GWEN_Buffer_GetPosPointer(gbuf))=0;
      return 0;
    }
    else {
      DBG_DEBUG(0, "Type \"%s\" not supported by MsgEngineHBCI", type);
      return 1;
    }

    /* I know, we'll never get here ;-) */
    return 1;

  }



  GWEN_DB_VALUETYPE MessageEngineXML::typeCheck(const char *tname) {

    DBG_VERBOUS(0, "MsgEngineHBCI_TypeCheck (%s)", tname);
    if (strcasecmp(tname, "date")==0 ||
        strcasecmp(tname, "time")==0 ||
        strcasecmp(tname, "dta")==0) {
      return GWEN_DB_VALUETYPE_CHAR;
    }
    else
      return GWEN_DB_VALUETYPE_UNKNOWN;
  }



  const char *MessageEngineXML::getCharValue(const char *name,
                                             const char *defValue){
    const char *result;

    result=0;

    if (!_sessionOwner.isValid()) {
      DBG_WARN(0, "No session owner");
      return defValue;
    }
    if (strcasecmp(name, "customerid")==0) {
      result=_sessionOwner.ref().custId().c_str();
    }
    else if (strcasecmp(name, "userid")==0) {
      result=_sessionOwner.ref().user().ref().userId().c_str();
    }
    else if (strcasecmp(name, "bankcode")==0) {
      result=_sessionOwner.ref().user().ref().bank().ref().bankCode().c_str();
    }
    else if (strcasecmp(name, "systemid")==0) {
      result=_sessionOwner.ref().user().ref().systemId().c_str();
    }
    else if (strcasecmp(name, "cryptKeyOwner")==0) {
      Pointer<Medium> m;

      m=_sessionOwner.ref().user().ref().medium();
      if (m.ref().securityMode()==HBCI_SECURITY_RDH) {
        Pointer<MediumRDHBase> mrdh;

        mrdh=m.cast<MediumRDHBase>();
        if (!mrdh.isValid()) {
          DBG_WARN(0, "Could not cast an RDH medium !");
        }
        else {
          result=mrdh.ref().cryptKeyOwner().c_str();
        }
      }
    }
    else if (strcasecmp(name, "date")==0) {
      DateTime ct;
      static char tbuffer[16];

      ct=DateTime::currentTime();
      snprintf(tbuffer, sizeof(tbuffer), "%04d%02d%02d",
	       ct.year(), ct.month(), ct.day());
      return tbuffer;
    }
    else if (strcasecmp(name, "time")==0) {
      DateTime ct;
      static char tbuffer[16];

      ct=DateTime::currentTime();
      snprintf(tbuffer, sizeof(tbuffer), "%02d%02d%02d",
	       ct.hours(), ct.minutes(), ct.seconds());
      return tbuffer;
    }
    else {
      return defValue;
    }

    if (result)
      if (*result) {
        DBG_DEBUG(0, "Returning value for \"%s\"", name);
        return result;
      }

    return defValue;
  }



  int MessageEngineXML::getIntValue(const char *name,
                                    int defValue){

    if (!_sessionOwner.isValid()) {
      DBG_WARN(0, "No session owner");
      return defValue;
    }
    DBG_DEBUG(0, "checking value for \"%s\"", name);

    if (strcasecmp(name, "cryptkeynumber")==0) {
      DBG_DEBUG(0, "Returning value for \"%s\"", name);
      return _sessionOwner.ref().user().ref().medium().ref().cryptKeyNumber();
    }
    else if (strcasecmp(name, "cryptkeyversion")==0) {
      DBG_DEBUG(0, "Returning value for \"%s\"", name);
      return _sessionOwner.ref().user().ref().medium().ref().cryptKeyVersion();
    }
    else if (strcasecmp(name, "signkeynumber")==0) {
      DBG_DEBUG(0, "Returning value for \"%s\"", name);
      return _sessionOwner.ref().user().ref().medium().ref().signKeyNumber();
    }
    else if (strcasecmp(name, "signkeyversion")==0) {
      DBG_DEBUG(0, "Returning value for \"%s\"", name);
      return _sessionOwner.ref().user().ref().medium().ref().signKeyVersion();
    }
    else if (strcasecmp(name, "updversion")==0) {
      GWEN_DB_NODE *updgrp;

      updgrp=GWEN_DB_GetGroup(_config,
                              GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                              _sessionOwner.ref().user().ref().userId().c_str());
      if (updgrp) {
        DBG_DEBUG(0, "Returning value for \"%s\"", name);
        return GWEN_DB_GetIntValue(updgrp,
                                   "version",0,
                                   defValue);
      }
    }
    else if (strcasecmp(name, "bpdversion")==0) {
      DBG_DEBUG(0, "Returning value for \"%s\"", name);
      return GWEN_DB_GetIntValue(_config,
                                 "bpd/version",0, defValue);
    }
    else if (strcasecmp(name, "country")==0) {
      DBG_DEBUG(0, "Returning value for \"%s\"", name);
      return _sessionOwner.ref().user().ref().bank().ref().country();
    }

    return defValue;
  }


  int MessageEngineXML::readBinType(GWEN_XMLNODE *node,
                                    GWEN_DB_NODE *gr,
                                    GWEN_BUFFER *gbuf) {
    Pointer<ParserPlugin> pp;
    const char *type;
    const char *subtype;

    type=GWEN_XMLNode_GetProperty(node, "bintype", "");
    subtype=GWEN_XMLNode_GetProperty(node, "binsubtype", "");

    if (*type==0) {
      DBG_DEBUG(0, "No binary type given");
      return 1;
    }

    try {
      pp=hbci()->getParser(type);
    }
    catch(Error xerr) {
      DBG_ERROR(0,"Error getting parser for \"%s\": %s",
		type, xerr.errorString().c_str());
      return -1;
    }

    return pp.ref().readBinType(node, gr, gbuf);
  }



  int MessageEngineXML::writeBinType(GWEN_XMLNODE *node,
				     GWEN_DB_NODE *gr,
                                     GWEN_BUFFER *gbuf) {
    Pointer<ParserPlugin> pp;
    const char *type;
    const char *subtype;

    type=GWEN_XMLNode_GetProperty(node, "bintype", "");
    subtype=GWEN_XMLNode_GetProperty(node, "binsubtype", "");

    if (*type==0) {
      DBG_DEBUG(0, "No binary type given");
      return 1;
    }

    try {
      pp=hbci()->getParser(type);
    }
    catch(Error xerr) {
      DBG_ERROR(0,"Error getting parser for \"%s\": %s",
		type, xerr.errorString().c_str());
      return -1;
    }

    return pp.ref().writeBinType(node, gr, gbuf);
  }


  Error MessageEngineXML::startSession(Pointer<Customer> owner,
                                       Pointer<Transporter> tx){
    DateTime dt;
    string f;

    DBG_INFO(0, "Starting session for \"%s\"",
             owner.ref().custId().c_str());
    _sessionOwner=owner;
    _sessionTransporter=tx;
    f=hbci()->applicationDataDir();
    f+="/";
    f+=String::num2string(country(), true, 3);
    f+="/";
    f+=bankCode();
    f+="/logs/";
    dt=DateTime::currentTime();
    f+=dt.timeString("%Y%N%D-%h%m%s");
    f+=".log";
    if (MsgXML_CreatePath(f.c_str(),
                          GWEN_PATH_FLAGS_VARIABLE |
                          GWEN_PATH_FLAGS_NAMEMUSTNOTEXIST)) {
      return Error("MessageEngineXML::startSession",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Could not create path",
                   f);
    }
    _logFile=f;
    return Error();
  }


  Error MessageEngineXML::endSession(){
    Error err;

    DBG_INFO(0, "Ending session");
    if (_dialogOpen) {
      // there still is an open dialog
      hbci()->interactor().ref().actionStarted(ACT_CLOSEDIALOG,
                                               "Closing dialog");
      err=_closeDialog();
      if (!err.isOk()) {
        DBG_WARN(0,"Could not close dialog (%s)",
                 err.errorString().c_str());
      }
      _dialogOpen=false;
      _sessionOwner=0;
      _sessionTransporter=0;
    }
    _logFile="";
    return Error();
  }

}; // namespace


/*
 * C-Wrappers for the MessageEngine. These are called by MsgEngineHBCI.
 */
extern "C" {
  int HBCI_MsgEngineXML_ReadBinType(HBCI_MessageEngineXML *e,
				    GWEN_XMLNODE *node,
                                    GWEN_DB_NODE *gr,
                                    GWEN_BUFFER *dbuf) {
    assert(e);
    return e->readBinType(node, gr, dbuf);
  }

  int HBCI_MsgEngineXML_WriteBinType(HBCI_MessageEngineXML *e,
				     GWEN_XMLNODE *node,
                                     GWEN_DB_NODE *gr,
                                     GWEN_BUFFER *dbuf) {
    assert(e);
    return e->writeBinType(node, gr, dbuf);
  }

  int HBCI_MsgEngineXML_TypeRead(HBCI_MessageEngineXML *e,
                                 GWEN_BUFFER *msgbuf,
                                 GWEN_XMLNODE *node,
                                 GWEN_BUFFER *vbuf,
                                 char escapeChar,
                                 const char *delimiters){
    assert(e);
    return e->typeRead(msgbuf, node, vbuf, escapeChar, delimiters);
  }

  int HBCI_MsgEngineXML_TypeWrite(HBCI_MessageEngineXML *e,
                                  GWEN_BUFFER *gbuf,
                                  GWEN_BUFFER *data,
                                  GWEN_XMLNODE *node){
    assert(e);
    return e->typeWrite(gbuf, data, node);
  }

  GWEN_DB_VALUETYPE HBCI_MsgEngineXML_TypeCheck(HBCI_MessageEngineXML *e,
                                                const char *tname){
    assert(e);
    return e->typeCheck(tname);
  }

  const char *HBCI_MsgEngineXML_GetCharValue(HBCI_MessageEngineXML *e,
                                             const char *name,
                                             const char *defValue){
    assert(e);
    return e->getCharValue(name, defValue);
  }

  int HBCI_MsgEngineXML_GetIntValue(HBCI_MessageEngineXML *e,
                                    const char *name,
                                    int defValue){
    assert(e);
    return e->getIntValue(name, defValue);
  }


} // extern "C"



