/***************************************************************************
  $RCSfile: msgxmljob.cpp,v $
  -------------------
  cvs         : $Id: msgxmljob.cpp,v 1.11 2004/01/19 18:53:50 aquamaniac Exp $
  begin       : Mon Jul 07 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "msgxmljob.h"
#include <gwenhywfar/text.h>
#include <gwenhywfar/debug.h>


namespace HBCI {


  MsgJobXML::MsgJobXML(MessageEngineXML *engine,
                       Pointer<Customer> cu,
		       const string &name,
		       const string &accountCode,
                       const string &secmode,
		       GWEN_DB_NODE *cfg,
                       GWEN_XMLNODE *node)
  :MsgJob(cu, name, accountCode)
  ,_engine(engine)
  ,_jobData(cfg)
  ,_node(node)
  ,_secmode(secmode)
  ,_firstseg(0)
  ,_lastseg(0)
  ,_signsNeeded(1)
  ,_jobsPerMessage(1)
  ,_attachable(false)
  ,_single(true)
  ,_needresponse(true){
    // get some data from cfg and node
    _jobsPerMessage=GWEN_DB_GetIntValue(_jobData,
                                        "params/jobspermsg", 0,
                                        1);
    _signsNeeded=GWEN_DB_GetIntValue(_jobData,
                                     "params/jobspermsg", 0,
                                     1);
    _attachable=atoi(GWEN_XMLNode_GetProperty(_node, "attachable", "0"));
    _description=GWEN_XMLNode_GetProperty(_node, "description", "");
    _dlg=atoi(GWEN_XMLNode_GetProperty(_node, "dlg", "0"));
    _single=atoi(GWEN_XMLNode_GetProperty(_node, "single", "0")) || _dlg;
    _needresponse=atoi(GWEN_XMLNode_GetProperty(_node,
                                                   "needresponse",
                                                   "1"));
  }


  MsgJobXML::~MsgJobXML(){
    GWEN_DB_Group_free(_jobData);
  }

  bool MsgJobXML::hasSegment(int i) {
    return (_firstseg<=i && _lastseg>=i) || _dlg;
  }


  const char *MsgJobXML::getProperty(const string &name,
				const string &defValue){
    return GWEN_DB_GetCharValue(_jobData,
                                name.c_str(), 0,
                                defValue.c_str());
  }


  int MsgJobXML::getIntProperty(const string &name,
				int defValue){
    return GWEN_DB_GetIntValue(_jobData, name.c_str(), 0, defValue);
  }


  string MsgJobXML::getBinProperty(const string &name,
                                   const string &defValue){
    const void *d;
    unsigned int retsize;

    d=GWEN_DB_GetBinValue(_jobData, name.c_str(),
                          0,
                          (void*)defValue.data(),
                          defValue.length(),
                          &retsize);
    if (d)
      return string((char*)d, retsize);

    return defValue;
  }


  void MsgJobXML::setBinProperty(const string &name, const string &value){
    if (GWEN_DB_SetBinValue(_jobData,
                            GWEN_DB_FLAGS_DEFAULT |
                            GWEN_DB_FLAGS_OVERWRITE_VARS,
                            name.c_str(),
                            value.c_str(),
                            value.length())) {
      throw Error("MsgJobXML::setBinPropertyt()",
                  ERROR_LEVEL_NORMAL,
                  HBCI_ERROR_CODE_UNKNOWN,
                  ERROR_ADVISE_DONTKNOW,
                  "Error setting binary value.",
                  name);
    }
  }


  void MsgJobXML::setIntProperty(const string &name, int value){
    GWEN_DB_SetIntValue(_jobData,
                        GWEN_DB_FLAGS_DEFAULT |
                        GWEN_DB_FLAGS_OVERWRITE_VARS,
                        name.c_str(), value);
  }


  void MsgJobXML::setProperty(const string &name, const string &value){
    GWEN_DB_SetCharValue(_jobData,
                         GWEN_DB_FLAGS_DEFAULT |
                         GWEN_DB_FLAGS_OVERWRITE_VARS,
                         name.c_str(), value.c_str());
  }


  void MsgJobXML::addRequestData(const string &path, GWEN_DB_NODE *gr){
    GWEN_DB_NODE *tgr;

    tgr=GWEN_DB_GetGroup(_jobData,
                         GWEN_DB_FLAGS_DEFAULT,
                         "request");
    assert(tgr);
    if (!path.empty()) {
      tgr=GWEN_DB_GetGroup(tgr,
                           GWEN_DB_FLAGS_DEFAULT |
                           GWEN_PATH_FLAGS_CREATE_GROUP,
                           path.c_str());
      assert(tgr);
    }
    GWEN_DB_AddGroupChildren(tgr, gr);
  }


  bool MsgJobXML::needsSignature() const{
    const char *p;

    p=GWEN_XMLNode_GetProperty(_node, "sign", "1");
    return atoi(p);
  }


  bool MsgJobXML::needsEncryption() const{
    const char *p;

    p=GWEN_XMLNode_GetProperty(_node, "crypt", "1");
    return atoi(p);
  }


  const list<MsgJob::Result> &MsgJobXML::results() const{
    return _results;
  }


  void MsgJobXML::addSigner(Pointer<Customer> cu){
    list<Pointer<Customer> >::iterator it;

    for (it=_signers.begin(); it!=_signers.end(); it++) {
      if ((*it).ref().custId()==cu.ref().custId()) {
	DBG_WARN(0, "Signer already in list (%s: %s)",
                 name().c_str(), cu.ref().custId().c_str());
	return;
      }
    } /* for */
    _signers.push_back(cu);
  }

  const list<Pointer<Customer> > &MsgJobXML::signers() const {
    return _signers;
  }

  void MsgJobXML::addResult(const Result &r){
    _results.push_back(r);
  }


  void MsgJobXML::removeResults(unsigned int code){
    list<Result>::iterator it;
    bool removed;

    for (;;) {
      removed=false;
      for (it=_results.begin(); it!=_results.end(); it++) {
        if ((*it).resultCode()==(int)code) {
          _results.erase(it);
          removed=true;
          break;
        }
      } // for
      if (!removed)
        break;
    } // for
  }



};

