/***************************************************************************
 $RCSfile: swiftparser.cpp,v $
                             -------------------
    cvs         : $Id: swiftparser.cpp,v 1.14 2004/04/29 08:01:02 cstim Exp $
    begin       : Sun Aug 05 2001
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de


 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


/*
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef __declspec
# if BUILDING_DLL
#  define DLLIMPORT __declspec (dllexport)
# else /* Not BUILDING_DLL */
#  define DLLIMPORT __declspec (dllimport)
# endif /* Not BUILDING_DLL */
#else
# define DLLIMPORT
#endif

#include <list>
#include <string>
#include <stdio.h>
#include <iostream>

#include "hbcistring.h"
#include "swiftparser.h"

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>

namespace HBCI {


string SWIFTparser::_normalize(const string &s) {
  unsigned int i;
  string result;

  i=0;
  while(i<s.length()) {
    if (s.substr(i,2)=="@@") {
      result+=13;
      result+=10;
      i++; // skip both
    }
    else
      result+=s[i];
    i++;
  } // while
  return result;
}


string SWIFTparser::nextTAG(const string &str, unsigned int pos){
  unsigned int lpos, spos;


  spos=pos;
  while(pos<str.length()) {
    /* The end of a tag is indicated by:
     * - sequence <CR> <LF> ':'
     *   that would mean we found the beginning of the next tag
     * - sequence <CR> <LF> <CR>
     *   because a S.W.I.F.T. document is terminated by a <CR> <LF> <->
     *   sequence the given sequence means that the end of the document
     *   is reached which implies that the tag itself ends, too ;-)
     */
    if (str.at(pos)==13) {
      /* check for sequences discussed above */
      lpos=pos;
      if (++lpos<str.length())
	if (str.at(lpos)==10)
	  if (++lpos<str.length())
	    if ((str.at(lpos)=='-') || (str.at(lpos)==':'))
	      return str.substr(spos,lpos-spos);
    } // if
    pos++;
  } // while
  /* we did not find any of the sequences discussed above, so this could not
   * be a valid swift document.
   * well, that would normally be handled this way ;-)
   * but since some institutes do not comply to SWIFT we have to
   * lighten up a bit. So if we did not find any of the sequences
   * discussed above we simply return the rest of the string given.
   */
  return str.substr(spos);
}



string SWIFTparser::nextFIELD(const string &str, unsigned int pos){
  unsigned int startpos;
  unsigned int endpos;

  if (str.at(pos)!='?')
    return "";

  startpos=pos;

  // skip possible CRLF after "?"
  pos++;
  while(pos<str.length()) {
    if ((unsigned char)(str.at(pos))>31)
      break;
    pos++;
  }
  if (pos>=str.length())
    return "";

  // now find end
  endpos=pos;
  while(endpos<str.length()) {
    //if ((str.at(endpos)==13) || (str.at(endpos)=='?'))
    if (str.at(endpos)=='?')
      return str.substr(startpos,endpos-startpos);
    endpos++;
  } // while
  /* we found neither a '?' nor a <CR> so this can not
   * be a valid swift document. */
  return str.substr(startpos);
}



string SWIFTparser::tagId(const string &tag){
  unsigned int pos;

  /* tag must start with a colon, return if it doesnt */
  if (tag.at(0)!=':')
    return "";

  pos=1; // skip first colon
  /* now search for the second colon */
  while(pos<tag.length()) {
    if (tag.at(pos)==':')
      /* create new string containing only the id */
      return tag.substr(1,pos-1);
    pos++;
  }
  /* didnt find second colon because of a bad tag */
  return "";
}


string SWIFTparser::tagContent(const string &tag){
  unsigned int pos;

  /* tag must start with a colon, return if it doesnt */
  if (tag.at(0)!=':')
    return "";

  pos=1; // skip first colon
  /* now search for the second colon */
  while (pos<tag.length()) {
    if (tag.at(pos)==':') {
      // we have it ;-) content begins directly after this
      pos++;
      if ((pos+2)<tag.length())
	return tag.substr(pos,tag.length()-pos-2);
      else
	return "";
    }
    pos++;
  }
  return "";
}



int SWIFTparser::fieldId(const string &field, unsigned int &pos){
  string tmp;

  /* tag must start with a question mark, return if it doesnt */
  if (field.at(pos)!='?')
    return -1;
  pos++;

  /* ------------------------------------ get first character after "?" */
  /* skip possible CRLF */
  while(pos<field.length()) {
    if ((unsigned char)(field.at(pos))>31)
      break;
    pos++;
  }
  if (pos>=field.length())
    return -1;
  tmp=field.at(pos);
  pos++;

  /* ----------------------------------- get second character after "?" */
  /* skip possible CRLF */
  while(pos<field.length()) {
    if ((unsigned char)(field.at(pos))>31)
      break;
    pos++;
  }
  if (pos>=field.length())
    return -1;
  /* get next char */
  tmp+=field.at(pos);
  pos++;
  /* convert them into an integer */
  return atoi(tmp.c_str());
}



int SWIFTparser::fieldId(const string &tag){
  unsigned int pos;

  pos=0;
  return fieldId(tag,pos);
}


string SWIFTparser::fieldContent(const string &field){
  unsigned int lpos;
  string result;

  lpos=0;
  // seek position of data begin
  if (-1==fieldId(field,lpos))
    return "";
  //field=field.substr(lpos);
  //lpos=0;

  // search for a <CR> <LF> sequence, because we want to exclude them
  while(lpos<field.length()) {
    if (field.at(lpos)=='?')
      break;
    // skip CR/LF in fields
    if ((unsigned char)(field.at(lpos))>31)
      result+=field.at(lpos);
    lpos++;
  }
  return result;
}



bool SWIFTparser::_mt940_61(const string &tc, GWEN_DB_NODE *cfg,
			    const string &curr){
  string tmp;
  unsigned int pos;
  unsigned int lpos;
  double d;
  Date td;
  Date vd;
  Value val;
  bool debit;
  string currency;

  pos=0;

  // get valuta date
  if (pos+6>=tc.length()) {
    // bad string
    DBG_INFO(0, "Bad data");
    return false;
  }
  tmp=tc.substr(pos,6);
  vd=Date(tmp,2);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "ValutaDate", vd.toString().c_str());
  pos+=6;

  // check for transaction date
  if (pos>=tc.length())
    // bad string
    return false;
  if (tc.at(pos)<='9' && tc.at(pos)>='0') {
    // get transaction date
    if (pos+4>=tc.length())
      // bad string
      return false;
    tmp=tc.substr(pos,4);
    td=Date(tmp,0); // year not given !
    td=Date(td.day(),
	    td.month(),
	    vd.year()); // take year from valuta date
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "date", td.toString().c_str());
    if (!td.isValid()) {
      DBG_INFO(0, "Bad date");
      return false;
    }
    pos+=4;
  }

  // get credit/debit mark
  if (pos+2>=tc.length()) {
    // bad string
    DBG_INFO(0, "Bad data");
    return false;
  }
  if (tc.at(pos)=='R') {
    tmp=tc.substr(pos,2);
  }
  else
    tmp=tc.substr(pos,1);
  // D and RC are debit...
  debit=((tmp=="D") || (tmp=="RC"));
  pos+=tmp.length();

  // check if there is  anything but a number
  if (pos>=tc.length()) {
    // bad string
    DBG_INFO(0, "Bad data");
    return false;
  }

  // check for third character of currency
  if (!isdigit(tc[pos])) {
    if (tc[pos]!='M' && tc[pos]!='R') {
      fprintf(stderr,
	      "Warning[MT94X]: Bad third character for currency (%c)\n",
              tc[pos]);
    }
    pos++;
  }

  // read value
  // find 'N'
  lpos=pos;
  while(pos<tc.length()) {
    if (tc.at(pos)=='N')
      break;
    pos++;
  }
  if (pos>=tc.length()) {
    // did not find the f*cking "N"
    DBG_INFO(0, "Bad data");
    return false;
  }
  tmp=tc.substr(lpos,pos-lpos);
  d=String::string2double(tmp);
  if (debit)
    d=-d;
  val=Value(d,curr);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "value/value", val.toString(true).c_str());
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "value/currency", val.getCurrency().c_str());

  pos++;

  // read the transaction key (3 chars)
  if (pos+2>=tc.length()) {
    // bad string
    DBG_INFO(0, "Bad data");
    return false;
  }
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "key", tc.substr(pos,3).c_str());
  pos+=3;

  // find "//" (if there is a bank reference) or <CR> (if there is more information)
  lpos=pos;
  while(pos<tc.length()) {
    if (tc.at(pos)=='/') {
      // does another "/" follow ?
      if (pos+1<tc.length())
	if (tc.at(pos+1)=='/')
	  // there is a "/", so this is the end of the customer reference
	  break;
    }
    if (tc.at(pos) == 13)
      break;
    pos++;
  } // while
  if (pos>lpos)
    GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "custref", tc.substr(lpos,pos-lpos).c_str());

  if (pos>=tc.length())
    // nothing follows, so we have finished
    return true;

  if (pos+1>=tc.length()) {
    // nothing more follows after that but it should
    DBG_INFO(0, "Bad data");
    return false;
  }

  if ((tc.at(pos)=='/') &&
      (tc.at(pos+1) == '/')) {
    // '//' means bank reference existing; OTOH one '/' belongs to
    // the additional data.
    pos+=2; // skip both '/'
    if (pos >= tc.length()) {
      DBG_INFO(0, "Bad data");
      return false; // nothing follows but should
    }

    lpos = pos;
    while (pos<tc.length()){
      if (tc.at(pos)== 13) // look for <CR>
	break;
      pos++;
    }
    if (pos>lpos)
      GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                           "bankref", tc.substr(lpos,pos-lpos).c_str());
    /* This used to say "custref" but that seemed to be an error. */
  }
  pos++;
  if (pos>=tc.length())
    // nothing follows, so we have finished
    return true;

  DBG_INFO(0, "Additional data: \"%s\"",
           tc.c_str()+pos);

  // otherwise there is some additional data to parse (it SUCKS)
  // need <LF> <"/"> sequence
  if (pos+2>=tc.length()){
    // nothing follows, but should
    DBG_INFO(0, "Bad data");
    return false;
  }

  // <LF> should follow
  if (tc.at(pos)!=10) {
    // false if it doesnt
    DBG_INFO(0, "Bad data");
    return false;
  }
  pos++;

  // handle "/OCMT/EUR20,00"
  if (tc.at(pos++)!='/') {
    DBG_INFO(0, "Bad data");
    return false;
  }

  if (pos>=tc.length()){
    // nothing follows, but should
    DBG_INFO(0, "Bad data");
    return false;
  }

  // read original value
  lpos=tc.find("/", pos);
  if (lpos==string::npos) {
    DBG_INFO(0, "Bad data");
    return false;
  }
  tmp=tc.substr(pos, lpos-pos);
  if (tmp!="OCMT") {
    DBG_INFO(0, "Bad subfield: \"%s\" (expected \"OCMT\")", tmp.c_str());
    return false;
  }
  pos=lpos;
  pos++;
  if (pos+3>=tc.length()){
    // preliminary end
    DBG_INFO(0, "Bad data");
    return false;
  }
  // get currency
  currency=tc.substr(pos, 3);
  pos+=3;
  // get value
  lpos=tc.find("/", pos);
  if (lpos==string::npos) {
    lpos=tc.length();
    tmp=tc.substr(pos);
  }
  else
    tmp=tc.substr(pos, lpos-pos);
  d=String::string2double(tmp);
  if (debit)
    d=-d;
  val=Value(d,currency);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
		       "origvalue/value", val.toString(true).c_str());
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
		       "origvalue/currency", val.getCurrency().c_str());
  pos=lpos;
  pos++;
  if (pos>=tc.length()){
    // done
    return true;
  }
  while (pos<tc.length()){
    if (tc.at(pos)!='/') // look for non-'/'
      break;
    pos++;
  }
  if (pos>=tc.length()){
    // done
    return true;
  }

  // read charges
  lpos=tc.find("/", pos);
  if (lpos==string::npos) {
    DBG_INFO(0, "Bad data");
    return false;
  }
  tmp=tc.substr(pos, lpos-pos);
  if (tmp!="CHGS") {
    DBG_INFO(0, "Bad subfield: \"%s\" (expected \"CHGS\")", tmp.c_str());
    return false;
  }
  pos=lpos;
  pos++;
  if (pos+3>=tc.length()){
    // preliminary end
    DBG_INFO(0, "Bad data");
    return false;
  }
  // get currency
  currency=tc.substr(pos, 3);
  pos+=3;
  // get value
  lpos=tc.find("/", pos);
  if (lpos==string::npos) {
    lpos=tc.length();
    tmp=tc.substr(pos);
  }
  else
    tmp=tc.substr(pos, lpos-pos);
  d=String::string2double(tmp);
  val=Value(d,currency);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
		       "charges/value", val.toString(true).c_str());
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
		       "charges/currency", val.getCurrency().c_str());
  pos=lpos;
  pos++;
  if (pos>=tc.length()){
    // done
    return true;
  }

  // now start reading additional data
  return true;
}



bool SWIFTparser::_mt940_86(const string &tc, GWEN_DB_NODE *cfg){
  unsigned int pos;
  string field;
  int fid;   // field id
  string fd; // field content
  string tmp;

  // check if tag size is valid (must at lease have a size of 3)
  if (tc.length()<3)
    return false;

  pos=0;
  // read transaction code
  if (pos+3>=tc.length())
    return false;
  tmp=tc.substr(0,3);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
		  "textkey", tmp.c_str());

  pos=3;

  // check if there are fields
  if (pos>tc.length())
    // no fields, finished
    return true;

  // otherwise parse all fields
  do {
    field=SWIFTparser::nextFIELD(tc,pos);
    pos+=field.length();
    if (!field.empty()) {
      int fdi;

      // handle the field:
      fid=fieldId(field);
      fd=fieldContent(field);
      // cut off trailing blanks
      for (fdi=fd.length()-1; fdi>=0; fdi--) {
	if (!isspace(fd[fdi]))
	  break;
      }
      if (fdi>=0)
	fd=fd.substr(0, fdi+1);
      else
	fd="";
      if (!fd.empty()) {
	switch(fid) {
	case 0:   // transaction text
	  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
			  "text", fd.c_str());
	  break;
  
	case 10:  // primanota
	  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
			  "primanota", fd.c_str());
	  break;
  
	case 20:  // Y
	case 21:
	case 22:
	case 23:
	case 24:
	case 25:
	case 26:
	case 27:
	case 28:
	case 29:
	case 60:
	case 61:
	case 62:
	case 63:  // description
	  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "purpose", fd.c_str());
	  break;
  
	case 30:  // partners institute code
	  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
			  "otherAccount/BankCode", fd.c_str());
	  break;
  
	case 31:  // partners account id
	  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
			  "otherAccount/AccountId", fd.c_str());
	  break;
  
	case 32:  // Y
	case 33:  // partners name
	  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_DEFAULT,
			  "othername", fd.c_str());
	  break;
  
	default:  // simply ignore all other fields ;-)
	  break;
	} // switch
      } // if !fd.empty()
    } // if field not empty
  } while((!(field.empty())) && pos<tc.length());
  return true;
}



bool SWIFTparser::_mt940_25(const string &tc, string &instCode, string &id){
  unsigned int pos;

  pos=0;
  // first read the institute code (BLZ)
  while(pos<tc.length()) {
    if (tc.at(pos)=='/'){
      // store our institute code
      instCode=tc.substr(0,pos);
      break;
    } // if
    pos++;
  } // while

  // then read the account id
  pos++; // skip "/"
  if (pos>=tc.length())
    // no id
    return false;
  id=tc.substr(pos);
  return true;
}



bool SWIFTparser::_mt940_60_62(const string &tc,
			       GWEN_DB_NODE *cfg,
			       string &curr) {
  string tmp;
  unsigned int pos;
  double d;
  bool debit;
  Date bd;
  Value val;

  pos=0;

  // get credit/debit mark
  if (pos>=tc.length())
    // bad string
    return false;
  debit=tc.at(pos)=='D'?true:false;
  pos++;

  // read date
  bd=Date(tc.substr(pos, 6), 2);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                  "date", bd.toString().c_str());
  pos+=6;

  // read currency (3 digit)
  if (pos+3>=tc.length())
    // bad string
    return false;
  curr=tc.substr(pos,3);
  pos+=3;

  // read value
  if (pos>=tc.length())
    // bad string
    return false;

  tmp=tc.substr(pos,tc.length()-pos);
  d=String::string2double(tmp);
  if (debit)
    d=-d;
  val=Value(d, curr);
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "value/value", val.toString(true).c_str());
  GWEN_DB_SetCharValue(cfg, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "value/currency", val.getCurrency().c_str());

  return true;
}



bool SWIFTparser::readMT940(const string &mt940raw,
			    GWEN_DB_NODE *cfg,
                            unsigned int &pos) {
  string ourInstCode;
  string ourId;
  string tag;
  string id, id2;
  string content;
  string curr;
  bool endofcycle;
  bool hasTData;
  string mt940record;
  GWEN_DB_NODE *currTrans = 0;
  GWEN_DB_NODE *tmpcfg;

  DBG_INFO(0, "SWIFT: Started handling mt940 (pos=%d of %d).\n",
           pos,mt940record.length());

  endofcycle=false;
  hasTData=false;

  /* a SWIFT document begins with a <CR> <LF> sequence,
   * and ends with a <CR><LF><-> sequence.
   * well, some institutes do NOT comply to this standard, so I had to
   * lighten up.
   */

  // ok, lets parse ;-)
  mt940record=_normalize(mt940raw);
  while (pos<mt940record.length()) {
    DBG_INFO(0, "SWIFT: Started loop (pos=%d of %d)",
             pos,mt940record.length());

    // first skip CRLF and "-" at beginning
    while (pos<mt940record.length()) {
      if ((mt940record.at(pos)!=13) &&
	  (mt940record.at(pos)!=10) &&
	  (mt940record.at(pos)!='-'))
	break;
      pos++;
    } // while
    if (pos>=mt940record.length()) {
      // we are through
      DBG_INFO(0, "Record done");
      break;
    }

    // copy the current TAG into a tmp var
    tag=nextTAG(mt940record,pos);
    if (tag.empty()) {
      DBG_INFO(0, "SWIFT: TAG is empty when it shouln't. (pos=%d)",
               pos);
      return false;
    }
    // advance TAG-pointer to the next TAG
    pos+=tag.length();
    // get TAGs id and content
    id=tagId(tag);
    id2=id.substr(0,2);
    content=tagContent(tag);

    // now parse the tag
    if (id=="25") {
      if (!_mt940_25(content,ourInstCode,ourId)) {
        DBG_INFO(0, "SWIFT: Error in mt940_25");
        return false;
      }
    }
    else if (id=="61") {
      // update and store the current Transaction
      if (hasTData) {
        DBG_INFO(0, "Adding transaction");
	GWEN_DB_SetCharValue(currTrans, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"ourAccount/BankCode", ourInstCode.c_str());
	GWEN_DB_SetCharValue(currTrans, GWEN_DB_FLAGS_OVERWRITE_VARS,
			"ourAccount/AccountId", ourId.c_str());
      }
      // allow further 86 tags
      endofcycle=false;
      currTrans=GWEN_DB_GetGroup(cfg,
                                 GWEN_PATH_FLAGS_CREATE_GROUP,
                                 "transaction");
      if (!_mt940_61(content,currTrans, curr)) {
        DBG_INFO(0, "SWIFT: Error in mt940_61");
	return false;
      }
      hasTData=true;
    }
    else if (id=="86") {
      if (!endofcycle) {
	if (!_mt940_86(content,currTrans)){
          DBG_INFO(0, "SWIFT: Error in mt940_86");
          return false;
	}
      } // if not endofcycle
    } // :86:
    else if (id2=="60" || id2=="62") {
      // if is it the closing saldo ?
      if (id2=="62") {
	// yes,  so ignore further ":86:" tags
	endofcycle=true;
	// read the closing saldo
        tmpcfg=GWEN_DB_GetGroup(cfg,
                                GWEN_PATH_FLAGS_CREATE_GROUP,
                                "endsaldo");
      }
      else {
        tmpcfg=GWEN_DB_GetGroup(cfg,
                                GWEN_PATH_FLAGS_CREATE_GROUP,
                                "startsaldo");
      }
      if (!_mt940_60_62(content,tmpcfg, curr)){
        DBG_INFO(0, "SWIFT: Error in mt940_62");
        return false;
      }
    }
  } // while

  // update and store last Transaction if any
  if (hasTData) {
    DBG_INFO(0, "Adding transaction");
    GWEN_DB_SetCharValue(currTrans, GWEN_DB_FLAGS_OVERWRITE_VARS,
		    "ourAccount/BankCode", ourInstCode.c_str());
    GWEN_DB_SetCharValue(currTrans, GWEN_DB_FLAGS_OVERWRITE_VARS,
		    "ourAccount/AccountId", ourId.c_str());
  }

  // thats it
  DBG_INFO(0, "Done");
  return true;
}



bool SWIFTparser::readMT942(const string &mt942raw,
			    GWEN_DB_NODE *cfg,
                            unsigned int &pos) {
  return readMT940(mt942raw, cfg, pos);
}


} // namespace HBCI






