/***************************************************************************
  $RCSfile: xptcp.cpp,v $
  -------------------
  cvs         : $Id: xptcp.cpp,v 1.6 2004/03/17 09:53:00 aquamaniac Exp $
  begin       : Sat Jul 05 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "xptcp.h"
#include <gwenhywfar/debug.h>
#include <gwenhywfar/inetsocket.h>
#include <gwenhywfar/text.h>

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>


namespace HBCI {

  TransporterTCP::TransporterTCP(const string &address,
				 const string & port,
				 int timeout)
  :Transporter(address, port, timeout)
  ,_lastReadPos(0)
  ,_isOpen(false)
  ,_socket(0){

  }


  TransporterTCP::~TransporterTCP(){
  }


  TransportType TransporterTCP::transportType() const {
    return TransportTypeTCP;
  }


  Error TransporterTCP::startOpen(){
    GWEN_INETADDRESS *addr;
    GWEN_ERRORCODE err;

    _socket=GWEN_Socket_new(GWEN_SocketTypeTCP);
    err=GWEN_Socket_Open(_socket);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      return Error("TransporterTCP::startOpen",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error opening socket.");
    }
    err=GWEN_Socket_SetBlocking(_socket,0);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      return Error("TransporterTCP::startOpen",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error setting socket to nonblocking mode.");
    }

    addr=GWEN_InetAddr_new(GWEN_AddressFamilyIP);
    err=GWEN_InetAddr_SetAddress(addr, address().c_str());
    if (!GWEN_Error_IsOk(err))
      err=GWEN_InetAddr_SetName(addr, address().c_str());
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      GWEN_InetAddr_free(addr);
      return Error("TransporterTCP::startOpen",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Error setting address.");
    }
    err=GWEN_InetAddr_SetPort(addr, atoi(port().c_str()));
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      GWEN_InetAddr_free(addr);
      return Error("TransporterTCP::startOpen",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error setting port.");
    }

    err=GWEN_Socket_Connect(_socket, addr);
    GWEN_InetAddr_free(addr);
    // not yet finished or real error ?
    if (!GWEN_Error_IsOk(err)) {
      if (GWEN_Error_GetType(err)!=GWEN_Error_FindType("Socket") ||
          GWEN_Error_GetCode(err)!=GWEN_SOCKET_ERROR_IN_PROGRESS) {
        DBG_ERROR_ERR(0, err);
        // real error, so return that error
        return Error("TransporterTCP::startOpen",
                     ERROR_LEVEL_NORMAL,
                     HBCI_ERROR_CODE_UNKNOWN,
                     ERROR_ADVISE_DONTKNOW,
                     "Error on connect.");
      }
    }

    return Error();
  }


  Error TransporterTCP::checkOpen(){
    GWEN_ERRORCODE err;
    GWEN_SOCKETSET *wset;

    assert(_socket);
    if (_isOpen)
      return Error();
    wset=GWEN_SocketSet_new();

    err=GWEN_SocketSet_AddSocket(wset, _socket);
    if (!GWEN_Error_IsOk(err)) {
      DBG_INFO_ERR(0, err);
      return Error("TransporterTCP::checkOpen",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Error on GWEN_SocketSet_AddSocket.");
    }
    err=GWEN_Socket_Select(0, wset, 0, timeout());
    GWEN_SocketSet_free(wset);
    if (!GWEN_Error_IsOk(err)) {
      if (GWEN_Error_GetType(err)==GWEN_Error_FindType("Socket")) {
        if (GWEN_Error_GetCode(err)==GWEN_SOCKET_ERROR_TIMEOUT) {
          DBG_INFO_ERR(0, err);
          return Error("TransporterTCP::checkOpen",
                       ERROR_LEVEL_NORMAL,
                       HBCI_ERROR_CODE_SOCKET_ERROR_TIMEOUT,
                       ERROR_ADVISE_RETRY,
                       "Timeout.");
        }
        else if (GWEN_Error_GetCode(err)==GWEN_SOCKET_ERROR_INTERRUPTED) {
          return Error("TransporterTCP::checkOpen",
                       ERROR_LEVEL_NORMAL,
                       HBCI_ERROR_CODE_SOCKET_ERROR_INTERRUPT,
                       ERROR_ADVISE_RETRY,
                       "Interrupted.");
        }
        else {
          DBG_ERROR_ERR(0, err);
          return Error("TransporterTCP::checkOpen",
                       ERROR_LEVEL_NORMAL,
                       HBCI_ERROR_CODE_UNKNOWN,
                       ERROR_ADVISE_DONTKNOW,
                       "Error on select.");
        }
      } /* if socket error */
      else {
        DBG_ERROR_ERR(0, err);
        return Error("TransporterTCP::checkOpen",
                     ERROR_LEVEL_NORMAL,
                     HBCI_ERROR_CODE_UNKNOWN,
                     ERROR_ADVISE_DONTKNOW,
                     "Error on select.");
      }
    } /* if error */

    // get socket error
    err=GWEN_Socket_GetSocketError(_socket);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      return Error("TransporterTCP::checkOpen",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Error on GWEN_Socket_GetSocketError.");
    }

    // make socket blocking again
    err=GWEN_Socket_SetBlocking(_socket,1);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      return Error("TransporterTCP::checkOpen",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Error on GWEN_Socket_GetSetBlocking.");
    }

    _isOpen=true;
    DBG_INFO(0, "Connected");
    return Error();
  }


  bool TransporterTCP::isOpen(){
    return _isOpen;
  }


  Error TransporterTCP::close(){
    if (_socket) {
      GWEN_ERRORCODE err;

      err=GWEN_Socket_Close(_socket);
      GWEN_Socket_free(_socket);
      _socket=0;
      _isOpen=false;
      if (!GWEN_Error_IsOk(err))
	return Error("TransporterTCP::close",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_UNKNOWN,
		     ERROR_ADVISE_DONTKNOW,
		     "Error on close.");
    }
    return Error();
  }


  string TransporterTCP::readBytes(int maxb){
    GWEN_ERRORCODE err;
    char buffer[1024];
    int s;
    string result;

    DBG_VERBOUS(0, "%d bytes wanted (timeout is %d)", maxb, timeout());
    err=GWEN_Socket_WaitForRead(_socket, timeout());
    if (!GWEN_Error_IsOk(err)) {
      if (GWEN_Error_GetType(err)==GWEN_Error_FindType("Socket")) {
	if (GWEN_Error_GetCode(err)==GWEN_SOCKET_ERROR_TIMEOUT) {
	  throw Error("TransporterTCP::readBytes",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_SOCKET_ERROR_TIMEOUT,
		      ERROR_ADVISE_RETRY,
		      "Timeout.");
	}
	else if (GWEN_Error_GetCode(err)==GWEN_SOCKET_ERROR_INTERRUPTED) {
	  throw Error("TransporterTCP::readBytes",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_SOCKET_ERROR_INTERRUPT,
		      ERROR_ADVISE_RETRY,
		      "Interrupted.");
	}
	else {
          DBG_ERROR_ERR(0, err);
	  throw Error("TransporterTCP::readBytes",
		      ERROR_LEVEL_NORMAL,
		      HBCI_ERROR_CODE_UNKNOWN,
		      ERROR_ADVISE_DONTKNOW,
		      "Error on select.");
	}
      } /* if socket error */
      else {
	DBG_ERROR_ERR(0, err);
	throw Error("TransporterTCP::readBytes",
		    ERROR_LEVEL_NORMAL,
		    HBCI_ERROR_CODE_UNKNOWN,
		    ERROR_ADVISE_DONTKNOW,
		    "Error on select.");
      }
    } // if error

    s=maxb;
    if (s>(int)sizeof(buffer))
      s=sizeof(buffer);
    err=GWEN_Socket_Read(_socket, buffer, &s);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      throw Error("TransporterTCP::readBytes",
		  ERROR_LEVEL_NORMAL,
		  HBCI_ERROR_CODE_UNKNOWN,
		  ERROR_ADVISE_DONTKNOW,
		  "Error on read.");

    }
    DBG_DEBUG(0, "We have read %d bytes", s);
    result=string(buffer, s);
    return result;
  }


  Error TransporterTCP::writeBytes(const string &buffer){
    GWEN_ERRORCODE err;
    int s;

    DBG_DEBUG(0, "Writing %d bytes", buffer.length());
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
      GWEN_Text_DumpString(buffer.c_str(), buffer.length(), stderr, 1);

    err=GWEN_Socket_WaitForWrite(_socket, timeout());
    if (!GWEN_Error_IsOk(err)) {
      if (GWEN_Error_GetType(err)==GWEN_Error_FindType("Socket")) {
	if (GWEN_Error_GetCode(err)==GWEN_SOCKET_ERROR_TIMEOUT) {
	  return Error("TransporterTCP::writeBytes",
		       ERROR_LEVEL_NORMAL,
		       HBCI_ERROR_CODE_SOCKET_ERROR_TIMEOUT,
		       ERROR_ADVISE_RETRY,
		       "Timeout.");
	}
	else if (GWEN_Error_GetCode(err)==GWEN_SOCKET_ERROR_INTERRUPTED) {
	  return Error("TransporterTCP::writeBytes",
		       ERROR_LEVEL_NORMAL,
		       HBCI_ERROR_CODE_SOCKET_ERROR_INTERRUPT,
		       ERROR_ADVISE_RETRY,
		       "Interrupted.");
	}
	else {
          DBG_ERROR_ERR(0, err);
	  return Error("TransporterTCP::writeBytes",
		       ERROR_LEVEL_NORMAL,
		       HBCI_ERROR_CODE_UNKNOWN,
		       ERROR_ADVISE_DONTKNOW,
		       "Error on select.");
	}
      } /* if socket error */
      else {
	DBG_ERROR_ERR(0, err);
	  return Error("TransporterTCP::writeBytes",
		     ERROR_LEVEL_NORMAL,
		     HBCI_ERROR_CODE_UNKNOWN,
		     ERROR_ADVISE_DONTKNOW,
		     "Error on select.");
      }
    } // if error

    s=buffer.length();
    err=GWEN_Socket_Write(_socket, buffer.c_str(), &s);
    if (!GWEN_Error_IsOk(err)) {
      DBG_ERROR_ERR(0, err);
      return Error("TransporterTCP::writeBytes",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error on write.");

    }

    if (s!=(int)buffer.length())
      return Error("TransporterTCP::writeBytes",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Not all data written.");
    return Error();
  }


  string TransporterTCP::transporterTypeName() const{
    return XPTCP_TYPENAME;
  }


} // namespace


