/***************************************************************************
 $RCSfile: outbox.h,v $
                             -------------------
    cvs         : $Id: outbox.h,v 1.7 2004/01/19 20:20:34 aquamaniac Exp $
    begin       : Sat Jun 08 2002
    copyright   : (C) 2001 by Martin Preuss
    email       : openhbci@aquamaniac.de

 ***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Lesser General Public            *
 *   License as published by the Free Software Foundation; either          *
 *   version 2.1 of the License, or (at your option) any later version.    *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston,                 *
 *   MA  02111-1307  USA                                                   *
 *                                                                         *
 ***************************************************************************/



#ifndef OUTBOX_H
#define OUTBOX_H

#ifdef __cplusplus
#include <list>
#endif /* __cplusplus */

#include <openhbci2/dllimport.h>
#include <openhbci2/pointer.h>
#include <openhbci2/customer.h>
#include <openhbci2/bank.h>
#include <gwenhywfar/db.h>

#ifdef __cplusplus
/* Forward declarations */
namespace HBCI {
  class Outbox;
  class customerQueue;
  class bankQueue;
}
#endif /* __cplusplus */

#include <openhbci2/outboxjob.h>

#ifdef __cplusplus
namespace HBCI {

#ifndef DOXYGEN_HIDE

/**
 * @internal
 * @short This queue holds all jobs for one customer.
 * This class is not part of the OpenHBCI API !
 * @author Martin Preuss<martin@libchipcard.de>
 */
class DLLIMPORT customerQueue {
private:
    Pointer<Customer> _cust;
    list<Pointer<OutboxJob> > _jobs;
public:
    customerQueue(Pointer<Customer> c);
    ~customerQueue();
    void addJob(Pointer<OutboxJob> job);
    const list<Pointer<OutboxJob> > &jobs() const;
    Pointer<Customer> customer() const { return _cust;};
    void removeByStatus(OutboxJob_Status s);
    void removeByState(OutboxJob_Status s);
    void removeByResult(OutboxJob_Result s);
    bool removeJob(Pointer<OutboxJob> job);
    bool empty() const;
    int sizeByResult(OutboxJob_Result s=HBCI_JOB_RESULT_NONE) const;
    int sizeByStatus(OutboxJob_Status s=HBCI_JOB_STATUS_NONE) const;
};


/*___________________________________________________________________________
 *AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                              bankQueue
 *YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */



/**
 * @internal
 * @short This queue holds all jobs for one bank.
 * This class is not part of the OpenHBCI API !
 * @author Martin Preuss<martin@libchipcard.de>
 */
class DLLIMPORT bankQueue {
private:
    Pointer<Bank> _bank;
    list<Pointer<customerQueue> > _custQueues;
public:
    bankQueue(Pointer<Bank> b);
    ~bankQueue();
    Pointer<Bank> bank() const { return _bank;};
    const list<Pointer<customerQueue> > &customerQueues() const;
    void addJob(Pointer<Customer> c,
                Pointer<OutboxJob> job);
    void removeByStatus(OutboxJob_Status s);
    void removeByState(OutboxJob_Status s);
    void removeByResult(OutboxJob_Result s);
    bool removeJob(Pointer<OutboxJob> job);
    bool empty() const;
    int sizeByResult(OutboxJob_Result s=HBCI_JOB_RESULT_NONE) const;
    int sizeByStatus(OutboxJob_Status s=HBCI_JOB_STATUS_NONE) const;
    int customerCount() const;
};


#endif /* DOXYGEN_HIDE */



/*___________________________________________________________________________
 *AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                              Outbox
 *YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */



/**
 * @short This is a queue for the OutboxJobs. 
 *
 * If you add a job it will automatically be sorted in the right
 * bank/customer queue. This way all jobs in this queue are sorted by
 * their bank and customer. This groups the jobs in a manner that
 * allows sending them with the least possible number of customer
 * switches (i.e. different PINs to be entered).  
 *
 * @author Martin Preuss<martin@libchipcard.de> */
class DLLIMPORT Outbox {
  friend class API;
  friend class Loader;
private:
  GWEN_DB_NODE *_response;
  list<MsgQueue::Result> _results;
protected:
  void addResponse(GWEN_DB_NODE *rsp);
  void addResults(list<MsgQueue::Result> results);

  /**
   * @internal
   * Each job that gets enqueued will be assigned a unique id to.
   * To retrieve this id this method can be used. It is also used
   * by the Loader for storage.
   * @author Martin Preuss<martin@libchipcard.de>
   */
  DLLIMPORT static int nextId();

  /**
   * @internal
   * Set the next job id to be used. This should only be called
   * by Loader upon loading of the configuration file. Otherwise
   * jobids are not guaranteed to be unique.
   * @author Martin Preuss<martin@libchipcard.de>
   */
  DLLIMPORT static void setNextId(int i);

public:
    Outbox();
    ~Outbox();

    /**
     * Add a job to the queue. This job will be sorted in according to
     * the bank and customer involved.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    void addJob(Pointer<OutboxJob> job);

    /**
     * Remove all jobs with the given status from the queue.
     * @author Martin Preuss<martin@libchipcard.de>
     * @param s status to match (HBCI_JOB_STATUS_NONE will remove ALL jobs)
     */
    void removeByStatus(OutboxJob_Status s);
    /**
     * DEPRECATED -- only exists due to inconsistency in spelling. Use
     * removeByStatus() instead. 
     */
    void removeByState(OutboxJob_Status s);

    /**
     * Remove all jobs with the given result from the queue.
     * @author Martin Preuss<martin@libchipcard.de>
     * @param s result to match (HBCI_JOB_RESULT_NONE will remove ALL jobs)
     */
    void removeByResult(OutboxJob_Result s);

    /**
     * Removes a given job.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    void removeJob(Pointer<OutboxJob> job);

    /**
     * Returns the list of bank queues. There is one queue for each bank
     * involved.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    const list<Pointer<bankQueue> > &bankQueues() const;

    /**
     * Returns true if the Outbox is empty.
     */
    bool empty() const;

    /**
     * Returns the number of jobs with the given result code.
     * @author Martin Preuss<martin@libchipcard.de>
     * @param s result code to look for (HBCI_JOB_RESULT_NONE matches all).
     */
    int sizeByResult(OutboxJob_Result s=HBCI_JOB_RESULT_NONE) const;

    /**
     * Returns the number of jobs with the given status code.
     * @param s status code to look for (HBCI_JOB_STATUS_NONE matches all).
     */
    int sizeByStatus(OutboxJob_Status s=HBCI_JOB_STATUS_NONE) const;

    /**
     * Returns the number of banks involved.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    int bankCount() const;

    /**
     * Returns the number of customers involved.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    int customerCount() const;

    /**
     * Removes all jobs from the queue.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    void clear();

    /**
     * Returns a list of all jobs in the queue (of all banks together).
     * @author Martin Preuss<martin@libchipcard.de>
     */
    list<Pointer<OutboxJob> > jobs() const;

    /**
     * Returns a private copy of the response received while executing this
     * queue. The caller is responsible for freeing the GWEN_DB_NODE returned
     * by calling Config_free().
     */
    GWEN_DB_NODE *response();

    /**
     * Returns the responses received and stored in this outbox.
     */
    const list<MsgQueue::Result> &results() const { return _results;};

private:
    list<Pointer<bankQueue> > _banks;
    DLLIMPORT static int _nextId;

};

} /* namespace HBCI */
typedef HBCI::Outbox HBCI_Outbox;
extern "C" {
#else /* __cplusplus */
typedef struct HBCI_Outbox HBCI_Outbox;
#endif /* __cplusplus */

  HBCI_Outbox *HBCI_Outbox_new();
  void HBCI_Outbox_delete(HBCI_Outbox *h);

  /**
   * Add a job to the queue. This job will be sorted in according to
   * the bank and customer involved.
   *
   * The HBCI_API will take ownership of this job and will delete it
   * when it gets deleted itself, or when you call
   * clearQueueBy{Result,Status}. So please note that your pointer to
   * the OutboxJob might get invalid if you call
   * clearQueueBy{Result,Status}.
   *
   * For now no error is returned, but maybe later this method will check
   * if the given job fits into the current queue.
   * @param h The object to perform this action on.
   *
   * @param job The job to add. The HBCI_API will take ownership of
   * this job and will delete it when it gets deleted itself, or
   * when you call clearQueueBy{Result,Status}.  
   */
  void HBCI_Outbox_addJob(HBCI_Outbox *h, HBCI_OutboxJob *job);

    /**
     * Remove all jobs with the given status from the queue.
     * @author Martin Preuss<martin@libchipcard.de>
     * @param s status to match (HBCI_JOB_STATUS_NONE will remove ALL jobs)
     */
    void HBCI_Outbox_removeByStatus(HBCI_Outbox *h, OutboxJob_Status s);

    /**
     * Remove all jobs with the given result from the queue.
     * @author Martin Preuss<martin@libchipcard.de>
     * @param s result to match (HBCI_JOB_RESULT_NONE will remove ALL jobs)
     */
    void HBCI_Outbox_removeByResult(HBCI_Outbox *h, OutboxJob_Result s);

    /**
     * Removes a given job.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    void HBCI_Outbox_removeJob(HBCI_Outbox *h, HBCI_OutboxJob *job);

    /**
     * Returns true if the Outbox is empty.
     */
    int HBCI_Outbox_empty(const HBCI_Outbox *h);

    /**
     * Returns the number of jobs with the given result code.
     * @author Martin Preuss<martin@libchipcard.de>
     * @param s result code to look for (HBCI_JOB_RESULT_NONE matches all).
     */
    int HBCI_Outbox_sizeByResult(const HBCI_Outbox *h, OutboxJob_Result s);

    /**
     * Returns the number of jobs with the given status code.
     * @param s status code to look for (HBCI_JOB_STATUS_NONE matches all).
     */
    int HBCI_Outbox_sizeByStatus(const HBCI_Outbox *h, OutboxJob_Status s);

    /**
     * Returns the number of banks involved.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    int HBCI_Outbox_bankCount(const HBCI_Outbox *h);

    /**
     * Returns the number of customers involved.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    int HBCI_Outbox_customerCount(const HBCI_Outbox *h);

    /**
     * Removes all jobs from the queue.
     * @author Martin Preuss<martin@libchipcard.de>
     */
    void HBCI_Outbox_clear(HBCI_Outbox *h);

    /**
     * Returns a list of all jobs in the queue (of all banks together).
     * @author Martin Preuss<martin@libchipcard.de>
     */
    /* const list_HBCI_OutboxJob * HBCI_Outbox_jobs(const HBCI_Outbox *h); */

  /**
   * Returns a private copy of the response received while executing this
   * queue. The caller is responsible for freeing the GWEN_DB_NODE returned
   * by calling Config_free().
   */
  GWEN_DB_NODE *HBCI_Outbox_response(HBCI_Outbox *h);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* OUTBOX_H */



